Package: odbc
Title: Connect to ODBC Compatible Databases (using the DBI Interface)
Version: 1.7.0
Authors@R: c(
    person("Jim", "Hester", role = "aut"),
    person("Hadley", "Wickham", , "hadley@posit.co", role = c("aut", "cre")),
    person("Oliver", "Gjoneski", role = "aut"),
    person("Simon", "Couch", email = "simon.couch@posit.co", role = "aut"),
    person("lexicalunit", role = "cph",
           comment = "nanodbc library"),
    person("Google Inc.", role = "cph",
           comment = "cctz library"),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: A DBI-compatible interface to ODBC databases.
License: MIT + file LICENSE
URL: https://odbc.r-dbi.org, https://github.com/r-dbi/odbc,
        https://solutions.posit.co/connections/db/
BugReports: https://github.com/r-dbi/odbc/issues
Depends: R (>= 4.0.0)
Imports: bit64, blob (>= 1.2.0), cli, DBI (>= 1.0.0), hms, lifecycle,
        methods, Rcpp (>= 0.12.11), rlang (>= 1.1.0), vctrs
Suggests: connectcreds (>= 0.2.0), covr, DBItest, httr2, knitr,
        magrittr, paws.common, rmarkdown, RSQLite, snowflakeauth,
        testthat (>= 3.0.0), tibble, withr
LinkingTo: Rcpp
ByteCompile: true
Config/Needs/check: pkgbuild
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
SystemRequirements: GNU make, An ODBC3 driver manager and drivers.
Collate: 'RcppExports.R' 'aaa-odbc-data-type.R' 'connection-pane.R'
        'dbi-connection.R' 'odbc-connection.R' 'db.R' 'dbi-driver.R'
        'dbi-result.R' 'dbi-table.R' 'dbi.R' 'driver-access.R'
        'driver-bigquery.R' 'driver-databricks.R' 'driver-db2.R'
        'driver-hana.R' 'driver-hive.R' 'driver-impala.R'
        'driver-mysql.R' 'driver-netezza.R' 'driver-oracle.R'
        'driver-postgres.R' 'driver-redshift.R' 'driver-snowflake.R'
        'driver-spark.R' 'driver-sql-server.R' 'driver-sqlite.R'
        'driver-teradata.R' 'driver-vertica.R'
        'import-standalone-obj-type.R'
        'import-standalone-types-check.R' 'odbc-config.R'
        'odbc-data-sources.R' 'odbc-drivers.R' 'odbc-package.R'
        'odbc.R' 'utils.R' 'zzz.R'
VignetteBuilder: knitr
NeedsCompilation: yes
Packaged: 2026-05-05 18:45:03 UTC; simoncouch
Author: Jim Hester [aut],
  Hadley Wickham [aut, cre],
  Oliver Gjoneski [aut],
  Simon Couch [aut],
  lexicalunit [cph] (nanodbc library),
  Google Inc. [cph] (cctz library),
  Posit Software, PBC [cph, fnd]
Maintainer: Hadley Wickham <hadley@posit.co>
Repository: CRAN
Date/Publication: 2026-05-09 15:00:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(odbcListColumns,OdbcConnection)
S3method(odbcListObjectTypes,default)
S3method(odbcListObjects,OdbcConnection)
S3method(odbcPreviewObject,OdbcConnection)
export(databricks)
export(isTempTable)
export(odbc)
export(odbcDataType)
export(odbcEditDrivers)
export(odbcEditSystemDSN)
export(odbcEditUserDSN)
export(odbcListColumns)
export(odbcListConfig)
export(odbcListDataSources)
export(odbcListDrivers)
export(odbcListObjectTypes)
export(odbcListObjects)
export(odbcPreviewObject)
export(odbcSetTransactionIsolationLevel)
export(quote_value)
export(redshift)
export(snowflake)
exportClasses("DB2/AIX64")
exportClasses("Microsoft SQL Server")
exportClasses("Spark SQL")
exportClasses("Vertica Database")
exportClasses(ACCESS)
exportClasses(BigQuery)
exportClasses(DatabricksOdbcDriver)
exportClasses(HDB)
exportClasses(Hive)
exportClasses(Impala)
exportClasses(MySQL)
exportClasses(NetezzaSQL)
exportClasses(OdbcConnection)
exportClasses(OdbcDriver)
exportClasses(OdbcResult)
exportClasses(Oracle)
exportClasses(PostgreSQL)
exportClasses(Redshift)
exportClasses(RedshiftOdbcDriver)
exportClasses(SQLite)
exportClasses(Snowflake)
exportClasses(SnowflakeOdbcDriver)
exportClasses(Teradata)
exportMethods(dbAppendTable)
exportMethods(dbBegin)
exportMethods(dbBind)
exportMethods(dbClearResult)
exportMethods(dbColumnInfo)
exportMethods(dbCommit)
exportMethods(dbConnect)
exportMethods(dbDataType)
exportMethods(dbDisconnect)
exportMethods(dbExecute)
exportMethods(dbExistsTable)
exportMethods(dbFetch)
exportMethods(dbGetInfo)
exportMethods(dbGetQuery)
exportMethods(dbGetRowCount)
exportMethods(dbGetRowsAffected)
exportMethods(dbGetStatement)
exportMethods(dbHasCompleted)
exportMethods(dbIsValid)
exportMethods(dbListFields)
exportMethods(dbListTables)
exportMethods(dbQuoteIdentifier)
exportMethods(dbQuoteString)
exportMethods(dbRemoveTable)
exportMethods(dbRollback)
exportMethods(dbSendQuery)
exportMethods(dbSendStatement)
exportMethods(dbWriteTable)
exportMethods(odbcDataType)
exportMethods(show)
exportMethods(sqlCreateTable)
exportMethods(sqlData)
import(DBI)
import(methods)
import(rlang)
importFrom(Rcpp,sourceCpp)
importFrom(bit64,integer64)
importFrom(blob,blob)
importFrom(hms,hms)
importFrom(lifecycle,deprecated)
importFrom(utils,packageVersion)
useDynLib(odbc, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# odbc <a href="https://odbc.r-dbi.org"><img src="man/figures/logo.png" align="right" height="138" alt="odbc website" /></a>

<!-- badges: start -->

[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](http://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/)
[![CRAN_Status_Badge](http://www.r-pkg.org/badges/version/odbc)](https://cran.r-project.org/package=odbc)
[![R-CMD-check](https://github.com/r-dbi/odbc/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-dbi/odbc/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-dbi/odbc/branch/main/graph/badge.svg)](https://app.codecov.io/gh/r-dbi/odbc?branch=main)
<!-- badges: end -->

The goal of the odbc package is to provide a
[DBI](https://dbi.r-dbi.org/)-compliant interface to
[ODBC](https://learn.microsoft.com/en-us/sql/odbc/microsoft-open-database-connectivity-odbc)
drivers. This makes it easy to connect databases such as [SQL
Server](https://www.microsoft.com/en-us/sql-server/), 
Oracle,
[Databricks](https://www.databricks.com/), and 
Snowflake.

The odbc package is an alternative to
[RODBC](https://cran.r-project.org/package=RODBC) and
[RODBCDBI](https://cran.r-project.org/package=RODBCDBI) packages, and is
typically much faster. See `vignette("benchmarks")` to learn more.

## Overview

The odbc package is one piece of the R interface to databases with
support for ODBC:

<img src="man/figures/whole-game.png" alt="A diagram containing four boxes with arrows linking each pointing left to right. The boxes read, in order, &quot;R interface,&quot; &quot;driver manager,&quot; &quot;ODBC driver,&quot; and &quot;DBMS.&quot; The left-most box, R interface, contains three smaller components, labeled &quot;dbplyr,&quot; &quot;DBI,&quot; and &quot;odbc.&quot;" width="1489" />

Support for a given DBMS is provided by an **ODBC driver**, which
defines how to interact with that DBMS using the standardized syntax of
ODBC and SQL. Drivers can be downloaded from the DBMS vendor or, if
you’re a Posit customer, using the [professional
drivers](https://docs.posit.co/pro-drivers/).

Drivers are managed by a **driver manager**, which is responsible for
configuring driver locations, and optionally named **data sources** that
describe how to connect to a specific database. Windows is bundled with
a driver manager, while MacOS and Linux require installation of
[unixODBC](https://www.unixodbc.org/). Drivers often require some manual
configuration; see `vignette("setup")` for details.

In the **R interface**, the [DBI package](https://dbi.r-dbi.org/)
provides a front-end while odbc implements a back-end to communicate
with the driver manager. The odbc package is built on top of the
[nanodbc](https://nanodbc.github.io/nanodbc/) C++ library. To interface
with DBMSs using R and odbc:

<img src="man/figures/r-interface.png" alt="A high-level workflow for using the R interface in 3 steps. In step 1, configure drivers and data sources, the functions odbcListDrivers() and odbcListDataSources() help to interface with the driver manager. In step 2, the dbConnect() function, called with the first argument odbc(), connects to a database using the specified ODBC driver to create a connection object &quot;con.&quot; Finally, in step 3, that connection object can be passed to various functions to retrieve information on database structure, iteratively develop queries, and query data objects." width="1302" />

You might also use the [dbplyr package](https://dbplyr.tidyverse.org/)
to automatically generate SQL from your dplyr code.

## Installation

Install the latest release of odbc from CRAN with the following code:

``` r
install.packages("odbc")
```

To get a bug fix or to use a feature from the development version, you
can install the development version of odbc from GitHub:

``` r
# install.packages("pak")
pak::pak("r-dbi/odbc")
```

## Usage

To use odbc, begin by creating a database connection, which might look
something like this:

``` r
library(DBI)

con <- dbConnect(
  odbc::odbc(),
  driver = "SQL Server",
  server = "my-server",
  database = "my-database",
  uid = "my-username",
  pwd = rstudioapi::askForPassword("Database password")
)
```

(See `vignette("setup")` for examples of connecting to a variety of
databases.)

`dbListTables()` is used for listing all existing tables in a database.

``` r
dbListTables(con)
```

`dbReadTable()` will read a full table into an R `data.frame()`.

``` r
data <- dbReadTable(con, "flights")
```

`dbWriteTable()` will write an R `data.frame()` to an SQL table.

``` r
dbWriteTable(con, "iris", iris)
```

`dbGetQuery()` will submit a SQL query and fetch the results:

``` r
df <- dbGetQuery(
  con,
  "SELECT flight, tailnum, origin FROM flights ORDER BY origin"
)
```

It is also possible to submit the query and fetch separately with
`dbSendQuery()` and `dbFetch()`. This allows you to use the `n` argument
to `dbFetch()` to iterate over results that would otherwise be too large
to fit in memory.

#' @include odbc-connection.R
NULL

#' Helper method used to determine if a table identifier is that
#' of a temporary table.
#'
#' Currently implemented only for select back-ends where
#' we have a use for it (SQL Server, for example).  Generic, in case
#' we develop a broader use case.
#' @param conn OdbcConnection
#' @param name Table name
#' @param ... additional parameters to methods
#' @rdname isTempTable
#' @export
setGeneric(
  "isTempTable",
  valueClass = "logical",
  function(conn, name, ...) {
    standardGeneric("isTempTable")
  }
)

#' @rdname isTempTable
setMethod("isTempTable", c("OdbcConnection", "Id"),
  function(conn, name, ...) {
    isTempTable(conn,
      name = id_field(name, "table"),
      catalog_name = id_field(name, "catalog"),
      schema_name = id_field(name, "schema"),
      ...
    )
  }
)

#' @rdname isTempTable
setMethod("isTempTable", c("OdbcConnection", "SQL"),
  function(conn, name, ...) {
    isTempTable(conn, dbUnquoteIdentifier(conn, name)[[1]], ...)
  }
)

#' @include dbi-connection.R
NULL

#' Convenience functions for reading/writing DBMS tables
#'
#' @param conn An [OdbcConnection] object, produced by [DBI::dbConnect()].
#' @param name a character string specifying a table name. Names will be
#'   automatically quoted so you can use any sequence of characters, not
#'   just any valid bare table name.
#' @param value A data.frame to write to the database.
#' @inheritParams DBI::sqlCreateTable
#' @examples
#' \dontrun{
#' library(DBI)
#' con <- dbConnect(odbc::odbc())
#' dbListTables(con)
#' dbWriteTable(con, "mtcars", mtcars, temporary = TRUE)
#' dbReadTable(con, "mtcars")
#'
#' dbListTables(con)
#' dbExistsTable(con, "mtcars")
#'
#' # A zero row data frame just creates a table definition.
#' dbWriteTable(con, "mtcars2", mtcars[0, ], temporary = TRUE)
#' dbReadTable(con, "mtcars2")
#'
#' dbDisconnect(con)
#' }
#' @name DBI-tables
NULL

odbc_write_table <- function(conn,
                             name,
                             value,
                             overwrite = FALSE,
                             append = FALSE,
                             temporary = FALSE,
                             row.names = NULL,
                             field.types = NULL,
                             batch_rows = getOption("odbc.batch_rows", NA),
                             ...) {
  call <- caller_env()
  check_bool(overwrite, call = call)
  check_bool(append, call = call)
  check_bool(temporary, call = call)
  check_number_whole(batch_rows, allow_na = TRUE, allow_null = TRUE, call = call)
  check_row.names(row.names, call = call)
  check_field.types(field.types, call = call)
  if (append && !is.null(field.types)) {
    cli::cli_abort(
      "Cannot specify {.arg field.types} with {.code append = TRUE}.",
      call = call
    )
  }
  if (overwrite && append) {
    cli::cli_abort(
      "{.arg overwrite} and {.arg append} cannot both be {.val TRUE}.",
      call = call
    )
  }

  found <- dbExistsTableForWrite(conn, name)
  if (found && !overwrite && !append) {
    cli::cli_abort(
      "Table {toString(name)} exists in database, and both overwrite and \\
       append are {.code FALSE}.",
      call = call
    )
  }
  if (found && overwrite) {
    dbRemoveTable(conn, name)
  }

  values <- sqlData(conn, row.names = row.names, value[, , drop = FALSE])
  if (!found || overwrite) {
    dbCreateTable(
      conn = conn,
      name = name,
      fields = values,
      field.types = field.types,
      row.names = NULL,
      temporary = temporary
    )
  }
  dbAppendTable(
    conn = conn,
    name = name,
    value = values,
    batch_rows = batch_rows,
    ...,
    row.names = NULL
  )
  invisible(TRUE)
}

#' @rdname DBI-tables
#' @inheritParams DBI::dbWriteTable
#' @param overwrite Allow overwriting the destination table. Cannot be
#'   `TRUE` if `append` is also `TRUE`.
#' @param append Allow appending to the destination table. Cannot be
#'   `TRUE` if `overwrite` is also `TRUE`.
#' @param batch_rows The number of rows to retrieve. Defaults to `NA`, which
#'   is set dynamically to the minimum of 1024 and the size of the input.
#'   Depending on the database, driver, dataset and free memory, setting this
#'   to a lower value may improve performance.
#' @export
setMethod("dbWriteTable", c("OdbcConnection", "character", "data.frame"),
  odbc_write_table
)

#' @rdname DBI-tables
#' @export
setMethod("dbWriteTable", c("OdbcConnection", "Id", "data.frame"),
  odbc_write_table
)

#' @rdname DBI-tables
#' @export
setMethod("dbWriteTable", c("OdbcConnection", "SQL", "data.frame"),
  odbc_write_table
)

#' @rdname DBI-tables
#' @inheritParams DBI::dbAppendTable
#' @inheritParams DBI::dbWriteTable
#' @export
setMethod("dbAppendTable", "OdbcConnection",
  function(conn, name, value,
           batch_rows = getOption("odbc.batch_rows", NA),
           ..., row.names = NULL) {
    if (!is.null(row.names)) {
      cli::cli_abort(
        "{.arg row.names} must be {.code NULL}, not \\
         {.obj_type_friendly {row.names}}."
      )
    }

    fieldDetails <- tryCatch({
      details <- odbcConnectionColumns(conn, name, exact = TRUE)
      details$param_index <- match(details$name, colnames(value))
      details[!is.na(details$param_index) & !is.na(details$data_type), ]
    },
    error = function(e) {
      return(NULL)
    })

    if (nrow(value) > 0) {
      name <- dbQuoteIdentifier(conn, name)
      fields <- dbQuoteIdentifier(conn, colnames(value))
      nparam <- length(fields)
      params <- rep("?", nparam)

      sql <- paste0(
        "INSERT INTO ", name, " (", paste0(fields, collapse = ", "), ")\n",
        "VALUES (", paste0(params, collapse = ", "), ")"
      )
      rs <- OdbcResult(conn, sql)

      if (!is.null(fieldDetails) && nrow(fieldDetails) <= nparam) {
        result_describe_parameters(rs@ptr, fieldDetails)
      }

      values <- sqlData(conn, row.names = row.names, value[, , drop = FALSE])
      if (is.na(batch_rows)) {
        batch_rows <- NROW(value)
        if (batch_rows == 0) {
          batch_rows <- 1
        }
        batch_rows <- min(1024, batch_rows)
      }
      batch_rows <- parse_size(batch_rows)
      tryCatch(
        result_insert_dataframe(rs@ptr, values, batch_rows),
        finally = dbClearResult(rs)
      )
    }

    invisible(NA_real_)
  })

#' @rdname DBI-methods
#' @export
setMethod("sqlData", "OdbcConnection",
  function(con, value, row.names = NA, ...) {

    chk <- attr(value, ".odbc.transformed", exact = TRUE)
    if (isTRUE(chk)) return(value)

    value <- sqlRownamesToColumn(value, row.names)

    # Convert POSIXlt to POSIXct
    is_POSIXlt <- vapply(value, function(x) is.object(x) && (is(x, "POSIXlt")), logical(1))
    value[is_POSIXlt] <- lapply(value[is_POSIXlt], as.POSIXct)

    # C code takes care of atomic vectors, dates, date times, and blobs just need to coerce other objects
    is_object <- vapply(value, function(x) is.object(x) && !(is(x, "POSIXct") || is(x, "Date") || is_blob(x) || is(x, "difftime")), logical(1))
    value[is_object] <- lapply(value[is_object], as.character)

    if (nzchar(con@encoding)) {
      is_character <- vapply(value, is.character, logical(1))
      value[is_character] <- lapply(value[is_character], enc2iconv, to = con@encoding)
    }

    attr(value, ".odbc.transformed") <- TRUE
    value
  }
)

#' @rdname DBI-tables
#' @inheritParams DBI::sqlCreateTable
#' @param field.types Additional field types used to override derived types.
#' @export
setMethod("sqlCreateTable", "OdbcConnection",
  function(con,
           table,
           fields,
           row.names = NA,
           temporary = FALSE,
           ...,
           field.types = NULL) {
    check_bool(temporary)
    check_row.names(row.names)
    check_field.types(field.types)
    table <- dbQuoteIdentifier(con, table)
    fields <- createFields(con, fields, field.types, row.names)

    SQL(paste0(
      "CREATE ", if (temporary) "TEMPORARY ", "TABLE ", table, " (\n",
      "  ", paste(fields, collapse = ",\n  "), "\n)\n"
    ))
  }
)

# Helper function useful for defining custom sqlCreateTable methods.
createFields <- function(con, fields, field.types, row.names) {
  if (is.data.frame(fields)) {
    fields <- sqlRownamesToColumn(fields, row.names)
    fields <- DBI::dbDataType(con, fields)
  }
  if (!is.null(field.types)) {
    is_field <- names(field.types) %in% names(fields)
    if (!all(is_field)) {
      warning(
        sprintf(
          "Some columns in `field.types` not in the input, missing columns:\n%s",
          paste0("  - '", names(field.types)[!is_field], "'", collapse = "\n")
        ),
        call. = FALSE, immediate. = TRUE
      )
    }

    fields[names(field.types)] <- field.types
  }

  field_names <- dbQuoteIdentifier(con, names(fields))
  field_types <- unname(fields)
  paste0(field_names, " ", field_types)
}

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbListFields
#' @export
setMethod("dbListFields", c("OdbcConnection", "Id"),
  function(conn, name, ...) {
    dbListFields(
      conn,
      name = id_field(name, "table"),
      catalog_name = id_field(name, "catalog"),
      schema_name = id_field(name, "schema")
    )
  }
)

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbListFields
#' @export
setMethod("dbListFields", c("OdbcConnection", "SQL"),
  function(conn, name, ...) {
    dbListFields(conn, dbUnquoteIdentifier(conn, name)[[1]], ...)
  }
)

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbListFields
#' @param catalog_name Catalog where table is located.
#' @param schema_name Schema where table is located.
#' @param column_name The name of the column to return, the default returns all columns.
#' @export
setMethod("dbListFields", c("OdbcConnection", "character"),
  function(conn,
           name,
           catalog_name = NULL,
           schema_name = NULL,
           column_name = NULL,
           ...) {
    check_string(name)
    check_string(catalog_name, allow_null = TRUE)
    check_string(schema_name, allow_null = TRUE)
    check_string(column_name, allow_null = TRUE)

    cols <- odbcConnectionColumns(
      conn,
      name = name,
      catalog_name = catalog_name,
      schema_name = schema_name,
      column_name = column_name,
      exact = TRUE
    )
    cols[["name"]]
  }
)

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbExistsTable
#' @export
setMethod("dbExistsTable", c("OdbcConnection", "Id"),
  function(conn, name, ...) {
    dbExistsTable(
      conn,
      name = id_field(name, "table"),
      catalog_name = id_field(name, "catalog"),
      schema_name = id_field(name, "schema")
    )
  }
)

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbExistsTable
#' @export
setMethod("dbExistsTable", c("OdbcConnection", "SQL"),
  function(conn, name, ...) {
    dbExistsTable(conn, dbUnquoteIdentifier(conn, name)[[1]], ...)
  }
)

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbExistsTable
#' @export
setMethod("dbExistsTable", c("OdbcConnection", "character"),
  function(conn, name, ...) {
    check_string(name)
    df <- odbcConnectionTables(conn, name = name, ..., exact = TRUE)
    NROW(df) > 0
  }
)


#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbRemoveTable
#' @export
setMethod("dbRemoveTable", c("OdbcConnection", "character"),
  function(conn, name, ...) {
    name <- dbQuoteIdentifier(conn, name)
    dbExecute(conn, paste("DROP TABLE ", name))
    on_connection_updated(conn, name)
    invisible(TRUE)
  }
)

skip_if_no_unixodbc()

test_that("SQLServer", {
  DBItest::make_context(
    odbc(),
    test_connection_string("SQLSERVER"),
    tweaks = DBItest::tweaks(temporary_tables = FALSE),
    name = "SQLServer"
  )

  DBItest::test_getting_started(c(
    "package_name", # Not an error
    NULL
  ))
  DBItest::test_driver(c(
    "connect_bigint_integer",
    "connect_bigint_character",
    "connect_bigint_integer64",
    NULL
  ))
  DBItest::test_connection(c(
    NULL
  ))
  DBItest::test_result(c(
    "get_query_n_zero_rows",
    "get_query_n_incomplete",
    # TODO: fails with Pro Driver due to no "good" results returned
    "fetch_n_good_after_bad",
    "fetch_no_return_value", # TODO
    "clear_result_return_statement",
    "cannot_clear_result_twice_statement",
    "send_statement.*", # Invalid CTAS syntax
    "execute_atomic", # Invalid CTAS syntax
    "execute_immediate", # Invalid CTAS syntax
    "data_character", # I think the test is bad
    "data_64_bit_numeric_warning", # Test does not explicitly set 64 bit columns
    "data_64_bit_lossless", # Test does not explicitly set 64 bit columns
    "data_date.*", # Date not a builtin function name
    "data_raw.*", # cast(1 bytea) is not valid `cannot cast type integer to bytea`
    "^data_time$", "^data_time_.*", # time objects not supported
    "^data_timestamp.*", # syntax not supported
    NULL
  ))
  DBItest::test_sql(c(
    "append_roundtrip_.*", # TODO
    "quote_string_na_is_null", # Invalid syntax
    "remove_table_missing_succeed",
    "roundtrip_character", # #10
    "roundtrip_character_native", # Possible false positive
    "roundtrip_factor", # #10
    "roundtrip_time", # TODO
    "roundtrip_timestamp", # We explicitly want to set tzone to UTC regardless of input
    "write_table_error", # TODO
    "quote_string_roundtrip",
    "quote_literal_roundtrip",
    "quote_literal_na_is_null",
    "quote_literal_na_is_null",
    "create_table_error",
    "roundtrip_64_bit_roundtrip",
    "write_table_row_names_default",
    "list_fields_wrong_table",
    "list_fields_quoted",
    "list_fields_object",
    "list_objects_features",
    NULL
  ))
  DBItest::test_meta(c(
    "column_info_consistent", # TODO
    "bind_empty",
    "rows_affected_query",
    "rows_affected_statement",
    "has_completed_statement",
    "get_statement_statement",
    "row_count_statement",
    NULL
  ))
  DBItest::test_transaction(c(
    NULL
  ))
  DBItest::test_compliance(c(
    "compliance", # We are defining additional subclasses for OdbcConnections
    "reexport",
    NULL
  ))
})

test_that("works with schemas (#197)", {
  con <- test_con("SQLSERVER")
  dbExecute(con, "DROP SCHEMA IF EXISTS testSchema")
  dbExecute(con, "CREATE SCHEMA testSchema")
  on.exit({
    dbExecute(con, "DROP TABLE testSchema.iris")
    dbExecute(con, "DROP SCHEMA testSchema")
  })

  ir <- iris
  ir$Species <- as.character(ir$Species)

  table_id <- Id(schema = "testSchema", table = "iris")
  dbWriteTable(conn = con, name = table_id, value = ir)
  dbWriteTable(conn = con, name = table_id, value = ir, append = TRUE)

  res <- dbReadTable(con, table_id)
  expect_equal(res, rbind(ir, ir))

  dbWriteTable(conn = con, name = table_id, value = ir, overwrite = TRUE)
  res <- dbReadTable(con, table_id)
  expect_equal(res, ir)

  # Test: We can enumerate schemas out of catalog ( #527 )
  # Part 1: Make sure we can see the schema we created in the
  # current catalog.
  res <- odbcConnectionSchemas(con)
  expect_true("testSchema" %in% res)
  # Part 2: Make sure we don't see that schema in the tempdb
  # listing ( out of catalog schema listing )
  res <- odbcConnectionSchemas(con, catalog_name = "tempdb")
  # Should, at least, have INFORMATION_SCHEMA and sys
  expect_true(length(res) > 1)
  expect_false("testSchema" %in% res)
})

test_that("works with dbAppendTable (#215)", {
  con <- test_con("SQLSERVER")

  ir <- iris
  ir$Species <- as.character(ir$Species)

  tbl <- local_table(con, "iris", ir)
  dbAppendTable(con, tbl, ir)

  res <- dbReadTable(con, tbl)
  expect_equal(res, rbind(ir, ir))
})

test_that("Subseconds are retained upon insertion (#208)", {
  con <- test_con("SQLSERVER")
  data <- data.frame(time = Sys.time())
  tbl <- local_table(con, "time", data)

  res <- dbReadTable(con, tbl)
  expect_equal(as.double(res$time), as.double(data$time))
})

test_that("dbWriteTable errors if field.types don't exist (#271)", {
  con <- test_con("SQLSERVER")

  expect_snapshot(
    sqlCreateTable(con, "foo", iris, field.types = list(bar = "[int]"))
  )
})

test_that("blobs can be retrieved out of order", {
  con <- test_con("SQLSERVER")
  values <- data.frame(
    c1 = 1,
    c2 = "this is varchar max",
    c3 = 11,
    c4 = "this is text",
    stringsAsFactors = FALSE
  )
  tbl <- local_table(con, "test_out_of_order_blob", values)

  received <- DBI::dbReadTable(con, tbl)
  expect_equal(received, values)

  # Also test retrival using a prepared statement
  received2 <- dbGetQuery(con,
    paste0("SELECT * FROM ", tbl, "  WHERE c1 = ?"),
    params = list(1L)
  )
  expect_equal(received2, values)
})

test_that("can bind NA values", {
  con <- test_con("SQLSERVER")
  # With SELECT ing with the OEM SQL Server driver, everything
  # after the first column should be unbound. Test null detection for
  # unbound columns (NULL is registered after a call to nanodbc::result::get)
  values <- data.frame(
    c1 = c("this is varchar max", NA_character_),
    c2 = c(1L, NA_integer_),
    c3 = c(1.0, NA_real_),
    c4 = c(TRUE, NA),
    c5 = c(Sys.Date(), NA),
    c6 = c(Sys.time(), NA),
    stringsAsFactors = FALSE
  )
  tbl <- local_table(con, "test_na", values)

  received <- DBI::dbReadTable(con, tbl)
  expect_equal(values[-6], received[-6])
  expect_equal(as.double(values[[6]]), as.double(received[[6]]))
})

test_that("can parse SQL server identifiers", {
  con <- test_con("SQLSERVER")
  input <- DBI::SQL(c(
    "testtable",
    "[testtable]",
    "[\"testtable\"]",
    "testta[ble",
    "testta]ble",
    "[testschema].[testtable]",
    "[testschema].testtable",
    "[testdb].[testschema].[testtable]",
    "[testdb].[testschema].testtable"
  ))
  expected <- c(
    DBI::Id("testtable"),
    DBI::Id("testtable"),
    DBI::Id("testtable"),
    DBI::Id("testta[ble"),
    DBI::Id("testta]ble"),
    DBI::Id("testschema", "testtable"),
    DBI::Id("testschema", "testtable"),
    DBI::Id("testdb", "testschema", "testtable"),
    DBI::Id("testdb", "testschema", "testtable")
  )
  expect_identical(DBI::dbUnquoteIdentifier(con, input), expected)
})

test_that("odbcPreviewObject doesn't warn about pending rows", {
  con <- test_con("SQLSERVER")
  tbl <- local_table(con, "test_preview", data.frame(a = 1:10L))

  # There should be no "Pending rows" warning
  expect_no_warning({
    res <- odbcPreviewObject(con, rowLimit = 3, table = tbl)
  })
  expect_equal(nrow(res), 3)
})

test_that("dates should always be interpreted in the system time zone (#398)", {
  con <- test_con("SQLSERVER")
  # TODO: resolve the issue requiring this skip
  skip_if(grepl("RStudio", dbGetInfo(con)$drivername), "Pro Drivers fail this test.")
  res <- dbGetQuery(con, "SELECT CAST(? AS date)", params = as.Date("2019-01-01"))
  expect_equal(res[[1]], as.Date("2019-01-01"))
})

test_that("UTF in VARCHAR is not truncated", {
  con <- test_con("SQLSERVER")
  value <- "grün"
  res <- dbGetQuery(
    con,
    paste0("SELECT '", value, "' AS colone")
  )
  expect_equal(value, res[[1]])
})

test_that("Zero-row-fetch does not move cursor", {
  con <- test_con("SQLSERVER")
  tbl <- local_table(con, "test_zero_row_fetch", mtcars[1:2, ])

  rs <- dbSendStatement(con, paste0("SELECT * FROM ", tbl))
  expect_equal(nrow(dbFetch(rs, n = 0)), 0)
  expect_equal(nrow(dbFetch(rs, n = 10)), 2)
  dbClearResult(rs)
})

test_that("isTempTable handles variety of temporary specifications", {
  con <- test_con("SQLSERVER")
  expect_true(isTempTable(con, "#myTmp"))
  expect_true(isTempTable(con, "#myTmp", catalog_name = "tempdb"))
  expect_true(isTempTable(con, "#myTmp", catalog_name = "%"))
  expect_true(isTempTable(con, "#myTmp", catalog_name = NULL))
  expect_true(!isTempTable(con, "##myTmp"))
  expect_true(!isTempTable(con, "#myTmp", catalog_name = "abc"))
})

test_that("dbExistsTable accounts for local temp tables", {
  con <- test_con("SQLSERVER")
  tbl_name <- "#myTemp"
  tbl_name2 <- "##myTemp"
  tbl_name3 <- "#myTemp2"
  DBI::dbExecute(con, paste0("CREATE TABLE ", tbl_name, " (
    id int not null,
    primary key (id) )"), immediate = TRUE)
  expect_true(dbExistsTable(con, tbl_name))
  expect_true(dbExistsTable(con, tbl_name, catalog_name = "tempdb"))
  # Fail because not recognized as temp table ( catalog not tempdb )
  expect_true(!dbExistsTable(con, tbl_name, catalog_name = "abc"))
  # Fail because not recognized as temp table ( second char "#" )
  expect_true(!dbExistsTable(con, tbl_name2, catalog_name = "tempdb"))
  # Fail because table not actually present
  expect_true(!dbExistsTable(con, tbl_name3, catalog_name = "tempdb"))

  # fail because table was created in another live session
  con2 <- test_con("SQLSERVER")
  expect_true(!dbExistsTable(con2, tbl_name))
})

test_that("can create / write to temp table", {
  con <- test_con("SQLSERVER")
  locTblName <- "#myloctmp"

  df <- data.frame(name = c("one", "two"), value = c(1, 2))
  values <- sqlData(con, row.names = FALSE, df[, , drop = FALSE])

  nm <- dbQuoteIdentifier(con, locTblName)
  fields <- createFields(con, values, row.names = FALSE, field.types = NULL)
  expected <- DBI::SQL(paste0(
    "CREATE TABLE ", nm, " (\n",
    "  ", paste(fields, collapse = ",\n  "), "\n)\n"
  ))

  ret1 <- sqlCreateTable(con, locTblName, values, temporary = TRUE)
  expect_equal(ret1, expected)

  ret2 <- sqlCreateTable(con, locTblName, values, temporary = FALSE)
  expect_equal(ret2, expected)

  dbWriteTable(con, locTblName, mtcars, row.names = TRUE)
  res <- dbGetQuery(con, paste0("SELECT * FROM ", locTblName))
  expect_equal(mtcars$mpg, res$mpg)
  dbAppendTable(con, locTblName, mtcars)
  res <- dbGetQuery(con, paste0("SELECT * FROM ", locTblName))
  expect_equal(nrow(res), 2 * nrow(mtcars))

  globTblName <- "##myglobtmp"
  expect_snapshot_warning(sqlCreateTable(con, globTblName, values, temporary = TRUE))
  expect_no_warning(sqlCreateTable(con, globTblName, values, temporary = FALSE))
  # ensure we include global temp tables in dbListTables output (#509)
  local_table(con, globTblName, mtcars)
  expect_true(globTblName %in% dbListTables(con))

  notTempTblName <- "nottemp"
  expect_snapshot_warning(sqlCreateTable(con, notTempTblName, values, temporary = TRUE))
  expect_no_warning(sqlCreateTable(con, notTempTblName, values, temporary = FALSE))
})

test_that("independent encoding of column entries and names (#834)", {
  skip_on_os("windows")
  # PRO Driver does not present "AutoTranslate" option.
  # Cursory investigation seems to indicate behavior is equivalent
  # to AutoTranslate being set to "yes" ( character data is returned
  # as UTF encoded, rather than in the code page corresponding to the
  # collation ).
  skip_if(Sys.getenv("ODBC_DRIVERS_VINTAGE") != "OEM")
  rawVal1 <- as.raw(c(0x72, 0xc3, 0xa6, 0x76, 0x65, 0x6e))
  rawVal2 <- as.raw(c(0xc3, 0xa5, 0x6c, 0x65, 0x6e, 0x73))
  rawVal3 <- as.raw(c(0xc3, 0xb8, 0x72, 0x72, 0x65, 0x64))
  rawCol <-  as.raw(c(0x62, 0xc3, 0xb8, 0x76, 0x73))
  df <- data.frame(
    var_char_col = c('kanin', rawToChar(rawVal1), rawToChar(rawVal2), rawToChar(rawVal3)),
    col = 1
  )
  colnames(df)[2] <- rawToChar(rawCol)
  conn <- test_con("SQLSERVER", encoding = "latin1", AutoTranslate = "no")
  on.exit({
    dbExecute(conn, "DROP TABLE deleteme_Danish_Norwegian_CI_AS.testschema.deleteme")
    dbExecute(conn, "DROP SCHEMA IF EXISTS testschema")
    dbExecute(conn, "USE tempdb")
    dbExecute(conn, "DROP DATABASE IF EXISTS deleteme_Danish_Norwegian_CI_AS")
  })
  DBI::dbExecute(conn, "CREATE DATABASE deleteme_Danish_Norwegian_CI_AS COLLATE Danish_Norwegian_CI_AS")
  DBI::dbExecute(conn, "USE deleteme_Danish_Norwegian_CI_AS")
  DBI::dbExecute(conn, "CREATE SCHEMA testschema")
  tbl_id <- DBI::Id(catalog = "deleteme_Danish_Norwegian_CI_AS", schema = "testschema", name = "deleteme")
  DBI::dbWriteTable(conn, name = tbl_id, df, field.types = c('var_char_col' = 'varchar(5)'), overwrite = TRUE)
  res <- DBI::dbReadTable(conn, tbl_id)
  expect_identical(df, res)
})

test_that("DATETIME2 precision (#790)", {
  con <- test_con("SQLSERVER")

  seed <- as.POSIXlt("2025-01-25 18:45:39.395682")
  val <- seed + runif(500, min = 0, max = 1)
  df <- data.frame(dtm = val, dtm2 = val)

  tbl <- local_table(con, "test_datetime2_precision", df,
    field.types = list("dtm" = "DATETIME", "dtm2" = "DATETIME2(6)"))
  res <- DBI::dbReadTable(con, tbl)
  expect_equal(as.POSIXlt(df[[2]])$sec, as.POSIXlt(res[[2]])$sec, tolerance = 1E-7)
})

test_that("DATETIMEOFFSET", {
  con <- test_con("SQLSERVER")

  # Test writing strings to a `DATETIMEOFFSET` target and reading it back as
  # POSIXct with properly recorded offset.
  df <- data.frame(tz_char = rep("2025-05-10 19:35:03.123 -02:00", 3), tz = rep("2025-05-10 19:35:03.123 -02:00", 3))

  tbl_name <- "test_datetimeoffset"
  tbl <- local_table(con, tbl_name, df,
    field.types = list("tz_char" = "VARCHAR(50)", "tz" = "DATETIMEOFFSET"), overwrite = TRUE)
  res <- DBI::dbReadTable(con, tbl)
  expect_s3_class(res[[2]], "POSIXct")
  expect_equal(as.double(res[[2]][1] - as.POSIXct(res[[1]][1]), units = "hours"), 2, tolerance = 1E-4)

  # Test writing POSIXct to a `DATETIMEOFFSET` target
  tbl_name <- "test_datetimeoffset2"
  df <- data.frame(tz = c(as.POSIXct("2022-04-01 12:00:00", tz = "Europe/Stockholm"), as.POSIXct("2022-04-01 13:00:00", tz = "Europe/Stockholm")))
  tbl <- local_table(con, tbl_name, df,
    field.types = list("tz" = "DATETIMEOFFSET"), overwrite = TRUE)
  # This result comes back as POSIXct in the timezone of the connection --- UTC in this case
  # However difftime can handle subtracting POSIXct with different timezones.
  res <- DBI::dbReadTable(con, tbl)
  expect_true(all(df$tz - res$tz == 0))
})

test_that("package:odbc roundtrip test", {
  test_roundtrip(test_con("SQLSERVER"))
})

test_that("Mixed success multiple result-sets (#924)", {
  con <- test_con("SQLSERVER")

  df <- data.frame(ID = LETTERS[1:10])

  tbl <- local_table(con, "test_mixed_success_param_retrieval", df)
  res <- DBI::dbGetQuery(
      con,
      "SELECT * FROM test_mixed_success_param_retrieval WHERE ID = ?",
      params = list(c("A", "B", "C"))
  )
  expect_equal(nrow(res), 3)
  res <- DBI::dbGetQuery(
      con,
      "SELECT * FROM test_mixed_success_param_retrieval WHERE ID = ?",
      params = list(c("Z", "A", "B", "C"))
  )
  expect_equal(nrow(res), 3)
  res <- DBI::dbGetQuery(
      con,
      "SELECT * FROM test_mixed_success_param_retrieval WHERE ID = ?",
      params = list(c("A", "Z", "B", "C"))
  )
  expect_equal(nrow(res), 3)
  res <- DBI::dbGetQuery(
      con,
      "SELECT * FROM test_mixed_success_param_retrieval WHERE ID = ?",
      params = list(c("A", "B", "C", "Z"))
  )
  expect_equal(nrow(res), 3)
})

test_that("Table-valued parameters", {
  # PRO Driver does not, at this time, handle TVPs.
  # In particular, `SQLDescribeParam`
  # applied to the TVP param does *not* return
  # SQL_SS_TABLE
  skip_if(Sys.getenv("ODBC_DRIVERS_VINTAGE") != "OEM")
  con <- test_con("SQLSERVER")
  dbExecute(con, "CREATE TYPE tvp_param AS TABLE (col0 INT, col1 BIGINT, col2 VARCHAR(MAX), col3 VARCHAR(MAX), col4 VARCHAR(MAX));")
  # tvp is second argument to sproc
  dbExecute(con, "CREATE PROCEDURE tvp_test(@p0 INT, @p1 tvp_param READONLY, @p2 NVARCHAR(MAX)) AS
                  BEGIN
                    SET NOCOUNT ON;
                    SELECT @p0 as p0, col0, col1, col2, col3, col4, @p2 as p2
                    FROM @p1
                    ORDER BY col0;
                    RETURN 0;
                  END")
  # a second sproc that takes the tvp as the first argument
  dbExecute(con, "CREATE PROCEDURE tvp_test2(@p0 tvp_param READONLY, @p1 INT, @p2 NVARCHAR(MAX)) AS
                  BEGIN
                    SET NOCOUNT ON;
                    SELECT @p1 as p0, col0, col1, col2, col3, col4, @p2 as p2
                    FROM @p0
                    ORDER BY col0;
                    RETURN 0;
                  END")
  on.exit({
    dbExecute(con, "DROP PROCEDURE tvp_test")
    dbExecute(con, "DROP PROCEDURE tvp_test2")
    dbExecute(con, "DROP TYPE tvp_param")
  })

  df.param <- data.frame(int = 1:10, bigint = 1:10, vrchr = rownames(mtcars)[1:10],
    vrchr2 = as.character(iris$Species[1:10]), vrchr3 = LETTERS[1:10], stringsAsFactors = FALSE)
  res <- dbGetQuery(con, "{ CALL tvp_test(?, ?, ?) }", params = list(100, df.param, "Lorem ipsum dolor sit amet"))

  expected <- cbind(data.frame("p0" = as.integer(100)), df.param,
    data.frame("p2" = "Lorem ipsum dolor sit amet", stringsAsFactors = FALSE))
  colnames(expected) <- c("p0", "col0", "col1", "col2", "col3", "col4", "p2")
  expected$col1 <- bit64::as.integer64(expected$col1)
  expect_identical(res, expected)

  res <- dbGetQuery(con, "{ CALL tvp_test2(?, ?, ?) }",
    params = list(df.param, 100, "Lorem ipsum dolor sit amet"))
  expect_identical(res, expected)
})

# #952: Binding date types that may have INTSXP OR REALSXP storage
test_that("Variable date type storage", {
  con <- test_con("SQLSERVER")
  tbl_name <- "datevariablestoragetest"
  # INTSXP storage
  data <- data.frame(
       "dates" = seq.Date(from = as.Date("2020-01-01"), to = as.Date("2020-01-02"), by = "1 day"),
       "dtms" = seq.POSIXt(from = as.POSIXct("2020-01-01", tz = "UTC"), to = as.POSIXct("2020-01-02", tz = "UTC"), length.out = 2L)
  )
  # REALSXP storage
  data_real <- data.frame(
       "dates" = c(as.Date("2020-01-01"), as.Date("2020-01-02")),
       "dtms" = c(as.POSIXct("2020-01-01", tz = "UTC"), as.POSIXct("2020-01-02", tz = "UTC"))
  )
  # 1. Writing both should succeed
  # 2. They should both come back with REALSXP storage
  local_table(con, tbl_name, data)
  res <- dbReadTable(con, tbl_name)
  expect_identical(data_real, res)

  DBI::dbWriteTable(con, tbl_name, data_real, overwrite = TRUE)
  res <- dbReadTable(con, tbl_name)
  expect_identical(data_real, res)
})

test_that("Recycling in dbBind works (#491)", {
  dat_in <- data.frame(id = 1000:1004, timestamp = sprintf("2022-04-01 12:00:%02d -04:00", 1:5))
  dat_expect <- subset(dat_in, id %in% 1003:1004) |>
    transform(timestamp = as.POSIXct(timestamp, tz = "America/New_York"))
  rownames(dat_expect) <- NULL

  con <- test_con("SQLSERVER")
  tbl_name <- local_table(con, "recyclingtemptable", dat_in,
                          field.types = c(id = "int", timestamp = "DATETIMEOFFSET"))

  res <- dbSendStatement(con, paste("select * from ", tbl_name, " where id = ? and timestamp > ? order by id"))
  expect_silent( dbBind(res, list(1003:1004, "2022-04-01 12:00:02.000000 +00:00")) )
  expect_silent( ret <- dbFetch(res) )
  dbClearResult(res)
  attr(ret$timestamp, "tzone") <- attr(dat_expect$timestamp, "tzone")
  expect_equal(ret, dat_expect)

  res <- dbSendStatement(con, paste("select * from ", tbl_name, " where id = ? and timestamp > ? order by id"))
  expect_error(
    # different lengths: 3 and 2
    dbBind(res, list(1003:1005, rep("2022-04-01 12:00:02.000000 +00:00", 2))),
    "Can't recycle .*"
  )
  dbClearResult(res)
})

#' Return the object hierarchy supported by a connection.
#'
#' Lists the object types and metadata known by the connection, and how those
#' object types relate to each other.
#'
#' The returned hierarchy takes the form of a nested list, in which each object
#' type supported by the connection is a named list with the following
#' attributes:
#'
#' \describe{
#'   \item{contains}{A list of other object types contained by the object, or
#'       "data" if the object contains data}
#'   \item{icon}{An optional path to an icon representing the type}
#' }
#
#' For instance, a connection in which the top-level object is a schema that
#' contains tables and views, the function will return a list like the
#' following:
#'
#' \preformatted{list(schema = list(contains = list(
#'                    list(name = "table", contains = "data")
#'                    list(name = "view", contains = "data"))))
#'
#' }
#' @param connection A connection object, as returned by `dbConnect()`.
#' @return The hierarchy of object types supported by the connection.
#' @export
# nocov start
odbcListObjectTypes <- function(connection) {
  UseMethod("odbcListObjectTypes")
}

#' @export
odbcListObjectTypes.default <- function(connection) {
  # slurp all the objects in the database so we can determine the correct
  # object hierarchy

  # all databases contain tables, at a minimum
  obj_types <- list(table = list(contains = "data"))

  # See if we have views too.  A little more elaborate than just
  # checking if any( table_types == "VIEW" ), because some back-ends
  # may contain VIEW "look-alikes".  For example PostgreSQL/"MATVIEW".
  # The icon url is that of the default Rstudio IDE resource location.
  # It gets automatically pulled in for an object type "view" ( but
  # not for some of the other ones ).  Should we not want to encode that
  # dependency, we can just bring that PNG into this package.
  table_types <- tolower(string_values(odbcConnectionTableTypes(connection)))
  viewlike <- grep("view", table_types, value = TRUE)
  viewlike_types <- sapply(
    viewlike,
    function(i) list(contains = "data", icon = "connections/objects/view.png"),
    simplify = FALSE
  )
  obj_types <- c(obj_types, viewlike_types)

  # check for schema support
  if (connection@info$supports.schema) {
    obj_types <- list(schema = list(contains = obj_types))
  }

  # check for multiple catalogs
  if (connection@info$supports.catalogs) {
    obj_types <- list(catalog = list(contains = obj_types))
  }

  obj_types
}

#' List objects in a connection.
#'
#' Lists all of the objects in the connection, or all the objects which have
#' specific attributes.
#'
#' When used without parameters, this function returns all of the objects known
#' by the connection. Any parameters passed will filter the list to only objects
#' which have the given attributes; for instance, passing \code{schema = "foo"}
#' will return only objects matching the schema \code{foo}.
#'
#' @param connection A connection object, as returned by `dbConnect()`.
#' @param ... Attributes to filter by.
#' @return A data frame with \code{name} and \code{type} columns, listing the
#'   objects.
#' @export
odbcListObjects <- function(connection, ...) {
  UseMethod("odbcListObjects")
}

#' @export
odbcListObjects.OdbcConnection <- function(connection,
                                           catalog = NULL,
                                           schema = NULL,
                                           name = NULL,
                                           type = NULL,
                                           ...) {
  check_string(catalog, allow_null = TRUE)
  check_string(schema, allow_null = TRUE)
  check_string(name, allow_null = TRUE)
  check_string(type, allow_null = TRUE)

  # if no catalog was supplied but this database has catalogs, return a list of
  # catalogs
  if (is.null(catalog)) {
    catalogs <- tryCatch(
      string_values(odbcConnectionCatalogs(connection)),
      error = function(err) character()
    )
    if (length(catalogs) > 0) {
      return(
        data.frame(
          name = catalogs,
          type = rep("catalog", times = length(catalogs)),
          stringsAsFactors = FALSE
        )
      )
    }
  }

  # if no schema was supplied but this database has schema, return a list of
  # schema
  if (is.null(schema)) {
    schemas <- string_values(odbcConnectionSchemas(connection, catalog))
    if (length(schemas) > 0) {
      return(
        data.frame(
          name = schemas,
          type = rep("schema", times = length(schemas)),
          stringsAsFactors = FALSE
        )
      )
    }
  }

  objs <- tryCatch(
    odbcConnectionTables(connection, name, catalog, schema, table_type = type),
    error = function(e) NULL
  )
  # just return a list of the objects and their types, possibly filtered by the
  # options above
  data.frame(
    name = objs[["table_name"]],
    type = tolower(objs[["table_type"]]),
    stringsAsFactors = FALSE
  )
}

#' List columns in an object.
#'
#' Lists the names and types of each column (field) of a specified object.
#'
#' The object to inspect must be specified as one of the arguments
#' (e.g. \code{table = "employees"}); depending on the driver and underlying
#' data store, additional specification arguments may be required.
#'
#' @param connection A connection object, as returned by `dbConnect()`.
#' @param ... Parameters specifying the object.
#' @return A data frame with \code{name} and \code{type} columns, listing the
#'   object's fields.
#' @export
odbcListColumns <- function(connection, ...) {
  UseMethod("odbcListColumns")
}

# given a connection, returns its "host name" (a unique string which identifies it)
computeHostName <- function(connection) {
  paste(
    collapse = "_",
    string_values(c(
      connection@info$username,
      connection@info$dbname,
      if (!identical(connection@info$servername, connection@info$dbname)) connection@info$servername
    ))
  )
}

computeDisplayName <- function(connection) {
  # use DSN if present
  dsn <- connection@info$sourcename
  if (!is.null(dsn) & dsn != "") {
    return(dsn)
  }

  # use the database name as the display name
  display_name <- connection@info$dbname
  server_name <- connection@info$servername
  user_name <- connection@info$username

  # prepend username if present
  server_name <- paste(collapse = "@", string_values(c(user_name, server_name)))

  # add server name (if it isn't already the display name, which can be the case
  # for serverless DBMS)
  if (!identical(server_name, display_name)) {
    display_name <- paste(collapse = " - ", string_values(c(display_name, server_name)))
  }

  display_name
}

validateObjectName <- function(table, view, ..., call = caller_env()) {

  # Handle view look-alike object types
  # ( e.g. PostgreSQL/"matview" )
  args <- list(...)
  arg_names <- names(args)
  viewlike <- grep("view", arg_names, value = TRUE)
  view <- Reduce(`%||%`, args[viewlike], view)

  # Error if both table and view are passed
  check_exclusive(table, view, .frame = call)

  table %||% view
}

#' @export
odbcListColumns.OdbcConnection <- function(connection,
                                           table = NULL,
                                           view = NULL,
                                           catalog = NULL,
                                           schema = NULL,
                                           ...) {
  check_string(table, allow_null = TRUE)
  check_string(view, allow_null = TRUE)
  check_string(catalog, allow_null = TRUE)
  check_string(schema, allow_null = TRUE)

  name <- validateObjectName(table, view, ...)
  # specify schema or catalog if given
  cols <- odbcConnectionColumns(connection,
    name = name,
    catalog_name = catalog,
    schema_name = schema
  )

  # extract and name fields for observer
  data.frame(
    name = cols[["name"]],
    type = cols[["field.type"]],
    stringsAsFactors = FALSE
  )
}

#' Preview the data in an object.
#'
#' Return the data inside an object as a data frame.
#'
#' The object to previewed must be specified as one of the arguments
#' (e.g. \code{table = "employees"}); depending on the driver and underlying
#' data store, additional specification arguments may be required.
#'
#' @param connection A connection object, as returned by `dbConnect()`.
#' @param rowLimit The maximum number of rows to display.
#' @param ... Parameters specifying the object.
#' @return A data frame containing the data in the object.
#' @export
odbcPreviewObject <- function(connection, rowLimit, ...) {
  UseMethod("odbcPreviewObject")
}

#' @export
odbcPreviewObject.OdbcConnection <- function(connection,
                                             rowLimit,
                                             table = NULL,
                                             view = NULL,
                                             schema = NULL,
                                             catalog = NULL,
                                             ...) {
  check_number_whole(rowLimit)
  check_string(table, allow_null = TRUE)
  check_string(view, allow_null = TRUE)
  check_string(schema, allow_null = TRUE)
  check_string(catalog, allow_null = TRUE)

  # extract object name from arguments
  name <- validateObjectName(table, view, ...)

  # prepend schema if specified
  if (!is.null(schema)) {
    name <- paste(dbQuoteIdentifier(connection, schema),
      dbQuoteIdentifier(connection, name),
      sep = "."
    )
  }

  # prepend catalog if specified
  if (!is.null(catalog)) {
    name <- paste(dbQuoteIdentifier(connection, catalog), name, sep = ".")
  }

  dbGetQuery(connection, odbcPreviewQuery(connection, rowLimit, name),
    n = rowLimit
  )
}

#' Create a preview query.
#'
#' Optimize against the rowLimit argument.  S3 since some
#' back-ends do not parse the LIMIT syntax.  Internal, not expected that
#' users would interact with this method.
#'
#' @param connection A connection object, as returned by `dbConnect()`.
#' @param rowLimit The maximum number of rows to display.
#' @param name Name of the object to be previewed
#' @keywords internal
odbcPreviewQuery <- function(connection, rowLimit, name) {
  UseMethod("odbcPreviewQuery")
}

#' Common top-N syntax ( MYSQL, PSQL, DB2, SNOWFLAKE, etc )
#' @rdname odbcPreviewQuery
odbcPreviewQuery.OdbcConnection <- function(connection, rowLimit, name) {
  paste0("SELECT * FROM ", name, " LIMIT ", rowLimit)
}

#' SQL Server specific top-N syntax
#' @rdname odbcPreviewQuery
`odbcPreviewQuery.Microsoft SQL Server` <- function(connection, rowLimit, name) {
  paste0("SELECT TOP ", rowLimit, " * FROM ", name)
}

#' Teradata specific top-N syntax
#' @rdname odbcPreviewQuery
`odbcPreviewQuery.Teradata` <- function(connection, rowLimit, name) {
  paste0("SELECT TOP ", rowLimit, " * FROM ", name)
}

#' Oracle specific top-N syntax
#' @rdname odbcPreviewQuery
odbcPreviewQuery.Oracle <- function(connection, rowLimit, name) {
  paste0("SELECT * FROM ", name, " WHERE ROWNUM <= ", rowLimit)
}

#' Get an icon representing a connection.
#'
#' @description
#' Return the path on disk to an icon representing a connection.
#'
#' @details
#' The icon returned should be a 32x32 square image file.
#'
#' @param connection A connection object, as returned by `dbConnect()`.
#' @return The path to an icon file on disk.
#' @keywords internal
odbcConnectionIcon <- function(connection) {
  # no icon is returned by default
  ""
}

#' List the actions supported for the connection
#'
#' @description
#' Return a list of actions that can be performed on the connection.
#'
#' @details
#' The list returned is a named list of actions, where each action has the
#' following properties:
#'
#' \describe{
#'   \item{callback}{A function to be invoked to perform the action}
#'   \item{icon}{An optional path to an icon representing the action}
#' }
#'
#' @param connection A connection object, as returned by `dbConnect()`.
#' @return A named list of actions that can be performed on the connection.
#' @keywords internal
odbcConnectionActions <- function(connection) {
  actions <- list()

  if (exists(".rs.api.documentNew")) {
    documentNew <- get(".rs.api.documentNew")
    actions <- c(
      actions,
      list(
        SQL = list(
          icon = system.file("icons/edit-sql.png", package = "odbc"),
          callback = function() {
            varname <- Filter(
              function(e) identical(get(e, envir = .GlobalEnv), connection),
              ls(envir = .GlobalEnv)
            )

            tables <- odbcConnectionTables(connection)
            columnPos <- 6
            if (nrow(tables) == 0) {
              contents <- paste(
                paste0("-- !preview conn=", ifelse(length(varname) > 0, varname[[1]], "")),
                "",
                "SELECT 1",
                "",
                sep = "\n"
              )
            } else {
              firstTable <- tables[1, ]

              tableName <- dbQuoteIdentifier(connection, firstTable$table_name)

              # add schema
              if (!is.na(firstTable$table_schema) && nchar(firstTable$table_schema) > 0) {
                tableName <- paste(dbQuoteIdentifier(connection, firstTable$table_schema), tableName, sep = ".")
              }

              # add catalog
              if (!is.na(firstTable$table_catalog) && nchar(firstTable$table_catalog) > 0) {
                tableName <- paste(dbQuoteIdentifier(connection, firstTable$table_catalog), tableName, sep = ".")
              }

              contents <- paste(
                paste0("-- !preview conn=", varname),
                "",
                paste0("SELECT * FROM ", tableName),
                "",
                sep = "\n"
              )

              columnPos <- 14
            }

            documentNew("sql", contents, row = 2, column = columnPos, execute = FALSE)
          }
        )
      )
    )
  }

  actions <- c(actions, list(
    Help = list(
      # show README for this package as the help; we will update to a more
      # helpful (and/or more driver-specific) website once one exists
      icon = "",
      callback = function() {
        utils::browseURL("https://github.com/rstats-db/odbc/blob/master/README.md")
      }
    )
  ))

  actions
}

on_connection_closed <- function(con) {
  # make sure we have an observer
  observer <- getOption("connectionObserver")
  if (is.null(observer)) {
    return(invisible(NULL))
  }

  type <- con@info$dbms.name
  host <- computeHostName(con)
  observer$connectionClosed(type, host)
}

on_connection_updated <- function(con, hint) {
  # make sure we have an observer
  observer <- getOption("connectionObserver")
  if (is.null(observer)) {
    return(invisible(NULL))
  }

  type <- con@info$dbms.name
  host <- computeHostName(con)
  observer$connectionUpdated(type, host, hint = hint)
}

on_connection_opened <- function(connection, code) {
  # make sure we have an observer
  observer <- getOption("connectionObserver")
  if (is.null(observer)) {
    return(invisible(NULL))
  }

  # find an icon for this DBMS
  icon <- odbcConnectionIcon(connection)

  # let observer know that connection has opened
  observer$connectionOpened(
    # connection type
    type = connection@info$dbms.name,

    # name displayed in connection pane
    displayName = computeDisplayName(connection),

    # host key
    host = computeHostName(connection),

    # icon for connection
    icon = icon,

    # connection code
    connectCode = code,

    # disconnection code
    disconnect = function() {
      dbDisconnect(connection)
    },
    listObjectTypes = function() {
      odbcListObjectTypes(connection)
    },

    # table enumeration code
    listObjects = function(...) {
      odbcListObjects(connection, ...)
    },

    # column enumeration code
    listColumns = function(...) {
      odbcListColumns(connection, ...)
    },

    # table preview code
    previewObject = function(rowLimit, ...) {
      odbcPreviewObject(connection, rowLimit, ...)
    },

    # other actions that can be executed on this connection
    actions = odbcConnectionActions(connection),

    # raw connection object
    connectionObject = connection
  )
}
# nocov end

skip_if_no_unixodbc()

test_that("MySQL", {
  DBItest::make_context(
    odbc(),
    test_connection_string("MYSQL"),
    tweaks = DBItest::tweaks(temporary_tables = FALSE),
    name = "MySQL"
  )

  DBItest::test_getting_started(c(
    "package_name", # Not an error
    NULL
  ))
  DBItest::test_driver("connect_format")
  DBItest::test_connection(c(
    NULL
  ))
  DBItest::test_result(c(
    "fetch_n_bad", # TODO
    "fetch_n_good_after_bad", # TODO
    "fetch_n_multi_row_inf", # TODO
    "fetch_no_return_value", # TODO
    "get_query_n_bad", # todo
    "get_query_good_after_bad_n", # todo
    "get_query_n_zero_rows", # todo
    "get_query_n_incomplete", # todo
    "get_query_n_multi_row_inf", # todo
    "fetch_no_return_value", # TODO
    "data_logical($|_.+)", # Not an error, PostgreSQL has a logical data type
    "data_raw.*", # cast(1 bytea) is not valid `cannot cast type integer to bytea`
    "^data_timestamp.*", # MySQL converts the timestamps from local times, so they roundtrip unexpectedly
    "data_character", # Strange MySQL Error only reproducible on travis
    "data_type_create_table", # 2023/6/23: Strange MySQL Error (warning/Result already cleared) only reproducible on GHA
    NULL
  ))
  DBItest::test_sql(c(
    "quote_identifier_vectorized", # Can't implement until https://github.com/rstats-db/DBI/issues/71 is closed
    "roundtrip_logical", # Not an error, PostgreSQL has a logical data type
    "roundtrip_timestamp", # We explicitly want to set tzone to UTC regardless of input
    "roundtrip_time",
    "roundtrip_raw", # TODO
    # TODO: fails with Pro Driver with shuffled names
    "list_fields",
    "list_tables",
    ".*_table_name",
    "write_table_error", # TODO
    "roundtrip_character", # Strange MySQL Error only reproducible on travis
    "roundtrip_character_native", # Strange MySQL Error only reproducible on travis
    "roundtrip_factor", # Strange MySQL Error only reproducible on travis
    "unquote_identifier_vectorized",
    "unquote_identifier_roundtrip",
    "unquote_identifier_special",
    "create_table_error",
    "append_roundtrip_.*",
    "roundtrip_64_bit_roundtrip",
    "write_table_row_names_default",
    "remove_table_temporary_arg",
    "remove_table_missing_succeed",
    "remove_table_temporary",
    "list_objects",
    "list_objects_features",
    "list_fields_wrong_table",
    "list_fields_quoted",
    "list_fields_object",
    "list_tables_quote",
    "list_objects_quote",
    NULL
  ))
  DBItest::test_meta(c(
    "rows_affected_query", # The MySQL Driver returns 1 affected row
    "rows_affected_statement",
    "row_count_statement",
    "column_info_consistent",
    "bind_.*",
    "has_completed_statement",
    "get_statement_statement",
    NULL
  ))
  DBItest::test_transaction(c(
    NULL
  ))
  DBItest::test_compliance(c(
    "compliance", # We are defining additional subclasses for OdbcConnections
    "reexport",
    NULL
  ))
})

test_that("can roundtrip columns", {
  test_roundtrip(test_con("MYSQL"), columns = c("logical", "binary"))
})

test_that("odbcPreviewObject", {
  con <- test_con("MYSQL")
  tbl <- local_table(con, "test_preview", data.frame(a = 1:10L))

  # There should be no "Pending rows" warning
  expect_no_warning({
    res <- odbcPreviewObject(con, rowLimit = 3, table = tbl)
  })
  expect_equal(nrow(res), 3)
})

test_that("sproc result retrieval", {
  con <- test_con("MYSQL")

  # TODO: resolve the issue requiring this skip
  skip_if(grepl("libmysqlodbc_sb64", dbGetInfo(con)$drivername), "Pro Drivers fail this test.")

  sprocName <- "testSproc"
  DBI::dbExecute(
    con,
    paste0("CREATE PROCEDURE ", sprocName, "(IN arg INT) BEGIN SELECT 'abc' as TestCol; END")
  )
  on.exit(DBI::dbExecute(con, paste0("DROP PROCEDURE ", sprocName)))

  expect_no_error({
    res <- dbGetQuery(con, paste0("CALL ", sprocName, "(1)"))
  })
  expect_identical(
    res,
    data.frame("TestCol" = "abc", stringsAsFactors = FALSE)
  )
})

skip_if_no_unixodbc()

test_that("PostgreSQL", {
  DBItest::make_context(
    odbc(),
    test_connection_string("POSTGRES"),
    tweaks = DBItest::tweaks(temporary_tables = FALSE, placeholder_pattern = "?"),
    name = "PostgreSQL"
  )

  DBItest::test_getting_started(c(
    "package_name", # Not an error
    NULL
  ))
  DBItest::test_driver("connect_format")
  DBItest::test_connection(c(
    "clear_result_return_statement",
    "cannot_clear_result_twice_statement",
    NULL
  ))
  DBItest::test_result(c(
    "fetch_n_bad", # TODO
    "fetch_n_good_after_bad", # TODO
    "fetch_no_return_value", # TODO
    "get_query_n_bad", # todo
    "get_query_good_after_bad_n", # todo
    "get_query_n_zero_rows", # todo
    "fetch_no_return_value", # TODO
    "data_raw.*", # cast(1 bytea) is not valid `cannot cast type integer to bytea`
    "^data_time$", "^data_time_.*", # `time()` function is not valid syntax
    "^data_timestamp.*", # We explicitly want to set tzone to UTC
    "data_64_bit_numeric_warning", # TODO
    "data_64_bit_lossless", # TODO
    "data_integer", # Fails, unreliably on Windows
    # TODO: fails with Pro Driver with "a statement is already in progress"
    "send_query_immediate",
    # TODO: fails with Pro Driver with "a statement is already in progress"
    "send_query_only_one_result_set",
    "send_query_syntax_error", # TODO
    "get_query_syntax_error", # TODO
    "send_query_params", # TODO
    "fetch_n_multi_row_inf", # TODO
    "get_query_n_multi_row_inf", # TODO
    "get_query_n_incomplete", # TODO
    "get_query_params", # TODO
    "send_statement_params", # TODO
    "execute_params", # TODO
    NULL
  ))
  DBItest::test_sql(c(
    "quote_identifier_vectorized", # Can't implement until https://github.com/rstats-db/DBI/issues/71 is closed
    "quote_identifier_special", # TODO
    "roundtrip_timestamp", # We explicitly want to set tzone to UTC
    "roundtrip_time",
    "roundtrip_raw", # TODO
    "list_tables",
    ".*_table_name",
    "write_table_error", # TODO
    "unquote_identifier_vectorized", # TODO
    "create_table_overwrite", # TODO
    "create_table_error", # TODO
    "append_table_.*", # TODO
    "append_roundtrip_.*", # TODO
    "append_table_.*", # TODO
    "roundtrip_64_bit_roundtrip", # TODO
    "roundtrip_character", # TODO
    "roundtrip_field_types", # TODO
    "write_table_append_incompatible",
    "write_table_row_names_default", # TODO
    "remove_table_temporary_arg", # TODO
    "remove_table_missing_succeed", # TODO
    "remove_table_temporary", # TODO
    "list_objects_features", # TODO
    "list_fields_wrong_table", # TODO
    "list_fields_quoted", # TODO
    "list_fields_object", # TODO
    NULL
  ))
  DBItest::test_meta(c(
    "arrow_.*",
    "stream_bind.*",
    "bind_.*", # TODO
    "has_completed_statement",
    "get_statement_statement",
    "column_info_consistent", # TODO
    "row_count_statement", # TODO
    "rows_affected_statement", # TODO
    "rows_affected_query", # TODO
    "get_info_result", # TODO
    NULL
  ))
  DBItest::test_transaction(c(
    NULL
  ))
  DBItest::test_compliance(c(
    "compliance", # We are defining additional subclasses for OdbcConnections
    "reexport", # TODO
    NULL
  ))
})

test_that("can roundtrip columns", {
  con <- test_con("POSTGRES")
  test_roundtrip(con)
})

test_that("show method works as expected with real connection", {
  skip_on_os("windows")
  con <- test_con("POSTGRES")

  expect_output(show(con), "@localhost")
  expect_output(show(con), "Database: [a-z]+")
  expect_output(show(con), "PostgreSQL Version: ")
})

test_that("64 bit integers work with alternate mappings", {
  con_integer64 <- test_con("POSTGRES", bigint = "integer64")
  con_integer <- test_con("POSTGRES", bigint = "integer")
  con_numeric <- test_con("POSTGRES", bigint = "numeric")
  con_character <- test_con("POSTGRES", bigint = "character")

  tbl <- local_table(
    con_integer64,
    "test",
    data.frame(a = 1:10L),
    field.types = c(a = "BIGINT")
  )

  expect_s3_class(dbReadTable(con_integer64, tbl)$a, "integer64")
  expect_type(dbReadTable(con_integer, tbl)$a, "integer")
  expect_type(dbReadTable(con_numeric, tbl)$a, "double")
  expect_type(dbReadTable(con_character, tbl)$a, "character")
})

# This test checks whether when writing to a table and using
# result_describe_parameters to offer descriptions of the data
# we are attempting to write, our logic remains robust to the
# case when the data being written has columns ordered
# differently than the table we are targetting.
test_that("Writing data.frame with column ordering different than target table", {
  con <- test_con("POSTGRES")
  values <- data.frame(
    datetime = as.POSIXct(c(14, 15), origin = "2016-01-01", tz = "UTC"),
    name = c("one", "two"),
    num = 1:2,
    stringsAsFactors = FALSE
  )
  tbl <- "test_order_write"
  dbCreateTable(con, tbl, values)
  dbAppendTable(con, tbl, values[c(2, 3, 1)])
  on.exit(dbRemoveTable(con, tbl))

  received <- dbReadTable(con, tbl)
  received <- received[order(received$num), ]
  row.names(received) <- NULL
  expect_equal(values, received)
})

test_that("odbcPreviewObject", {
  con <- test_con("POSTGRES")
  tbl <- local_table(con, "test_preview", data.frame(a = 1:10L))

  # There should be no "Pending rows" warning
  expect_no_warning({
    res <- odbcPreviewObject(con, rowLimit = 3, table = tbl)
  })
  expect_equal(nrow(res), 3)
})

skip_if_no_unixodbc()

test_that("SQLite", {
  DBItest::make_context(
    odbc(),
    test_connection_string("SQLITE"),
    tweaks = DBItest::tweaks(placeholder_pattern = "?", strict_identifier = TRUE),
    name = "SQLite"
  )

  DBItest::test_getting_started(c(
    "package_name", # Not an error
    NULL
  ))
  DBItest::test_driver(c(
    "connect_format",
    "connect_bigint_numeric",
    "connect_bigint_character",
    "connect_bigint_integer64",
    NULL
  ))
  DBItest::test_connection(c(
    "data_type_connection",
    NULL
  ))
  DBItest::test_result(c(
    "data_logical$", # Not an error
    "data_64_bit.*", # TODO
    "data_integer", # These tests are returned as strings by SQLite (bug?)
    "data_raw.*", # cast(1 bytea) is not valid `cannot cast type integer to bytea`
    "^data_time$", "^data_time_.*", # time objects not supported
    "^data_timestamp.*", # SQLite doesn't do timestamps
    "^data_date.*", # SQLite doesn't do dates
    "send_query_params", # TODO
    "fetch_n_bad", # TODO
    "fetch_n_good_after_bad", # TODO
    "fetch_no_return_value", # TODO
    "fetch_n_multi_row_inf", # TODO
    "get_query_n_bad", # TODO
    "get_query_good_after_bad_n", # TODO
    "get_query_n_multi_row_inf", # TODO
    "get_query_n_zero_rows", # TODO
    "get_query_n_incomplete", # TODO
    "get_query_params", # TODO
    "send_statement_params", # TODO
    "execute_params", # TODO
    "data_numeric", # TODO
    "clear_result_return_statement",
    "cannot_clear_result_twice_statement",
    NULL
  ))
  DBItest::test_sql(c(
    "quote_identifier_special", # #7
    "roundtrip_timestamp.*", # SQLite doesn't do timestamps
    "roundtrip_date.*", # SQLite doesn't do timestamps
    "roundtrip_logical", # Not an error
    "read_table", # #7

    "exists_table_temporary",
    "list_tables_temporary",
    "list_objects_temporary",
    "list_fields_temporary",

    # These work locally but fail on travis due to an old SQLite version
    "roundtrip_integer",
    "roundtrip_numeric.*",
    "roundtrip_character",
    "roundtrip_factor",
    "roundtrip_raw",
    "quote_identifier_vectorized", # TODO
    "quote_identifier_string", # TODO
    "unquote_identifier_vectorized", # TODO
    "read_table_empty", # TODO
    "read_table_row_names_na_missing", # TODO
    "create_table_overwrite", # TODO
    "create_table_error", # TODO
    "create_table_visible_in_other_connection", # TODO
    "append_table_missing", # TODO
    "append_table_append_incompatible", # TODO
    "append_roundtrip_.*", # TODO
    "append_table_name", # TODO
    "append_table_row_names_false", # TODO
    "write_table_error", # TODO
    "overwrite_table", # TODO
    "overwrite_table_missing", # TODO
    "append_table", # TODO
    "append_table_new", # TODO
    "temporary_table", # TODO
    "table_visible_in_other_connection", # TODO
    "roundtrip_64_bit_roundtrip", # TODO
    "roundtrip_time", # TODO
    "roundtrip_field_types", # TODO
    "write_table_row_names_.*", # TODO
    "list_tables", # TODO
    "exists_table", # TODO
    "remove_table_temporary_arg", # TODO
    "remove_table_missing_succeed", # TODO
    "remove_table_temporary", # TODO
    "list_objects", # TODO
    "list_objects_features", # TODO
    "list_fields", # TODO
    "list_fields_wrong_table", # TODO
    "list_fields_quoted", # TODO
    "list_fields_object", # TODO
    "exists_table_name",
    "read_table_name",
    "write_table_name",
    "remove_table_name",
    "write_table_append_incompatible",
    NULL
  ))
  DBItest::test_meta(c(
    "arrow_.*",
    "stream_bind.*",
    "column_info_consistent", # TODO
    "row_count_statement", # TODO
    "rows_affected_statement", # TODO
    "rows_affected_query", # TODO
    "has_completed_statement",
    "get_statement_statement",
    "bind_.*", # TODO
    NULL
  ))
  DBItest::test_transaction(c(
    "begin_write_disconnect",
    NULL
  ))
  DBItest::test_compliance(c(
    "reexport", # TODO
    NULL
  ))
})

test_that("unsupported types gives informative error", {
  con <- test_con("SQLITE")

  df <- data.frame(foo = complex(1))
  expect_snapshot(
    error = TRUE,
    dbWriteTable(con, "df", df)
  )
})

test_that("odbcPreviewObject works", {
  con <- test_con("SQLITE")
  tbl <- local_table(con, "test_preview", data.frame(a = 1:10L))

  # There should be no "Pending rows" warning
  expect_no_warning({
    res <- odbcPreviewObject(con, rowLimit = 3, table = tbl)
  })
  expect_equal(nrow(res), 3)
})

test_that("dbWriteTable() with `field.types` with `append = TRUE`", {
  con <- test_con("SQLITE")

  expect_snapshot(
    error = TRUE,
    dbWriteTable(
      con,
      "boopery",
      data.frame(bop = 1),
      field.types = c(bop = "numeric"),
      append = TRUE
    )
  )
})

test_that("WriteTable() with `overwrite = TRUE` and `append = TRUE`", {
  con <- test_con("SQLITE")

  expect_snapshot(
    error = TRUE,
    dbWriteTable(
      con,
      "boopery",
      data.frame(bop = 1),
      overwrite = TRUE,
      append = TRUE
    )
  )
})

test_that("dbWriteTable(), existing table, `overwrite = FALSE`, `append = FALSE`", {
  con <- test_con("SQLITE")

  local_table(con, "boopery", data.frame(bop = 1))

  expect_snapshot(
    error = TRUE,
    dbWriteTable(
      con,
      "boopery",
      data.frame(bop = 1),
      overwrite = FALSE,
      append = FALSE
    )
  )
})

# nocov start
.onUnload <- function(libpath) {
  gc() # Force garbage collection of connections
  library.dynam.unload("odbc", libpath)
}

.onLoad <- function(libname, pkgname) {
  # If TZDIR is not set we need to set it to R's timezone database on windows,
  # we can use the standard timezone database elsewhere.
  if (is_windows()) {
    if (!nzchar(Sys.getenv("TZDIR"))) Sys.setenv("TZDIR" = file.path(R.home(), "share", "zoneinfo"))
  } else {
    set_odbcsysini()
  }
}

# nocov end

#' @include dbi-connection.R
NULL



class_cache <- new.env(parent = emptyenv())

OdbcConnection <- function(
    ...,
    timezone = "UTC",
    timezone_out = "UTC",
    encoding = "",
    name_encoding = "",
    bigint = c("integer64", "integer", "numeric", "character"),
    timeout = Inf,
    dbms.name = NULL,
    attributes = NULL,
    interruptible = getOption("odbc.interruptible", interactive()),
    .connection_string = NULL,
    call = caller_env(2)
) {
  check_attributes(attributes, call = call)

  args <- compact(list(...))
  check_args(args)
  connection_string <- build_connection_string(args, .connection_string)

  bigint <- bigint_mappings()[match.arg(bigint, names(bigint_mappings()))]

  if (is.infinite(timeout) || is.na(timeout)) {
    timeout <- 0
  }

  withCallingHandlers(
    ptr <- odbc_connect(
      connection_string,
      timezone = timezone,
      timezone_out = timezone_out,
      encoding = encoding,
      name_encoding = name_encoding,
      bigint = bigint,
      timeout = timeout,
      r_attributes = attributes,
      interruptible_execution = interruptible
    ),
    error = function(cnd) {
      check_quoting(args)
    }
  )
  quote <- connection_quote(ptr)

  info <- connection_info(ptr)
  if (!is.null(dbms.name)) {
    info$dbms.name <- dbms.name
  }
  if (!nzchar(info$dbms.name)) {
    cli::cli_abort(
      c(
        "!" = "The ODBC driver failed to generate a DBMS name.",
        "i" = "Please provide one manually with {.arg dbms.name}."
      ),
      call = call
    )
  }

  class <- getClassDef(info$dbms.name, inherits = FALSE)
  if (is.null(class)) {
    class <- getClassDef(info$dbms.name, where = class_cache, inherits = FALSE)
    if (is.null(class) || methods::isVirtualClass(class)) {
      setClass(info$dbms.name, contains = "OdbcConnection", where = class_cache)
    }
  }
  new(
    info$dbms.name,
    ptr = ptr,
    quote = quote,
    info = info,
    encoding = encoding,
    connectionString = decompose_connection_string(connection_string)
  )
}

#' Quote special character when connecting
#'
#' @description
#' When connecting to a database using odbc, all the arguments are concatenated
#' into a single connection string that looks like `name1=value1;name2=value2`.
#' That means if your value contains `=` or `;` then it needs to be quoted.
#' Other rules mean that you need to quote any values that starts or ends with
#' white space, or contains `{` or `}`.
#'
#' This function quotes a string in a way that should work for most drivers,
#' but unfortunately there doesn't seem to be an approach that works everywhere.
#' If this function doesn't work for you, you'll need to carefully read the
#' docs for your driver.
#'
#' @export
#' @param x A string to quote.
#' @return A quoted string, wrapped in `I()`.
#' @examples
#' quote_value("abc")
#' quote_value("ab'c")
#'
#' # Real usage is more likely to look like:
#' \dontrun{
#' library(DBI)
#'
#' con <- dbConnect(
#'   odbc::odbc(),
#'   dsn = "reallycooldatabase"
#'   password = odbc::quote_value(Sys.getenv("MY_PASSWORD"))
#' )
#' }
quote_value <- function(x) {
  has_single <- grepl("'", x, fixed = TRUE)
  has_double <- grepl('"', x, fixed = TRUE)

  if (has_single && has_double) {
    abort("Don't know how to escape a value with both single and double quotes.")
  } else if (has_double) {
    quote <- "'"
  } else {
    quote <- '"'
  }

  I(paste0(quote, x, quote))
}

check_args <- function(args) {
  if (length(args) == 0) {
    return()
  }

  if (!all(has_names(args))) {
    abort("All elements of ... must be named.", call = quote(DBI::dbConnect()))
  }

  name_groups <- split(names(args), tolower(names(args)))
  bad_names <- name_groups[lengths(name_groups) > 1]
  if (length(bad_names) > 0) {
    bullets <- vapply(bad_names, paste0, collapse = ", ", FUN.VALUE = character(1))

    abort(
      c(
        "After ignoring case, some arguments have the same name:",
        set_names(bullets, "*")
      ),
      call = quote(DBI::dbConnect())
    )
  }
}

check_quoting <- function(args) {
  needs_quoting <- vapply(args, needs_quoting, FUN.VALUE = logical(1))
  if (any(needs_quoting)) {
    # TODO: use cli pluralisation
    args <- paste0("`", names(args)[needs_quoting], "`", collapse = ", ")

    inform(c(
      paste0(args, " contains a special character that may need quoting."),
      i = "Wrap the value in `odbc::quote_value()` to use a heuristic that should work for most drivers.",
      i = "If that still doesn't work, consult your driver's documentation.",
      i = "Otherwise, you can suppress this message by wrapping the value in `I()`."
    ))
  }
}

needs_quoting <- function(x) {
  if (inherits(x, "AsIs")) {
    return(FALSE)
  }

  if (grepl("^\\{.*\\}$", x) || grepl("^(['\\\"]).*\\1$", x)) {
    return(FALSE)
  }

  grepl("[{}=;]| $|^ ", x)
}

# -------------------------------------------------------------------------

#' Supported Connection Attributes
#'
#' These (pre) connection attributes are supported and can be passed as
#' part of the `dbConnect` call in the named list `attributes` parameter:
#'
#' * `azure_token`: This should be a string scalar; in particular Azure Active
#'   Directory authentication token.  Only for use with Microsoft SQL Server and
#'   with limited support away from the OEM Microsoft driver.
#' * `sf_private_key`: This parameter is specific to establishing a connection
#'   to `snowflake` and is understood by both OEM, as well as `Posit`
#'   pro drivers.  Argument should be a string (scalar); in particular a
#'   PEM-encoded private key.  Note,
#'   if using private key authentication, the `authenticator` connection
#'   string attribute must be set to `SNOWFLAKE_JWT`.
#'   Using this *connection* attribute is an alternative to using the
#'   `PRIV_KEY_FILE` connection string attribute.
#' * `sf_private_key_password`: If key passed using `sf_private_key` is
#'   encrypted, you can use this attribute to communicate the password.
#' @rdname ConnectionAttributes
#' @keywords internal
#' @aliases ConnectionAttributes
#' @usage NULL
#' @format NULL
#' @examples
#' \dontrun{
#' conn <- dbConnect(
#'   odbc::odbc(),
#'   dsn = "my_azure_mssql_db",
#'   Encrypt = "yes",
#'   attributes = list("azure_token" = .token)
#'
#' conn <- dbConnect(
#'  odbc::odbc(),
#'  dsn = "snowflake",
#'  attributes = list("sf_private_key" = paste(readLines("<path-to-private-key-file>"), collapse="\n"),
#'                    "sf_private_key_password" = "<optional-private-key-encryption-password>"),
#'  authenticator = "SNOWFLAKE_JWT")
#' }
SUPPORTED_CONNECTION_ATTRIBUTES <- c("azure_token", "sf_private_key", "sf_private_key_password")

#' Odbc Connection Methods
#'
#' Implementations of pure virtual functions defined in the `DBI` package
#' for OdbcConnection objects.
#' @name OdbcConnection
NULL

#' odbcConnectionColumns
#'
#' @description
#'
#' For a given table this function returns detailed information on
#' all fields / columns.  The expectation is that this is a relatively thin
#' wrapper around the ODBC `SQLColumns` function call, with some of the field
#' names renamed / re-ordered according to the return specifications below.
#'
#' @details
#' In [DBI::dbWriteTable()] we make a call to this method
#' to get details on the fields of the table we are writing to.  In particular
#' the columns `data_type`, `column_size`, and `decimal_digits` are used.  An
#' implementation is not necessary for [DBI::dbWriteTable()] to work.
#'
#' `odbcConnectionColumns` is routed through the `SQLColumns` ODBC
#' method.  This function, together with remaining catalog functions
#' (`SQLTables`, etc), by default ( `SQL_ATTR_METADATA_ID == false` ) expect
#' either ordinary arguments (OA) in the case of the catalog, or pattern value
#' arguments (PV) in the case of schema/table name.  For these, quoted
#' identifiers do not make sense, so we unquote identifiers prior to the call.
#'
#' @param conn OdbcConnection
#' @param name Table identifiers.
#'   By default, are interpreted as a ODBC search pattern where `_` and `%` are
#'   wild cards. Set `exact = TRUE` to match `_` exactly.
#' @param ... additional parameters to methods
#' @param exact Set to `TRUE` to escape `_` in identifier names so that it
#'   matches exactly, rather than matching any single character. `%` always
#'   matches any number of characters as this is unlikely to appear in a
#'   table name.
#'
#' @seealso The ODBC documentation on
#' [SQLColumns](https://learn.microsoft.com/en-us/sql/odbc/reference/syntax/sqlcolumns-function)
#' for further details.
#'
#' @return data.frame with columns
#' - name
#' - field.type - equivalent to type_name in SQLColumns output
#' - table_name
#' - schema_name
#' - catalog_name
#' - data_type
#' - column_size
#' - buffer_length
#' - decimal_digits
#' - numeric_precision_radix
#' - remarks
#' - column_default
#' - sql_data_type
#' - sql_datetime_subtype
#' - char_octet_length
#' - ordinal_position
#' - nullable
#'
#' @rdname odbcConnectionColumns
#' @keywords internal
setGeneric(
  "odbcConnectionColumns",
  valueClass = "data.frame",
  function(conn, name, ..., exact = FALSE) {
    standardGeneric("odbcConnectionColumns")
  }
)

setMethod("odbcConnectionColumns", c("OdbcConnection", "Id"),
  function(conn, name, ..., column_name = NULL, exact = FALSE) {
    odbcConnectionColumns(conn,
      name = id_field(name, "table"),
      catalog_name = id_field(name, "catalog"),
      schema_name = id_field(name, "schema"),
      column_name = column_name,
      exact = exact
    )
  }
)

setMethod("odbcConnectionColumns", c("OdbcConnection", "character"),
  function(conn,
           name,
           ...,
           catalog_name = NULL,
           schema_name = NULL,
           column_name = NULL,
           exact = FALSE) {
    if (exact) {
      schema_name <- escapePattern(schema_name)
      name <- escapePattern(name)
      column_name <- escapePattern(column_name)
    }
    connection_sql_columns(
      conn@ptr,
      table_name = name,
      catalog_name = catalog_name,
      schema_name = schema_name,
      column_name = column_name
    )
  }
)

setMethod("odbcConnectionColumns", c("OdbcConnection", "SQL"),
  function(conn, name, ..., exact = FALSE) {
    odbcConnectionColumns(conn, dbUnquoteIdentifier(conn, name)[[1]], ..., exact = exact)
  }
)

# This function returns a listing of tables accessible
# to the connected user.
# The expectation is that this is a relatively thin
# wrapper around the ODBC `SQLTables` function call,
# albeit returning a subset of the fields.
#
# It is important to note that, similar to the ODBC/API
# call, this method also accomodates pattern-value arguments
# for the catalog, schema, and table name arguments.
#
# If extending this method, be aware that `package:odbc`
# internally uses this method to satisfy both
# DBI::dbListTables and DBI::dbExistsTable methods.
# ( The former also advertises pattern value arguments )
setGeneric(
  "odbcConnectionTables",
  valueClass = "data.frame",
  function(conn, name, ...) {
    standardGeneric("odbcConnectionTables")
  }
)

setMethod("odbcConnectionTables", c("OdbcConnection", "Id"),
  function(conn, name, table_type = NULL, exact = FALSE) {
    odbcConnectionTables(
      conn,
      name = id_field(name, "table"),
      catalog_name = id_field(name, "catalog"),
      schema_name = id_field(name, "schema"),
      table_type = table_type,
      exact = exact
    )
  }
)

setMethod("odbcConnectionTables", c("OdbcConnection", "character"),
  function(conn,
           name,
           catalog_name = NULL,
           schema_name = NULL,
           table_type = NULL,
           exact = FALSE) {
    if (exact) {
      catalog_name <- escapePattern(catalog_name)
      schema_name <- escapePattern(schema_name)
      name <- escapePattern(name)
    }
    connection_sql_tables(
      conn@ptr,
      catalog_name = catalog_name,
      schema_name = schema_name,
      table_name = name,
      table_type = table_type
    )
  }
)

setMethod("odbcConnectionTables", "OdbcConnection",
  function(conn,
           name = NULL,
           catalog_name = NULL,
           schema_name = NULL,
           table_type = NULL,
           exact = FALSE) {
    odbcConnectionTables(
      conn,
      name = "%",
      catalog_name = catalog_name,
      schema_name = schema_name,
      table_type = table_type,
      exact = exact
    )
  }
)

setMethod("odbcConnectionTables", c("OdbcConnection", "SQL"),
  function(conn, name, table_type = NULL, exact = FALSE) {
    odbcConnectionTables(
      conn,
      dbUnquoteIdentifier(conn, name)[[1]],
      table_type = table_type,
      exact = exact
    )
  }
)

setGeneric(
  "odbcConnectionCatalogs",
  valueClass = "character",
  function(conn) {
    standardGeneric("odbcConnectionCatalogs")
  }
)

setMethod("odbcConnectionCatalogs", "OdbcConnection",
  function(conn) {
    connection_sql_catalogs(conn@ptr)
  }
)

# Currently, for a generic connection the
# catalog_name argument is ignored.
setGeneric(
  "odbcConnectionSchemas",
  valueClass = "character",
  function(conn, catalog_name = NULL) {
    standardGeneric("odbcConnectionSchemas")
  }
)

setMethod("odbcConnectionSchemas", "OdbcConnection",
  function(conn, catalog_name = NULL) {
    connection_sql_schemas(conn@ptr)
  }
)

setGeneric(
  "odbcConnectionTableTypes",
  valueClass = "character",
  function(conn) {
    standardGeneric("odbcConnectionTableTypes")
  }
)

setMethod("odbcConnectionTableTypes", "OdbcConnection",
  function(conn) {
    connection_sql_table_types(conn@ptr)
  }
)


# Internal method, used to check for target existence
# in dbWriteTable.  Historically we used DBI::dbExistsTable
# however its mission may be interpreted more broadly
# ( does this table exists period ) versus the question
# we are trying to answer in dbWriteTable ( does this
# table identifier I may be writing to exist? )  In
# particular, the former may mean, across catalogs or schema;
# the latter we understand to mean:
# 1. At the specified catalog/schema; if not specified then
# 2A. In the current catalog/schema; OR
# 2B. In a temporary table.
#
# While by default we continue to route requests to dbExistsTable,
# this narrower interpretation may allow some back-ends
# to optimize this search.
setGeneric(
  "dbExistsTableForWrite",
  valueClass = "logical",
  function(conn, name, ...) {
    standardGeneric("dbExistsTableForWrite")
  }
)

setMethod("dbExistsTableForWrite", c("OdbcConnection", "Id"),
  function(conn, name, ...) {
    dbExistsTableForWrite(
      conn,
      name = id_field(name, "table"),
      catalog_name = id_field(name, "catalog"),
      schema_name = id_field(name, "schema")
    )
  }
)

setMethod("dbExistsTableForWrite", c("OdbcConnection", "SQL"),
  function(conn, name, ...) {
    dbExistsTableForWrite(conn, dbUnquoteIdentifier(conn, name)[[1]], ...)
  }
)

setMethod("dbExistsTableForWrite", c("OdbcConnection", "character"),
  function(conn, name, ...) {
    dbExistsTable(conn = conn, name = name, ...)
  }
)

test_that("errors if unnamed arguments", {
  expect_snapshot(check_args(list(1, 2, 3)), error = TRUE)
})

test_that("errors about case-insensitve arguments", {
  expect_no_error(check_args(list()))
  expect_no_error(check_args(list(x = 1, y = 1)))

  expect_snapshot(error = TRUE, {
    check_args(list(xa = 1, xA = 1))
    check_args(list(xa = 1, xA = 1, XA = 1))
    check_args(list(xa = 1, xA = 1, xb = 1, xB = 1))
  })
})

test_that("messages if values might need quoting", {
  expect_snapshot(check_quoting(list(foo = "f{oo", bar = "b{ar", baz = "baz")))
})

test_that("correctly detects values that need escaping", {
  # I() opts-out
  expect_false(needs_quoting(I("f{o")))

  # already quoted
  expect_false(needs_quoting("'f{o'"))
  expect_false(needs_quoting('"f{o"'))
  expect_false(needs_quoting("{f{o}"))

  # no special values
  expect_false(needs_quoting("foo"))

  # has a special character
  expect_true(needs_quoting("foo{"))
  expect_true(needs_quoting("foo}"))
  expect_true(needs_quoting("foo="))
  expect_true(needs_quoting("foo;"))

  # space at start or end
  expect_true(needs_quoting("foo "))
  expect_true(needs_quoting(" foo"))
  expect_false(needs_quoting("fo o"))
})

test_that("automatically picks correct quote type", {
  expect_equal(quote_value("x"), I('"x"'))
  expect_equal(quote_value("x'"), I('"x\'"'))
  expect_equal(quote_value("x'"), I('"x\'"'))

  expect_snapshot(quote_value("'\""), error = TRUE)
})

# connections pane -------------------------------------------------------------
test_that("validateObjectName() errors informatively", {
  skip_if_no_unixodbc()
  con <- test_con("SQLITE")

  expect_snapshot(
    error = TRUE,
    odbcListColumns(con, table = "boop", view = "bop")
  )

  expect_snapshot(
    error = TRUE,
    odbcListColumns(con)
  )
})

#' @export
#' @rdname DBI-classes
setClass("Spark SQL", contains = "OdbcConnection")

# Databricks supports multiple catalogs.  On the other hand,
# the default implementation of `odbcConnectionSchemas` which routes through
# `SQLTables` is likely to enumerate the schemas in the currently active
# catalog only.
#
# This implementation will respect the `catalog_name` arrgument.
setMethod("odbcConnectionSchemas", "Spark SQL",
  function(conn, catalog_name = NULL) {
    if (is.null(catalog_name)) {
      return(callNextMethod())
    }
    res <- dbGetQuery(conn, paste0("SHOW SCHEMAS IN ", catalog_name))
    if (nrow(res)) {
      return(res$databaseName)
    }
    return(character())
  }
)


#' @export
#' @rdname odbcDataType
#' @usage NULL
setMethod("odbcDataType", "Spark SQL",
  function(con, obj, ...) {
    switch_type(
      obj,
      factor = "VARCHAR(255)",
      datetime = "TIMESTAMP",
      date = "DATE",
      binary = "BINARY",
      integer = "INT",
      int64 = "INT",
      double = "DOUBLE",
      character = "VARCHAR(255)",
      logical = "BOOLEAN",
      list = "VARCHAR(255)",
      time = ,
      stop("Unsupported type", call. = FALSE)
    )
  }
)

#' @include dbi-connection.R
NULL

#' @export
#' @rdname DBI-classes
setClass("Vertica Database", contains = "OdbcConnection")

#' @export
#' @rdname odbcDataType
#' @usage NULL
setMethod("odbcDataType", "Vertica Database",
  function(con, obj, ...) {
    switch_type(
      obj,
      factor = "VARCHAR",
      datetime = "TIMESTAMP",
      date = "DATE",
      integer = "INTEGER",
      int64 = "INTEGER",
      double = "DOUBLE PRECISION",
      character = "VARCHAR",
      logical = "BOOLEAN",
      list = "VARCHAR",
      time = "TIME",
      binary = "VARBINARY",
      stop("Unsupported type", call. = FALSE)
    )
  }
)

has_names <- function(x) {
  nms <- names(x)
  if (is.null(nms)) {
    rep(FALSE, length(x))
  } else {
    !(is.na(nms) | nms == "")
  }
}

`%||%` <- function(x, y) if (is.null(x)) y else x

string_values <- function(x) {
  # TODO: Throw a condition object that can be caught for debugging purposes
  x <- tryCatch(x, error = function(x) "")
  unique(x[nzchar(x)])
}

# version of iconv that respects input Encoding, which bare iconv does not.
enc2iconv <- function(x, to, ...) {
  encodings <- Encoding(x)
  for (enc in unique(encodings)) {
    if (enc == to) {
      next
    }
    current <- enc == encodings
    if (enc == "unknown") {
      enc <- ""
    }
    x[current] <- iconv(x[current], from = enc, to = to, ...)
  }
  x
}

choices_rd <- function(x) {
  paste0(collapse = ", ", paste0("'", x, "'"))
}

lengths <- function(x) {
  vapply(x, length, integer(1))
}

# A 'size' must be an integer greater than 1, returned as a double so we have a larger range
parse_size <- function(x, arg = caller_arg(x), call = caller_env()) {
  check_number_whole(x, min = 1, arg = arg, call = call)
  as.numeric(x)
}

id_field <- function(id,
                     field = c("catalog", "schema", "table"),
                     error_call = caller_env()) {
  arg_match(field, error_call = error_call)

  if (length(id@name) == 1) {
    switch(field,
      catalog = NULL,
      schema = NULL,
      table = id@name[[1]],
    )
  } else if (length(id@name) == 2) {
    switch(field,
      catalog = NULL,
      schema = id@name[[1]],
      table = id@name[[2]],
    )
  } else if (length(id@name) == 3) {
    switch(field,
      catalog = id@name[[1]],
      schema = id@name[[2]],
      table = id@name[[3]],
    )
  } else {
    abort("Identifier must be length 1, 2, or 3.", call = error_call)
  }
}

isPatternValue <- function(val) {
  grepl("[%|_]", val)
}

# Per ODBC spec, "%" is synonymous with
# zero or more of any characters. "_" is
# any one character.  Here we convert
# a string with pattern value style wildcards
# to one that can be used in a regex search.
convertWildCards <- function(val) {
  val <- gsub("%", "(.*)", val)
  gsub("_", "(.)", val)
}

# Will return " AND key [= | LIKE] value"
# where "=" or "LIKE" is chosen depending
# on the the `exact` argument, and whether
# the `value` contains any wild card characters.
#
# Note, wild cards are not escaped in `value`
# as part of this call.
getSelector <- function(key, value, exact) {
  if (is.null(value)) {
    return("")
  }
  comp <- " = "
  if ((value == "%" || !exact) &&
    isPatternValue(value)) {
    comp <- " LIKE "
  }
  value <- paste0("'", value, "'")

  paste0(" AND ", key, comp, value)
}

# Will iterate over charsToEsc argument and for each:
# will escape any un-escaped occurance in `x`.
escapePattern <- function(x, charsToEsc = c("_"), escChar = "\\\\") {
  if (is.null(x) || inherits(x, "AsIs")) {
    return(x)
  }
  matchGroup <- paste(charsToEsc, collapse = "|")
  pattern <- paste0("([^", escChar, "])(", matchGroup, ")")
  replace <- paste0("\\1", escChar, "\\2")
  x <- gsub(pattern, replace, x)
  I(x)
}

is_windows <- function() {
  identical(.Platform$OS.type, "windows")
}

is_macos <- function() {
  identical(Sys.info()[["sysname"]], "Darwin")
}

compact <- function(x) x[!vapply(x, is.null, logical(1))]

set_odbcsysini <- function() {
  odbcsysini <- Sys.getenv("ODBCSYSINI")
  if (!identical(odbcsysini, "")) {
    return(invisible())
  }

  tryCatch(
    {
      path <- dirname(odbcListConfig()[["drivers"]])
      Sys.setenv(ODBCSYSINI = path)
    },
    error = function(err) NULL
  )

  invisible()
}

random_name <- function(prefix = "") {
  vals <- c(letters, LETTERS, 0:9)
  name <- paste0(sample(vals, 10, replace = TRUE), collapse = "")
  paste0(prefix, "odbc_", name)
}

# error handling ---------------------------------------------------------------
# `call` looks to the "user-called" function rather than the usual `caller_env(n)`.
# this function is called from C++, and parent.frame() (and thus caller_env())
# is unable to traverse across the Rcpp code to find the caller env; setting
# `call = caller_env(n)` with n of 0/1/2/3 gives no error context (#789).
rethrow_database_error <- function(msg, call = trace_back()$call[[1]]) {
  tryCatch(
    res <- parse_database_error(msg),
    error = function(e) cli::cli_abort("{msg}", call = call)
  )

  cli::cli_abort(
    c(
      "!" = "ODBC failed with error {res$cnd_context_code} from \\
             {.field {paste0(res$cnd_context_driver, collapse = '')}}.",
      set_database_error_names(res$cnd_body),
      "i" = "From {.file {res$cnd_context_nanodbc}}."
    ),
    class = "odbc_database_error",
    call = call
  )
}

parse_database_error <- function(msg) {
  # Split nanodbc's context, error code, and message returned from the database
  cnd_msg <- strsplit(msg, "\n", fixed = TRUE)[[1]]

  # Parse out nanodbc/nanodbc.cpp:<line#> and 5-char error code
  cnd_context <- strsplit(cnd_msg[1], ": ")[[1]]
  cnd_context_nanodbc <- cnd_context[1]
  cnd_context_code <- cnd_context[2]

  # In error returned from the database, find the square-bracketed context
  # on driver and driver manager
  cnd_context_driver <- gregexpr("(\\[.*?\\])", cnd_msg[-1])
  cnd_context_driver <- regmatches(cnd_msg[-1], cnd_context_driver)[[1]]

  # Remove the square-bracketed context from the database error
  cnd_body <- Reduce(
    function(p, x) gsub(p, "", x, fixed = TRUE),
    cnd_context_driver,
    cnd_msg[-1],
    right = TRUE
  )
  # Limit error message length to something sane (#914)
  max_len <- 5000
  cnd_body <- sapply(cnd_body, FUN = function(x) {
    if (nchar(x) > max_len) {
      x <- substr(x, 0, max_len)
      x <- paste0(x, "...")
    }
    x
  }, USE.NAMES = FALSE)

  cnd_body <- contextualize_database_error(cnd_body)

  list(
    cnd_context_nanodbc = cnd_context_nanodbc,
    cnd_context_code = cnd_context_code,
    cnd_context_driver = cnd_context_driver,
    cnd_body = cnd_body
  )
}

contextualize_database_error <- function(cnd_body) {
  if (length(cnd_body) == 0 || !is_character(cnd_body)) {
    return(cnd_body)
  }

  if (any(grepl("Data source name not found", cnd_body))) {
    cnd_body <-
      c(
        cnd_body,
        "i" = "See {.help odbc::odbcListDataSources} to learn more."
      )
  }

  cnd_body
}

set_database_error_names <- function(cnd_body) {
  # Respect names from `contextualize_database_error()`, otherwise ensure
  # that the first is an "x" and the remainder are bulleted.
  if (length(cnd_body) == 0) {
    return(cnd_body)
  }

  cnd_body <- escape_curly_brackets(cnd_body)

  if (is.null(names(cnd_body))) {
    return(
      set_names(cnd_body, nm = c("x", rep("*", length(cnd_body) - 1)))
    )
  }

  set_names(
    cnd_body,
    c(
      ifelse(names(cnd_body[1]) == "", "x", names(cnd_body[1])),
      ifelse(names(cnd_body[-1]) == "", "*", names(cnd_body[-1]))
    )
  )
}

# Escape curly brackets before formatting with cli (#859). Do so only to
# unnamed elements so that formatting is preserved for context
# added with `contextualize_database_error()`.
escape_curly_brackets <- function(cnd_body) {
  if (is.null(names(cnd_body))) {
    return(gsub("\\{", "{{", gsub("\\}", "}}", cnd_body)))
  }

  unnamed <- names(cnd_body) == ""
  cnd_body[unnamed] <- gsub("\\{", "{{", gsub("\\}", "}}", cnd_body[unnamed]))
  cnd_body
}

# check helpers for common odbc arguments --------------------------------------
check_row.names <- function(row.names, call = caller_env()) {
  if (is.null(row.names) || is_scalar_logical(row.names) || is_string(row.names)) {
    return()
  }

  cli::cli_abort(
    "{.arg row.names} must be {.code NULL}, {.code TRUE}, {.code FALSE}, \\
     {.code NA}, or a single string, not {.obj_type_friendly {row.names}}.",
    call = call
  )
}

check_field.types <- function(field.types, call = caller_env()) {
  if (!(is_null(field.types) || (is_named(field.types)))) {
    cli::cli_abort(
      "{.arg field.types} must be {.code NULL} or a named vector of field \\
       types, not {.obj_type_friendly {field.types}}.",
      call = call
    )
  }
}

check_attributes <- function(attributes, call = caller_env()) {
  if (!all(has_names(attributes))) {
    cli::cli_abort("All elements of {.arg attributes} must be named.", call = call)
  }

  attributes_supported <- names(attributes) %in% SUPPORTED_CONNECTION_ATTRIBUTES
  if (!all(attributes_supported)) {
    cli::cli_abort(
      c("!" = "{.arg attributes} does not support the connection attribute{?s} \\
               {.val {names(attributes)[!attributes_supported]}}.",
        "i" = "Allowed connection attribute{?s} {?is/are} \\
               {.val {SUPPORTED_CONNECTION_ATTRIBUTES}}."),
      call = call
    )
  }
}

# apple + spark drive config (#651) --------------------------------------------
# Method will attempt to:
# 1. Locate an installation of unixodbc / error out otherwise.
# 2. Verify the driver_config argument.  Expect this to be a list with
#    two fields:
#    * path Vector of viable driver paths ( only first one is used )
#    * url A location where the user can downlaod the driver from.
#    See spark_simba_config, for example.  Its return value is used as
#    the value for this argument.
# 3. Inspect the config for some settings that can impact how our package
#    performs.
# 4. If action == "modify" then we attempt to modify the config in-situ.
# 5. Otherwise we throw a warning asking the user to revise.
configure_simba <- function(driver_config,
                            action = "modify", call = caller_env()) {
  if (!is_macos()) {
    return(invisible())
  }
  if (!is.null(getOption("odbc.no_config_override"))) {
    return(invisible())
  }

  unixodbc_install <- locate_install_unixodbc()
  if (length(unixodbc_install) == 0) {
    error_install_unixodbc(call)
  }

  simba_config <- driver_config$path
  if (length(simba_config) == 0) {
    func <- cli::cli_warn
    if (action == "modify") {
      func <- cli::cli_abort
    }
    func(
      c(i = "Please install the needed driver from {driver_config$url}"),
      call = call
    )
  }
  configure_unixodbc_simba(unixodbc_install, simba_config[1], action, call)
}

locate_install_unixodbc <- function() {
  libodbcinst <- c(
    system_safely("odbc_config --lib-prefix"),
    system_safely("pkg-config --variable=libdir odbc")
  )

  if (length(libodbcinst) > 0) {
    return(libodbcinst[1])
  }

  common_dirs <- c(
    "/usr/lib",
    "/usr/local",
    "/lib",
    "/usr/local/lib",
    "/usr/lib/x86_64-linux-gnu",
    "/usr/lib/aarch64-linux-gnu",
    "/opt/homebrew/lib",
    "/opt/homebrew/etc",
    "/opt/homebrew/opt/unixodbc/lib",
    "/opt/R/arm64/lib",
    "/opt/R/x86_64/lib"
  )

  return(libodbcinst_file(common_dirs))
}

system_safely <- function(x) {
  tryCatch(
    {
      unixodbc_prefix <- system(x, intern = TRUE, ignore.stderr = TRUE)
      return(libodbcinst_file(unixodbc_prefix))
    },
    error = function(e) {},
    warning = function(w) {}
  )
  character()
}

# Search for an instance of the unixodbc lib in
# the directories passed in the `dirs` argument.
# Preference is given to shared libraries.
libodbcinst_file <- function(dirs) {
  if (is_windows()) {
    return(character())
  }
  # Order matters, we prefer shlib (#919)
  file_names <- ifelse(is_macos(), "libodbcinst.dylib", "libodbcinst.so")
  file_names <- c(file_names, "libodbcinst.a")
  for (file_name in file_names) {
    candidates <- list.files(dirs,
      pattern = file_name, full.names = TRUE)
    if (length(candidates)) {
      return(candidates[1])
    }
  }
  return(character())
}

error_install_unixodbc <- function(call) {
  abort(
    c(
      "Unable to locate the unixODBC driver manager.",
      i = if (is_macos()) "Please install unixODBC using Homebrew with `brew install unixodbc`.",
      i = if (!is_macos()) "Please install unixODBC and retry."
    ),
    call = call
  )
}

configure_unixodbc_simba <- function(unixodbc_install, simba_config, action, call) {

  # As shipped, the simba spark ini has an incomplete final line
  suppressWarnings(
    simba_lines <- readLines(simba_config)
  )
  res <- replace_or_append(
    lines = simba_lines,
    key_pattern = "^ODBCInstLib=",
    accepted_value = unixodbc_install,
    replacement = paste0("ODBCInstLib=", unixodbc_install)
  )
  warnings <- character()
  if (action != "modify" && res$modified) {
     warnings <- c(warnings, c("*" = "Please consider revising the {.arg ODBCInstLib}
       field in {.file {simba_config}} and setting its value to {.val {unixodbc_install}}"))
  }
  simba_lines_new <- res$new_lines
  res <- replace_or_append(
    lines = simba_lines_new,
    key_pattern = "^DriverManagerEncoding=",
    accepted_value = "UTF-16|utf-16",
    replacement = "DriverManagerEncoding=UTF-16"
  )
  if (action != "modify" && res$modified) {
     warnings <- c(warnings, c("*" = "Please consider revising the
       {.arg DriverManagerEncoding} field in {.file {simba_config}} and setting its
       value to {.val UTF-16}."))
  }
  if (length(warnings)) {
    cli::cli_warn(c(
      c(i = "Detected potentially unsafe driver settings:"),
      warnings
    ))
  }
  simba_lines_new <- res$new_lines

  if (action == "modify") {
    write_simba_lines(simba_lines, simba_lines_new, simba_config, call)
  }

  invisible()
}

write_simba_lines <- function(spark_lines, spark_lines_new, spark_config, call) {
  if (identical(spark_lines, spark_lines_new)) {
    return(invisible())
  }

  # see assertthat::is.writeable
  if (!is_writeable(spark_config)) {
    abort(
      c(
        paste0(
          "Detected needed changes to the driver configuration file at ",
          spark_config,
          ", but the file was not writeable."
        ),
        i = "Please make the changes outlined at https://solutions.posit.co/connections/db/databases/databricks/#troubleshooting-apple-macos-users."
      ),
      call = call
    )
  }

  writeLines(spark_lines_new, spark_config)
}

# Interpret the argument as an `ODBC` driver
# and attempt to infer the directory housing it.
# It will return an empty character vector if unable to.
driver_dir <- function(driver) {
  # driver argument could be an outright path, or a name
  # of a driver specified in odbcinst.ini  Try to discern
  driver_spec <- subset(odbcListDrivers(), name == driver)
  if (nrow(driver_spec)) {
    driver_path <- subset(driver_spec, attribute == "Driver")$value
  } else {
    driver_path <- driver
  }

  ret <- dirname(driver_path)
  if (ret == ".") {
    ret <- character()
  }
  return(ret)
}

is_writeable <- function(path) {
  tryCatch(file.access(path, mode = 2)[[1]] == 0, error = function(e) FALSE)
}

# Given a vector of lines in an ini file, look for a given key pattern.
# If found:
# - No action if the `accepted_value` argument is found on line.
# - Replace otherwise.
# If not found: append.
# The `replacement` is the whole intended line, giving the desired
# "key=value" pair.
# @return a list with two elements:
# - new_lines: Potentially modified lines
# - modified: Whether method modified lines, where modified means
# both changed or appended.
replace_or_append <- function(lines, key_pattern, accepted_value, replacement) {
  matching_lines_loc <- grepl(key_pattern, lines)
  matching_lines <- lines[matching_lines_loc]
  found_ok <- length(matching_lines) != 0 &&
    any(grepl(accepted_value, lines[matching_lines_loc]))
  if (length(matching_lines) == 0) {
    lines <- c(lines, replacement)
  } else if (!found_ok) {
    lines[matching_lines_loc] <- replacement
  }
  return(list(new_lines = lines, modified = !found_ok))
}

running_on_connect <- function() {
  Sys.getenv("RSTUDIO_PRODUCT") == "CONNECT"
}

# Like Sys.getenv(), but checks passed keys in order and allows for non-string
# defaults.
getenv2 <- function(..., .unset = NULL) {
  x <- as.character(list(...))
  env <- Sys.getenv(x, NA_character_)
  if (all(is.na(env))) {
    return(.unset)
  }
  val <- env[!is.na(env)][1]
  unname(val)
}

find_default_driver <- function(paths, fallbacks, label, call = caller_env()) {
  paths <- paths[file.exists(paths)]
  if (length(paths) > 0) {
    return(paths[1])
  }
  fallbacks <- intersect(fallbacks, odbcListDrivers()$name)
  if (length(fallbacks) > 0) {
    return(fallbacks[1])
  }
  cli::cli_abort(
    c(
      "Failed to automatically find the {label} ODBC driver.",
      i = "Set {.arg driver} to known driver name or path."
    ),
    call = call
  )
}

#' Finalize a connection string by appending a list of arguments to an optional
#' starting connection string.
#'
#' @description
#' Arguments are expected in the formed of a named list (k1 = v1, k2 = v2),
#' and are appended as key-value pairs:
#' starting_string;k1=v1;k2=v2
#'
#' @param args A named list of arguments to append to connection string
#' @param string An initial / starting connection string
#' @return string A finalized connection string
#' @noRd
build_connection_string <- function(args = list(), string = NULL) {

  args_string <- paste(names(args), args, sep = "=", collapse = ";")

  if (!is.null(string) && !grepl(";$", string) && length(args) > 0) {
    string <- paste0(string, ";")
  }

  paste0(string, args_string)
}

#' Decompose a connection string into a named list.
#'
#' @description
#' This method essentially reverses `build_connection_string`,
#' with the addition that we attempt to scrub out any
#' sensitive key/value pairs when we return to the caller.
#'
#' @param string A connection string.  Expected format is
#'               `k1=v1;k2=v2;k3=v3`
#' @return list A named list (k1 = v1, k2 = v2, ...)
#' @noRd
decompose_connection_string <- function(string) {

  res_pairs <- strsplit(string, ";", fixed = TRUE)[[1]]
  res_key_value <- strsplit(res_pairs, "=", fixed = TRUE)
  res_key_value <- Filter(function(x) length(x) == 2 && nchar(x[1]),
                          res_key_value)
  values <- lapply(res_key_value, FUN = function(x) trimws(x[2]))
  # Remove enclosing quotation marks around value
  # we may have put them there using odbc::quote_value
  values <- lapply(values, FUN = function(x) {
      x <-gsub("(^')(.*)('$)", "\\2", x)
      gsub("(^\")(.*)(\"$)", "\\2", x)
  })
  nms <- sapply(res_key_value, FUN = function(x) x[1])
  names(values) <- trimws(nms)
  return(sanitize_connection_string(values))
}

#' Sanitize a named list representing a connection string
#' (see `decompose_connection_string`) by scrubbing out
#' values belonging to keys we identify as potentially
#' carrying authentication information.
#'
#' @description
#' Scrubs out pre-set vector of key patterns.
#'
#' @param lst A named list representing contents of
#'            connection string.
#' @return A named list (k1 = v1, k2 = v2, ...), much like
#'         input argument, with some values potentially filtered out.
#' @noRd
sanitize_connection_string <- function(lst) {
  EXCLUDE_PATTERNS <- c("*PWD*", "*PASS*", "*TOKEN*")

  regex <- paste(EXCLUDE_PATTERNS, tolower(EXCLUDE_PATTERNS), collapse = "|", sep = "|")
  return(lst[!grepl(regex, names(lst), fixed = FALSE)])
}

test_that("parse_size works", {
  expect_snapshot(error = TRUE, parse_size("foo"))

  expect_identical(parse_size(1L), 1)
  expect_identical(parse_size(1), 1)
  expect_identical(parse_size(.Machine$integer.max), as.numeric(.Machine$integer.max))
  expect_identical(parse_size(.Machine$integer.max + 1), .Machine$integer.max + 1)
})

test_that("id_field extracts correct elements", {
  expect_equal(id_field(DBI::Id("z"), "table"), "z")

  expect_equal(id_field(DBI::Id("y", "z"), "schema"), "y")
  expect_equal(id_field(DBI::Id("y", "z"), "table"), "z")

  expect_equal(id_field(DBI::Id("x", "y", "z"), "catalog"), "x")
  expect_equal(id_field(DBI::Id("x", "y", "z"), "schema"), "y")
  expect_equal(id_field(DBI::Id("x", "y", "z"), "table"), "z")
})

test_that("id_field checks inputs", {
  expect_snapshot(error = TRUE, {
    id_field(DBI::Id("a"), "foo")
    id_field(DBI::Id("a", "b", "c", "d"))
  })
})

test_that("getSelector", {
  # If no wild cards are detected always use exact comparison / ignore `exact` argument
  expect_equal(getSelector("mykey", "myvalue", exact = TRUE), " AND mykey = 'myvalue'")
  expect_equal(getSelector("mykey", "myvalue", exact = FALSE), " AND mykey = 'myvalue'")

  # If `value` contains wild cards, respect `exact`argument
  expect_equal(getSelector("mykey", "myvalu_", exact = TRUE), " AND mykey = 'myvalu_'")
  expect_equal(getSelector("mykey", "myvalu_", exact = FALSE), " AND mykey LIKE 'myvalu_'")

  expect_equal(getSelector("mykey", "myvalu%", exact = TRUE), " AND mykey = 'myvalu%'")
  expect_equal(getSelector("mykey", "myvalu%", exact = FALSE), " AND mykey LIKE 'myvalu%'")

  # ... unless argument is '%' - always use 'LIKE' since this is most likely the
  # desired comparison / ignore `exact` argument
  expect_equal(getSelector("mykey", "%", exact = TRUE), " AND mykey LIKE '%'")
  expect_equal(getSelector("mykey", "%", exact = FALSE), " AND mykey LIKE '%'")
})

test_that("errors are rethrown informatively (#643, #788)", {
  skip_if_not(has_unixodbc())

  expect_snapshot(error = TRUE, dbConnect(odbc(), dsn = "does_not_exist_db"))

  con <- test_con("SQLITE")
  expect_snapshot(error = TRUE, dbExecute(con, "SELECT * FROM boopbopbopbeep"))
})

test_that("rethrow_database_error() errors well when parse_database_error() fails", {
  expect_snapshot(
    error = TRUE,
    rethrow_database_error("boop", call = NULL)
  )
})

test_that("parse_database_error() works with messages from the wild", {
  msg <- "nanodbc/nanodbc.cpp:1135: 00000
          [unixODBC][Driver Manager]Data source name not found and no default driver specified"
  expect_snapshot(error = TRUE, rethrow_database_error(msg, call = NULL))

  msg <- "nanodbc/nanodbc.cpp:1594: 07002
          [Microsoft][ODBC Driver 17 for SQL Server]COUNT field incorrect or syntax error"
  expect_snapshot(error = TRUE, rethrow_database_error(msg, call = NULL))

  msg <- "nanodbc/nanodbc.cpp:1710: 07002
          [ODBC Firebird Driver]COUNT field incorrect"
  expect_snapshot(error = TRUE, rethrow_database_error(msg, call = NULL))

  msg <- "nanodbc/nanodbc.cpp:1655: HYT00
          [Microsoft][SQL Server Native Client 11.0]Query timeout expired"
  expect_snapshot(error = TRUE, rethrow_database_error(msg, call = NULL))

  msg <- "nanodbc/nanodbc.cpp:1147: 00000
          [Microsoft][ODBC Driver 18 for SQL Server]Login timeout expired
          [Microsoft][ODBC Driver 18 for SQL Server]TCP Provider: Error code 0x2726
          [Microsoft][ODBC Driver 18 for SQL Server]A network-related or instance-specific error has occurred while establishing a connection to 127.0.0.1. Server is not found or not accessible. Check if instance name is correct and if SQL Server is configured to allow remote connections. For more information see SQL Server Books Online. "
  expect_snapshot(error = TRUE, rethrow_database_error(msg, call = NULL))

  # message contains brackets that may be interpreted as inline markup (#859)
  msg <- "nanodbc/nanodbc.cpp:1722: S1000
          [SAP AG][LIBODBCHDB DLL][HDBODBC] General error;403 internal error: Error opening the cursor for the remote database <***.***> [SAP AG][LIBODBCHDB SO][HDBODBC] Connection not open;-10807 Connection down: [89013] Socket closed by peer {***.**.*.**:***** -> ***.**.***.**:***** TenantName:(none) SiteVolumeID:1:3 SiteType:PRIMARY ConnectionID:****** SessionID:************}
          <SQL> 'SELECT DISTINCT \"po_id\", ***CENSORED***'"
  expect_snapshot(error = TRUE, rethrow_database_error(msg, call = NULL))

  skip_on_cran()
  
  # Really long error message can cause rethrow_database_error to take a really long time.
  # This is due to cli::cli_abort, current version is 3.6.5 at the time of this writing. (#914)
  err_msg_w_really_long_query <- paste(
    c(
      "nanodbc/nanodbc.cpp:1544",
      "[Snowflake] SQL compilation error: error line 1 at position 92",
      paste0(
        "<SQL> 'SELECT STATECODE FROM AIR_QUALITY_DATA_UNITED_STATES.PUBLIC.AIR_QUALITY WHERE COUNTYCODE IN (",
        paste(rep('1', 35000), collapse = ", "),
        "'"
      )
    ),
    collapse = "\n"
  ) # Even if this really long error message gets passed to cli::cli_abort it should only take 10 - 20 sec to fail
  expect_no_condition(parse_database_error(err_msg_w_really_long_query))
  timing <- system.time({
    expect_error(rethrow_database_error(err_msg_w_really_long_query, call = NULL))
  })
  expect_lt(
    object = timing["elapsed"],
    expected = 5,
    label = "Runtime of `rethrow_database_error(err_msg_w_really_long_query)`",
    expected.label = "5 seconds"
  )
  
  msg <- "`parse_database_error()` will not {be able to parse this}, but it should still be successfully rethrown as-is."
  expect_snapshot(error = TRUE, rethrow_database_error(msg, call = NULL))
})

test_that("set_database_error_names() works (#840)", {
  expect_equal(
    set_database_error_names(c("unnamed", "vector")),
    c(x = "unnamed", `*` = "vector")
  )
  expect_equal(set_database_error_names("unnamed scalar"), c(x = "unnamed scalar"))
  expect_equal(
    set_database_error_names(c("i" = "partially", "named")),
    c(i = "partially", `*` = "named")
  )
  expect_equal(
    set_database_error_names(c("partially", "i" = "named")),
    c(x = "partially", i = "named")
  )
  expect_equal(set_database_error_names(c("i" = "named")), c(i = "named"))
})

test_that("set_odbcsysini() works (#791)", {
  skip_on_cran()
  skip_if(is_windows())

  expect_false(identical(Sys.getenv("ODBCSYSINI"), ""))
})

test_that("check_row.names()", {
  skip_if_no_unixodbc()
  con <- test_con("SQLITE")

  expect_snapshot(
    error = TRUE,
    dbWriteTable(con, "boopery", data.frame(bop = 1), row.names = c("no", "way"))
  )
})

test_that("check_field.types()", {
  skip_if_no_unixodbc()
  con <- test_con("SQLITE")

  expect_snapshot(
    error = TRUE,
    dbWriteTable(con, "boopery", data.frame(bop = 1), field.types = "numeric")
  )
})

test_that("check_attributes()", {
  skip_if_no_unixodbc()
  expect_snapshot(
    error = TRUE,
    con <- test_con("SQLITE", attributes = list(boop = "bop"))
  )

  expect_snapshot(
    error = TRUE,
    con <- test_con("SQLITE", attributes = list(boop = "bop", beep = "boop"))
  )
})

test_that("configure_simba() returns early on windows", {
  local_mocked_bindings(is_macos = function() {FALSE})

  expect_equal(configure_simba(), NULL)
})

test_that("configure_simba() errors informatively on failure to install unixODBC", {
  local_mocked_bindings(
    is_macos = function() {TRUE},
    locate_install_unixodbc = function() {character(0)}
  )

  expect_snapshot(configure_simba(), error = TRUE)
})

test_that("configure_simba() calls configure_unixodbc_simba() (#835)", {
  skip_if_not(is_macos())
  local_mocked_bindings(
    locate_install_unixodbc = function() "hey",
    configure_unixodbc_simba = function(...) TRUE
  )

  expect_true(configure_simba(list(path = "example")))
})

test_that("locate_install_unixodbc() returns reasonable values", {
  skip_if(!is_macos())
  skip_if(!has_unixodbc(), "odbcinst not available.")

  # odbc_config / pkg-config cflags point to nonexistent files on CRAN (#903)
  skip_on_cran()

  res <- locate_install_unixodbc()

  expect_true(file.exists(res[1]))
  expect_true(grepl("(\\.dylib|\\.a)$", res[1]))
})

test_that("databricks() errors informatively when spark ini isn't writeable", {
  local_mocked_bindings(is_writeable = function(path) {FALSE})
  expect_snapshot(
    write_simba_lines("", ".", ".", call2("databricks")),
    error = TRUE
  )
})

test_that("driver_dir(...) returns reasonable values", {
  path <- "/some/path/driver.so"
  local_mocked_bindings(odbcListDrivers = function() {
    data.frame(name = "OG Driver", attribute = "Driver",
      value = path, drop = FALSE)
  })
  expect_equal(driver_dir("OG Driver"), "/some/path")
  expect_equal(driver_dir(path), "/some/path")
})

test_that("spark_simba_config() returns reasonable values", {
  simba_spark_ini <- "some/folder/simba.sparkodbc.ini"
  withr::local_envvar(SIMBASPARKINI = simba_spark_ini)
  expect_equal(spark_simba_config("")$path, simba_spark_ini)
})

test_that("configure_unixodbc_simba() writes reasonable entries", {
  unixodbc_install_path <- "libodbcinst.dylib"
  spark_config_path <- "simba.sparkodbc.ini"

  withr::local_file(spark_config_path)

  # neither of the relevant fields already there:
  writeLines(
    c("some=entries", "not=relevant"),
    con = spark_config_path
  )

  expect_snapshot(configure_unixodbc_simba(
    unixodbc_install = unixodbc_install_path,
    simba_config = spark_config_path,
    action = "warn"
  ))
  configure_unixodbc_simba(
    unixodbc_install = unixodbc_install_path,
    simba_config = spark_config_path,
    action = "modify"
  )

  expect_equal(
    readLines(spark_config_path),
    c(
      "some=entries",
      "not=relevant",
      "ODBCInstLib=libodbcinst.dylib",
      "DriverManagerEncoding=UTF-16"
    )
  )

  # both of the relevant fields are already there
  # but point to incorrect values
  writeLines(
    c("some=entries",
      "not=relevant",
      "ODBCInstLib=somewhere.dylib",
      "DriverManagerEncoding=UTF-8"),
    con = spark_config_path
  )

  expect_snapshot(configure_unixodbc_simba(
    unixodbc_install = unixodbc_install_path,
    simba_config = spark_config_path,
    action = "warn"
  ))
  res <- configure_unixodbc_simba(
    unixodbc_install = unixodbc_install_path,
    simba_config = spark_config_path,
    action = "modify"
  )

  expect_equal(res, NULL)
  expect_equal(
    readLines(spark_config_path),
    c(
      "some=entries",
      "not=relevant",
      "ODBCInstLib=libodbcinst.dylib",
      "DriverManagerEncoding=UTF-16"
    )
  )

  # One entry correct, other incorrect
  # expect warning with single suggestion
  # when action is "warn"
  writeLines(
    c("some=entries",
      "not=relevant",
      "ODBCInstLib=libodbcinst.dylib",
      "DriverManagerEncoding=UTF-32"),
    con = spark_config_path
  )

  expect_snapshot(configure_unixodbc_simba(
    unixodbc_install = unixodbc_install_path,
    simba_config = spark_config_path,
    action = "warn"
  ))
  res <- configure_unixodbc_simba(
    unixodbc_install = unixodbc_install_path,
    simba_config = spark_config_path,
    action = "modify"
  )

  expect_equal(res, NULL)
  expect_equal(
    readLines(spark_config_path),
    c(
      "some=entries",
      "not=relevant",
      "ODBCInstLib=libodbcinst.dylib",
      "DriverManagerEncoding=UTF-16"
    )
  )

  # an entry is there but commented out
  writeLines(
    c("some=entries",
      "not=relevant",
      ";ODBCInstLib=somewhere.dylib",
      ";DriverManagerEncoding=UTF-8"),
    con = spark_config_path
  )

  expect_snapshot(configure_unixodbc_simba(
    unixodbc_install = unixodbc_install_path,
    simba_config = spark_config_path,
    action = "warn"
  ))
  configure_unixodbc_simba(
    unixodbc_install = unixodbc_install_path,
    simba_config = spark_config_path,
    action = "modify"
  )

  expect_equal(
    readLines(spark_config_path),
    c(
      "some=entries",
      "not=relevant",
      ";ODBCInstLib=somewhere.dylib",
      ";DriverManagerEncoding=UTF-8",
      "ODBCInstLib=libodbcinst.dylib",
      "DriverManagerEncoding=UTF-16"
    )
  )

  # Finally, a good config
  writeLines(
    c("some=entries",
      "not=relevant",
      "ODBCInstLib=libodbcinst.dylib",
      "DriverManagerEncoding=UTF-16"),
    con = spark_config_path
  )

  expect_no_warning(configure_unixodbc_simba(
    unixodbc_install = unixodbc_install_path,
    simba_config = spark_config_path,
    action = "warn"
  ))
})

# Connection string helpers: build_connection_string,
# decompose_connection_string, sanitize_connection_string

test_that("handles simple inputs", {
  expect_equal(build_connection_string(), "")
  expect_equal(build_connection_string(list(foo = "1")), "foo=1")
  expect_equal(build_connection_string(list(foo = "1", bar = "2")), "foo=1;bar=2")
})

test_that("combines with existing .connection string", {
  expect_equal(build_connection_string(string = "x=1"), "x=1")
  expect_equal(build_connection_string(list(foo = "1"), "x=1"), "x=1;foo=1")
  expect_equal(build_connection_string(list(foo = "1"), "x=1;"), "x=1;foo=1")
})

test_that("Handles connection string as expected", {
  conn_string <- build_connection_string(list(dsn = "abc",
    k1 = quote_value("Quoted Value"),
    k2 = quote_value("Quated value with a ' character")))
  expect_equal(decompose_connection_string(conn_string), list(dsn = "abc",
    k1 = "Quoted Value", k2 = "Quated value with a ' character"))
  # Robust to ending with semi-column
  expect_equal(decompose_connection_string("abc=1;def=2;"), list(abc = "1",
    def = "2"))
  # Robust to starting with semi-column
  expect_equal(decompose_connection_string(";abc=1;def=2"), list(abc = "1",
    def = "2"))
  # Robust to double semi-column
  expect_equal(decompose_connection_string("abc=1;;def=2"), list(abc = "1",
    def = "2"))
  # Key-value entries with missing key are filtered out
  expect_equal(decompose_connection_string("abc=1;def=2;=3"), list(abc = "1",
    def = "2"))
  # Key-value entries with missing value are filtered out
  expect_equal(decompose_connection_string("abc=1;def=2;ghi="), list(abc = "1",
    def = "2"))
  # Connection string segments that are not formatted as key-value
  # pairs are filtered out
  expect_equal(decompose_connection_string("abc=1;def=2;ghi"), list(abc = "1",
    def = "2"))
})

test_that("Sanitize filters out auth keys", {
  lst <- list(dsn = "abc", useNativeQuery = "def", pwd = "p", PWD = "P",
    password = "pass", PASSWORD = "PASSWORD", token = "token", TOKEN = "TOKEN",
    MYPWD = "mywd")
  expect_equal(sanitize_connection_string(lst), list(dsn = "abc",
    useNativeQuery = "def"))
})

test_connection_string <- function(db) {
  testthat::skip_on_cran()

  var <- paste0("ODBC_CS_", db)
  cs <- Sys.getenv(var)
  if (cs == "") {
    skip(paste0("env var '", var, "' not set"))
  }
  list(.connection_string = cs)
}

test_con <- function(db, ...) {
  dbConnect(
    odbc::odbc(),
    .connection_string = test_connection_string(db),
    ...
  )
}

local_table <- function(con, name, df, ..., envir = parent.frame()) {
  dbWriteTable(con, name, df, ...)
  withr::defer(dbRemoveTable(con, name), envir = envir)

  name
}

skip_if_no_drivers <- function() {
  if (nrow(odbcListDrivers()) == 0) {
    skip("No drivers installed")
  }
}

skip_if_no_unixodbc <- function() {
  if (!is_windows() && length(locate_install_unixodbc()) == 0) {
    skip("No unixodbc installation found.")
  }
}

#' Test round tripping a simple table
#'
#' This tests all the supported data types, including missing values. It first
#' writes them to the database, then reads them back and verifies the data is
#' identical to the original.
#'
#' This function is not exported and should only be used during tests and as a
#' sanity check when writing new `odbcDataType()` methods.
#'
#' @param con An established DBI connection.
#' @param columns Table columns to exclude (default) or include, dependent on
#' the value of `invert`. One of `datetime`, `date`, `binary`,
#' `integer`, `double`, `character`, `logical`.
#' @param invert If `TRUE`, change the definition of columns to be exclusive,
#' rather than inclusive.
#' @param force_sorted If `TRUE`, a sorted `id` column is added to the sent
#' data, and the received data is sorted by this column before doing the
#' comparison. This is necessary for some databases that do not preserve row
#' order.
#' @examples
#' \dontrun{
#' test_roundtrip(con)
#'
#' # exclude a few columns
#' test_roundtrip(con, c("integer", "double"))
#'
#' # Only test a specific column
#' test_roundtrip(con, "integer", invert = FALSE)
#' }
test_roundtrip <- function(con, columns = "", invert = TRUE, force_sorted = FALSE) {
  dbms <- dbGetInfo(con)$dbms.name
  res <- list()
  testthat::test_that(paste0("[", dbms, "] round tripping data.frames works"), {
    # on.exit(try(DBI::dbRemoveTable(con, "test_table"), silent = TRUE))
    set.seed(42)

    iris <- datasets::iris

    # We can't use the data.frame constructor directly as list columns don't work there.
    sent <- list(

      # We always return strings as factors
      # factor = iris$Species,
      datetime = as.POSIXct(as.numeric(iris$Petal.Length * 10), origin = "2016-01-01", tz = "UTC"),
      date = as.Date(iris$Sepal.Width * 100, origin = Sys.time()),
      time = hms::hms(seconds = sample.int(24 * 60 * 60, NROW(iris))),
      binary = blob::as_blob(lapply(seq_len(NROW(iris)), function(x) as.raw(sample(0:100, size = sample(0:25, 1))))),
      integer = as.integer(iris$Petal.Width * 100),
      double = iris$Sepal.Length,
      character = as.character(iris$Species),
      logical = sample(c(TRUE, FALSE), size = nrow(iris), replace = T)
    )
    attributes(sent) <- list(names = names(sent), row.names = c(NA_integer_, -length(sent[[1]])), class = "data.frame")

    # Add a proportion of NA values to a data frame
    add_na <- function(x, p = .1) {
      is.na(x) <- stats::runif(length(x)) < p
      x
    }
    sent[] <- lapply(sent, add_na, p = .1)
    if (isTRUE(invert)) {
      sent <- sent[, !names(sent) %in% columns]
    } else {
      sent <- sent[, names(sent) %in% columns]
    }
    if (force_sorted) sent$id <- seq_len(NROW(iris))

    DBI::dbWriteTable(con, "test_table", sent, overwrite = TRUE)
    received <- DBI::dbReadTable(con, "test_table")
    if (force_sorted) received <- received[order(received$id), ]
    row.names(received) <- NULL
    testthat::expect_equal(sent, received)
    res <<- list(sent = sent, received = received)
  })
  invisible(res)
}

#' Set the Transaction Isolation Level for a Connection
#'
#' @param levels One or more of `r odbc:::choices_rd(names(odbc:::transactionLevels()))`.
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @seealso <https://learn.microsoft.com/en-us/sql/odbc/reference/develop-app/setting-the-transaction-isolation-level>
#' @export
#' @examples
#' \dontrun{
#' # Can use spaces or underscores in between words.
#' odbcSetTransactionIsolationLevel(con, "read uncommitted")
#'
#' # Can also use the full constant name.
#' odbcSetTransactionIsolationLevel(con, "SQL_TXN_READ_UNCOMMITTED")
#' }
odbcSetTransactionIsolationLevel <- function(conn, levels) {
  arg_match(levels, values = names(transactionLevels()), multiple = TRUE)
  # Convert to lowercase, spaces to underscores, remove sql_txn prefix
  levels <- tolower(levels)
  levels <- gsub(" ", "_", levels)
  levels <- sub("sql_txn_", "", levels)
  levels <- match.arg(tolower(levels), names(transactionLevels()), several.ok = TRUE)

  set_transaction_isolation(conn@ptr, transactionLevels()[levels])
}

#' Return the corresponding ODBC data type for an R object
#'
#' @description
#'
#' This is used when creating a new table with `dbWriteTable()`.
#' Databases with default methods defined are:
#'
#' - MySQL
#' - PostgreSQL
#' - SQL Server
#' - Oracle
#' - SQLite
#' - Spark
#' - Hive
#' - Impala
#' - Redshift
#' - Vertica
#' - BigQuery
#' - Teradata
#' - Access
#' - Snowflake
#'
#' @details
#'
#' If you are using a different database and `dbWriteTable()` fails with a SQL
#' parsing error the default method is not appropriate, you will need to write
#' a new method. The object type for your method will be the database name
#' retrieved by `dbGetInfo(con)$dbms.name`. Use the documentation provided with
#' your database to determine appropriate values for each R data type.
#'
#' @param con A driver connection object, as returned by `dbConnect()`.
#' @param obj An R object.
#' @param ... Additional arguments passed to methods.
#' @return Corresponding SQL type for the `obj`.
#' @export
setGeneric(
  "odbcDataType",
  valueClass = "character",
  function(con, obj, ...) {
    standardGeneric("odbcDataType")
  }
)

#' @export
#' @rdname odbcDataType
#' @usage NULL
setMethod("odbcDataType", "ANY",
  function(con, obj, ...) {
    switch_type(
      obj,
      factor = "VARCHAR(255)",
      datetime = "TIMESTAMP",
      date = "DATE",
      time = "TIME",
      binary = "VARBINARY(255)",
      integer = "INTEGER",
      int64 = "INTEGER",
      double = "DOUBLE PRECISION",
      character = "VARCHAR(255)",
      logical = "BIT", # only valid if DB supports Null fields
      list = "VARCHAR(255)",
      stop("Unsupported type", call. = FALSE)
    )
  }
)

switch_type <- function(obj, ...) {
  switch(object_type(obj),
    ...
  )
}

object_type <- function(obj) {
  if (is.factor(obj)) {
    return("factor")
  }
  if (is(obj, "POSIXct")) {
    return("datetime")
  }
  if (is(obj, "Date")) {
    return("date")
  }
  if (is_blob(obj)) {
    return("binary")
  }
  if (is(obj, "difftime")) {
    return("time")
  }
  if (is(obj, "integer64")) {
    return("int64")
  }

  return(typeof(obj))
}

is_blob <- function(obj) {
  if (is(obj, "blob")) {
    return(TRUE)
  }
  if (is.object(obj) && any(class(obj) != "AsIs")) {
    return(FALSE)
  }
  if (!is.list(obj)) {
    return(FALSE)
  }

  # Assuming raw inside naked lists if the first non-NULL element is raw,
  # not checking the other elements
  for (i in seq_along(obj)) {
    x <- obj[[i]]
    if (!is.null(x)) {
      return(is.raw(x))
    }
  }

  TRUE
}

varchar <- function(x, type = "varchar") {
  # at least 255 characters, use max if more than 8000:
  max_length <- max(c(255, nchar(as.character(x))), na.rm = TRUE)

  if (max_length > 8000) {
    max_length <- "max"
  }

  paste0(type, "(", max_length, ")")
}

varbinary <- function(x, type = "varbinary") {
  # at least 255 bytes, use max if more than 8000:
  max_length <- max(c(255, lengths(x)), na.rm = TRUE)

  if (max_length > 8000) {
    max_length <- "max"
  }

  paste0(type, "(", max_length, ")")
}

#' @export
#' @rdname DBI-classes
setClass("DB2/AIX64", contains = "OdbcConnection")
setClass("DB2/LINUXX8664", contains = "DB2/AIX64")

#' @rdname DBI-methods
# Don't generate a usage because there's some buglet in R CMD check
# (probably because of the `/` in the class name) which flags a usage
# without corresponding alias
#' @usage NULL
setMethod("sqlCreateTable", "DB2/AIX64",
  function(con,
           table,
           fields,
           row.names = NA,
           temporary = FALSE,
           ...,
           field.types = NULL) {
    check_bool(temporary)
    check_row.names(row.names)
    check_field.types(field.types)
    table <- dbQuoteIdentifier(con, table)
    fields <- createFields(con, fields, field.types, row.names)

    SQL(paste0(
      if (temporary) "DECLARE GLOBAL TEMPORARY" else "CREATE",
      " TABLE ", table, " (\n",
      "  ", paste(fields, collapse = ",\n  "),
      "\n)\n", if (temporary) " ON COMMIT PRESERVE ROWS"
    ))
  }
)


#' @rdname DBI-methods
# In addition to the `ODBC` API call to `SQLTables`, also
# attempt to query SYSIBMADM.ADMINTEMPTABLES for any tables
# that match the arguments.
#' @usage NULL
setMethod("odbcConnectionTables", c("DB2/AIX64", "character"),
  function(conn,
           name,
           catalog_name = NULL,
           schema_name = NULL,
           table_type = NULL,
           exact = FALSE) {

    res <- callNextMethod()
    dfTempTables <- NULL
    if ( (is.null(schema_name) || schema_name == "SESSION") &&
         (is.null(catalog_name) || tolower(catalog_name) == conn@info$username) ) {
      # Also look for temp tables in this case
      # Notes on DB2 temp tables:
      # https://www.ibm.com/docs/en/db2-for-zos/12?topic=statements-declare-global-temporary-table
      # Notes on DB2 ADMINTEMPTABLES admin view:
      # https://www.ibm.com/docs/en/db2woc?topic=aracp-admin-get-temp-tables-retrieve-information-temporary-tables
      qTable <- getSelector("TABNAME", name, exact)
      qSchema <- getSelector("TABSCHEMA", schema_name, exact)
      query <- paste0(
        "SELECT
            INSTANTIATOR AS \"table_catalog\",
            TABSCHEMA AS \"table_schema\",
            TABNAME AS \"table_name\",
            'TABLE' AS \"table_type\",
            NULL AS \"table_remarks\"
         FROM SYSIBMADM.ADMINTEMPTABLES ",
        "WHERE 1 = 1 ", qTable, qSchema)
      tryCatch({
        dfTempTables <- dbGetQuery(conn, query)
        # NULL as colname is not well liked by DB2 it seems. Hack here. (#905).
        if (nrow(dfTempTables)  > 0) {
          dfTempTables$table_remarks <- NA_character_
        }
      }, odbc_database_error = function(e) {
        cli::cli_warn("Unable to query for temporary tables in SYSIBMADM.ADMINTEMPTABLES")
      })
    }
    rbind(res, dfTempTables)
  }
)

#' @export
#' @rdname DBI-classes
setClass("Hive", contains = "OdbcConnection")

# only need to override dbQuteString when x is character.
# DBI:::quote_string just returns x when it is of class SQL, so no need to override that.
#' @rdname DBI-methods
setMethod("dbQuoteString", c("Hive", "character"),
  function(conn, x, ...) {
    if (is(x, "SQL")) {
      return(x)
    }
    x <- gsub("'", "\\\\'", enc2utf8(x))
    if (length(x) == 0L) {
      DBI::SQL(character())
    } else {
      str <- paste0("'", x, "'")
      str[is.na(x)] <- "NULL"
      DBI::SQL(str)
    }
  }
)


#' @export
#' @rdname odbcDataType
#' @usage NULL
setMethod("odbcDataType", "Hive",
  function(con, obj, ...) {
    switch_type(
      obj,
      factor = "STRING",
      datetime = "TIMESTAMP",
      date = "DATE",
      binary = "BINARY",
      integer = "INT",
      int64 = "INT",
      double = "DOUBLE",
      character = "STRING",
      logical = "BOOLEAN",
      list = "STRING",
      time = ,
      stop("Unsupported type", call. = FALSE)
    )
  }
)

#' @rdname OdbcConnection
#' @keywords internal
#' @export
setClass(
  "OdbcConnection",
  contains = "DBIConnection",
  slots = list(
    ptr = "externalptr",
    quote = "character",
    info = "ANY",
    encoding = "character",
    connectionString = "list"
  )
)


# TODO: show encoding, timezone, bigint mapping
#' @rdname OdbcConnection
#' @inheritParams methods::show
#' @export
setMethod("show", "OdbcConnection",
  function(object) {
    info <- dbGetInfo(object)

    cat(
      sep = "", "<OdbcConnection>",
      if (nzchar(info[["servername"]])) {
        paste0(
          " ",
          if (nzchar(info[["username"]])) paste0(info[["username"]], "@"),
          info[["servername"]], "\n"
        )
      },
      if (!dbIsValid(object)) {
        "  DISCONNECTED\n"
      } else {
        paste0(
          collapse = "",
          if (nzchar(info[["dbname"]])) {
            paste0("  Database: ", info[["dbname"]], "\n")
          },
          if (nzchar(info[["dbms.name"]]) && nzchar(info[["db.version"]])) {
            paste0("  ", info[["dbms.name"]], " ", "Version: ", info[["db.version"]], "\n")
          },
          NULL
        )
      }
    )
  }
)


#' @rdname OdbcConnection
#' @param dbObj An object inheriting from `DBIObject`, i.e. `DBIDriver`,
#' `DBIConnection`, or a `DBIResult`.
#' @inheritParams DBI::dbIsValid
#' @export
setMethod("dbIsValid", "OdbcConnection",
  function(dbObj, ...) {
    connection_valid(dbObj@ptr)
  }
)

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbDisconnect
#' @export
setMethod("dbDisconnect", "OdbcConnection",
  function(conn, ...) {
    if (!dbIsValid(conn)) {
      warning("Connection already closed.", call. = FALSE)
    }

    on_connection_closed(conn)
    connection_release(conn@ptr)
    invisible(TRUE)
  }
)

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbSendQuery
#' @param params Optional query parameters, passed on to [dbBind()]
#' @param immediate If `TRUE`, SQLExecDirect will be used instead of
#'   SQLPrepare, and the `params` argument is ignored
#' @export
setMethod("dbSendQuery", c("OdbcConnection", "character"),
  function(conn, statement, params = NULL, ..., immediate = FALSE) {
    if (has_result(conn@ptr)) {
      cli::cli_warn("Cancelling previous query")
    }
    OdbcResult(
      connection = conn,
      statement = statement,
      params = params,
      immediate = immediate
    )
  }
)

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbExecute
#' @export
setMethod("dbExecute", c("OdbcConnection", "character"),
  function(conn, statement, params = NULL, ..., immediate = is.null(params)) {
    rs <- dbSendStatement(conn, statement, params = params, ..., immediate = immediate)
    on.exit(dbClearResult(rs))

    dbGetRowsAffected(rs)
  }
)

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbSendStatement
#' @param params Query parameters to pass to [DBI::dbBind()].
#'   See [DBI::dbBind()] for details.
#' @export
setMethod("dbSendStatement", c("OdbcConnection", "character"),
  function(conn, statement, params = NULL, ..., immediate = FALSE) {
    if (has_result(conn@ptr)) {
      cli::cli_warn("Cancelling previous query")
    }
    OdbcResult(
      connection = conn,
      statement = statement,
      params = params,
      immediate = immediate
    )
  }
)

#' @rdname OdbcConnection
#' @inheritParams DBI::dbDataType
#' @export
setMethod("dbDataType", "OdbcConnection",
  function(dbObj, obj, ...) {
    odbcDataType(dbObj, obj)
  }
)

#' @rdname OdbcConnection
#' @inheritParams DBI::dbDataType
#' @export
setMethod("dbDataType", c("OdbcConnection", "data.frame"),
  function(dbObj, obj, ...) {
    odbc_data_type_df(dbObj, obj)
  }
)

#' @rawNamespace exportMethods(dbQuoteString)
NULL

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbQuoteIdentifier
#' @export
setMethod("dbQuoteIdentifier", c("OdbcConnection", "character"),
  function(conn, x, ...) {
    if (length(x) == 0L) {
      return(DBI::SQL(character()))
    }
    if (any(is.na(x))) {
      cli::cli_abort("{.arg x} can't be {.code NA}.")
    }
    if (nzchar(conn@quote)) {
      x <- gsub(conn@quote, paste0(conn@quote, conn@quote), x, fixed = TRUE)
    }
    nms <- names(x)
    res <- DBI::SQL(paste(conn@quote, x, conn@quote, sep = ""))
    names(res) <- nms
    res
  }
)

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbQuoteIdentifier
#' @export
setMethod("dbQuoteIdentifier", c("OdbcConnection", "SQL"),
  getMethod("dbQuoteIdentifier", c("DBIConnection", "SQL"), asNamespace("DBI"))
)

#' List remote tables and fields for an ODBC connection
#'
#' `dbListTables()` provides names of remote tables accessible through this
#' connection; `dbListFields()` provides names of columns within a table.
#'
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbListTables
#' @param catalog_name,schema_name,table_name Catalog, schema, and table names.
#'
#'   By default, `catalog_name`, `schema_name` and `table_name` will
#'   automatically escape underscores to ensure that you match exactly one
#'   table. If you want to search for multiple tables using wild cards, you
#'   will need to use `odbcConnectionTables()` directly instead.
#'
#' @param table_type The type of the table to return, the default returns
#'   all table types.
#' @returns A character vector of table or field names respectively.
#' @export
setMethod("dbListTables", "OdbcConnection",
  function(conn,
           catalog_name = NULL,
           schema_name = NULL,
           table_name = NULL,
           table_type = NULL,
           ...) {
    check_string(catalog_name, allow_null = TRUE)
    check_string(schema_name, allow_null = TRUE)
    check_string(table_name, allow_null = TRUE)
    check_string(table_type, allow_null = TRUE)

    tables <- odbcConnectionTables(
      conn,
      name = table_name,
      catalog_name = catalog_name,
      schema_name = schema_name,
      table_type = table_type,
      exact = TRUE
    )
    tables[["table_name"]]
  }
)

#' @rdname OdbcConnection
#' @inheritParams DBI::dbGetInfo
#' @export
setMethod("dbGetInfo", "OdbcConnection",
  function(dbObj, ...) {
    connection_info(dbObj@ptr)
  }
)

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbGetQuery
#' @inheritParams DBI::dbFetch
#' @export
setMethod("dbGetQuery", c("OdbcConnection", "character"),
  function(conn,
           statement,
           n = -1,
           params = NULL,
           immediate = is.null(params),
           ...) {
    rs <- dbSendQuery(
      conn,
      statement,
      params = params,
      immediate = immediate,
      ...
    )
    on.exit(dbClearResult(rs))

    df <- dbFetch(rs, n = n, ...)

    if (!dbHasCompleted(rs)) {
      warning("Pending rows", call. = FALSE)
    }

    df
  }
)

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbBegin
#' @export
setMethod("dbBegin", "OdbcConnection",
  function(conn, ...) {
    connection_begin(conn@ptr)
    invisible(TRUE)
  }
)

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbCommit
#' @export
setMethod("dbCommit", "OdbcConnection",
  function(conn, ...) {
    connection_commit(conn@ptr)
    invisible(TRUE)
  }
)

#' @rdname OdbcConnection
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbRollback
#' @export
setMethod("dbRollback", "OdbcConnection",
  function(conn, ...) {
    connection_rollback(conn@ptr)
    invisible(TRUE)
  }
)

#' @include dbi-connection.R
NULL

#' @export
#' @rdname DBI-classes
setClass("MySQL", contains = "OdbcConnection")

#' @export
#' @rdname odbcDataType
#' @usage NULL
setMethod("odbcDataType", "MySQL",
  function(con, obj, ...) {
    switch_type(
      obj,
      factor = "TEXT",
      datetime = "DATETIME",
      date = "DATE",
      time = "TIME",
      binary = "BLOB",
      integer = "INTEGER",
      int64 = "INTEGER",
      double = "DOUBLE",
      character = "TEXT",
      logical = "TINYINT",
      list = "TEXT",
      stop("Unsupported type", call. = FALSE)
    )
  }
)

#' @include dbi-connection.R
#' @include dbi-driver.R
NULL


#' Helper for Connecting to Databricks via ODBC
#'
#' @description
#'
#' Connect to Databricks clusters and SQL warehouses via the [Databricks ODBC
#' driver](https://www.databricks.com/spark/odbc-drivers-download).
#'
#' In particular, the custom `dbConnect()` method for the Databricks ODBC driver
#' implements a variant of Databricks's [unified authentication](https://docs.databricks.com/aws/en/dev-tools/auth/unified-auth)
#' model when no `uid` or `pwd` is supplied, checking for ambient credentials in
#' the following order:
#'
#' * Viewer-based or service principal credentials supplied by Posit Connect
#'   (requires the \pkg{connectcreds} package).
#' * Personal access tokens.
#' * Workload identity federation.
#' * OAuth machine-to-machine credentials.
#' * OAuth user-to-machine credentials suplied via Posit Workbench or the
#'   Databricks CLI on desktop.
#'
#' This aims to provide broad compatibility between \pkg{odbc} and the standard
#' environment variables used by Databricks SDKs in R and other languages.
#'
#' In addition, on macOS platforms, the `dbConnect()` method will check
#' for irregularities with how the driver is configured,
#' and attempt to fix in-situ, unless the `odbc.no_config_override`
#' environment variable is set.
#'
#' @param drv an object that inherits from [DBI::DBIDriver-class],
#' or an existing [DBI::DBIConnection-class]
#' object (in order to clone an existing connection).
#' @param httpPath,HTTPPath To query a cluster, use the HTTP Path value found
#'   under `Advanced Options > JDBC/ODBC` in the Databricks UI. For SQL
#'   warehouses, this is found under `Connection Details` instead.
#' @param useNativeQuery Suppress the driver's conversion from ANSI SQL 92 to
#'   HiveSQL.  When set to true, we also set the following related flags:
#'   `EnableNativeParameterizedQuery`, and `PopulateParametersForNativeQuery`.
#'   The default (`TRUE`), gives greater performance but could cause issues
#'   with paramterised queries (and hence `dbWriteTable()`).
#' @param workspace The URL of a Databricks workspace, e.g.
#'   `"https://example.cloud.databricks.com"`.
#' @param driver The name of the Databricks ODBC driver, or `NULL` to use the
#'   default name.
#' @param uid,pwd Manually specify a username and password for authentication.
#'   Specifying these options will disable automated credential discovery.
#' @param ... Further arguments passed on to [`dbConnect()`].
#'
#' @returns An `OdbcConnection` object with an active connection to a Databricks
#'   cluster or SQL warehouse.
#'
#' @examples
#' \dontrun{
#' DBI::dbConnect(
#'   odbc::databricks(),
#'   httpPath = "sql/protocolv1/o/4425955464597947/1026-023828-vn51jugj"
#' )
#'
#' # Use credentials from the viewer or a service principal (when possible) in
#' # a Shiny app deployed to Posit Connect.
#' library(connectcreds)
#' server <- function(input, output, session) {
#'   conn <- DBI::dbConnect(
#'     odbc::databricks(),
#'     workspace = "https://example.cloud.databricks.com",
#'     httpPath = "sql/protocolv1/o/4425955464597947/1026-023828-vn51jugj"
#'   )
#' }
#' }
#' @export
databricks <- function() {
  new("DatabricksOdbcDriver")
}

#' @rdname databricks
#' @export
setClass("DatabricksOdbcDriver", contains = "OdbcDriver")

#' @rdname databricks
#' @export
setMethod("dbConnect", "DatabricksOdbcDriver",
  function(drv,
           httpPath,
           workspace = Sys.getenv("DATABRICKS_HOST"),
           useNativeQuery = TRUE,
           driver = NULL,
           HTTPPath,
           uid = NULL,
           pwd = NULL,
           ...) {
    call <- caller_env()
    # For backward compatibility with RStudio connection string
    http_path <- check_exclusive(httpPath, HTTPPath, .call = call)
    check_string(get(http_path), allow_null = TRUE, arg = http_path, call = call)
    check_string(workspace, allow_null = TRUE, call = call)
    check_bool(useNativeQuery, call = call)
    check_string(driver, allow_null = TRUE, call = call)
    check_string(uid, allow_null = TRUE, call = call)
    check_string(pwd, allow_null = TRUE, call = call)

    args <- databricks_args(
      httpPath = if (missing(httpPath)) HTTPPath else httpPath,
      workspace = workspace,
      useNativeQuery = useNativeQuery,
      driver = driver,
      uid = uid,
      pwd = pwd,
      ...
    )
    # Perform some sanity checks on MacOS
    configure_simba(spark_simba_config(args$driver),
      action = "modify")
    inject(dbConnect(odbc(), !!!args))
  }
)

databricks_args <- function(httpPath,
                            workspace = Sys.getenv("DATABRICKS_HOST"),
                            useNativeQuery = FALSE,
                            driver = NULL,
                            uid = NULL,
                            pwd = NULL,
                            ...) {
  host <- databricks_host(workspace)

  args <- databricks_default_args(
    driver = driver,
    host = host,
    httpPath = httpPath,
    useNativeQuery = useNativeQuery
  )

  auth <- databricks_auth_args(host, uid = uid, pwd = pwd)
  all <- utils::modifyList(c(args, auth), list(...))

  arg_names <- tolower(names(all))
  if (!"authmech" %in% arg_names && !all(c("uid", "pwd") %in% arg_names)) {
    msg <- c(
      "Failed to detect ambient Databricks credentials.",
      "i" = "Supply {.arg uid} and {.arg pwd} to authenticate manually."
    )
    if (running_on_connect()) {
      msg <- c(
        msg,
        "i" = "Or consider enabling Posit Connect's Databricks integration. \
              For viewer-based credentials, see {.url \
              https://docs.posit.co/connect/user/oauth-integrations/#viewer-oauth-integrations}
              for details. \
              For service principal credentials, see {.url \
              https://docs.posit.co/connect/user/oauth-integrations/#service-account-oauth-integrations}
              for details."
      )
    }
    cli::cli_abort(msg, call = quote(DBI::dbConnect()))
  }

  all
}

databricks_default_args <- function(driver, host, httpPath, useNativeQuery) {
  ret <- list(
    driver = driver %||% databricks_default_driver(),
    host = host,
    httpPath = httpPath,
    thriftTransport = 2,
    userAgentEntry = databricks_user_agent(),
    # Connections to Databricks are always over HTTPS.
    port = 443,
    protocol = "https",
    ssl = 1
  )

  nativeQueryArgs <- list(useNativeQuery = as.integer(useNativeQuery))
  if (useNativeQuery) {
    # Per driver documentation, when native query is enabled, the additional two flags help
    # with properly handling parametrized queries
    nativeQueryArgs <- c(nativeQueryArgs, EnableNativeParameterizedQuery = 1, PopulateParametersForNativeQuery = 1)
  }

  ret <- c(ret, nativeQueryArgs)
  ret
}



# Returns a sensible driver name even if odbc.ini and odbcinst.ini do not
# contain an entry for the Databricks ODBC driver. For Linux and macOS we
# default to known shared library paths used by the official installers.
# On Windows we use the official driver name.
databricks_default_driver <- function() {
  find_default_driver(
    databricks_default_driver_paths(),
    fallbacks = c("Databricks", "Simba Spark ODBC Driver"),
    label = "Databricks/Spark",
    call = quote(DBI::dbConnect())
  )
}

databricks_default_driver_paths <- function() {
  if (Sys.info()["sysname"] == "Linux") {
    paths <- Sys.glob(c(
      "/opt/rstudio-drivers/spark/bin/lib/libsparkodbc_sb*.so",
      "/opt/simba/spark/lib/64/libsparkodbc_sb*.so"
    ))
  } else if (Sys.info()["sysname"] == "Darwin") {
    paths <- Sys.glob(c(
      "/Library/simba/spark/lib/libsparkodbc_sb*.dylib",
      "/Library/databricks/databricksodbc/lib/libdatabricks*.dylib"
    ))
  } else {
    paths <- character()
  }
  paths
}

databricks_host <- function(workspace) {
  if (nchar(workspace) == 0) {
    abort(
      c(
        "No Databricks workspace URL provided.",
        i = "Either supply `workspace` argument or set env var `DATABRICKS_HOST`."
      ),
      call = quote(DBI::dbConnect())
    )
  }
  gsub("https://|/$", "", workspace)
}

#' @importFrom utils packageVersion
databricks_user_agent <- function() {
  user_agent <- paste0("r-odbc/", packageVersion("odbc"))
  if (nchar(Sys.getenv("SPARK_CONNECT_USER_AGENT")) != 0) {
    # Respect the existing user-agent if present. Normally we'd append, but the
    # Databricks ODBC driver does not yet support space-separated entries in
    # this field.
    user_agent <- Sys.getenv("SPARK_CONNECT_USER_AGENT")
  }
  user_agent
}

databricks_auth_args <- function(host, uid = NULL, pwd = NULL) {
  # Detect viewer-based credentials from Posit Connect.
  workspace <- paste0("https://", host)
  if (is_installed("connectcreds") && connectcreds::has_viewer_token(workspace)) {
    token <- connectcreds::connect_viewer_token(workspace)
    return(list(
      authMech = 11,
      auth_flow = 0,
      auth_accesstoken = token$access_token
    ))
  }

  if (is_installed("connectcreds") && connectcreds::has_service_account_token(workspace)) {
    token <- connectcreds::connect_service_account_token(workspace)
    return(list(
      authMech = 11,
      auth_flow = 0,
      auth_accesstoken = token$access_token
    ))
  }

  if (!is.null(uid) && !is.null(pwd)) {
    return(list(uid = uid, pwd = pwd, authMech = 3))
  } else if (xor(is.null(uid), is.null(pwd))) {
    abort(
      c(
        "Both `uid` and `pwd` must be specified for manual authentication.",
        i = "Or leave both unset for automated authentication."
      ),
      call = quote(DBI::dbConnect())
    )
  }

  # Check some standard Databricks environment variables. This is used to
  # implement a subset of the "Databricks client unified authentication" model.
  token <- Sys.getenv("DATABRICKS_TOKEN")
  auth_type <- Sys.getenv("DATABRICKS_AUTH_TYPE")
  client_id <- Sys.getenv("DATABRICKS_CLIENT_ID")
  client_secret <- Sys.getenv("DATABRICKS_CLIENT_SECRET")
  cli_path <- Sys.getenv("DATABRICKS_CLI_PATH", "databricks")
  cfg_file <- Sys.getenv("DATABRICKS_CONFIG_FILE")

  # Check for Workbench-provided credentials.
  wb_token <- NULL
  if (grepl("posit-workbench", cfg_file, fixed = TRUE)) {
    wb_token <- workbench_databricks_token(host, cfg_file)
  }

  if (nchar(token) != 0) {
    # An explicit PAT takes precedence over everything else.
    list(
      authMech = 3,
      uid = "token",
      pwd = token
    )
  } else if (auth_type %in% c("env-oidc", "file-oidc")) {
    # Next up is workload identity federation.
    if (nchar(client_id) == 0) {
      cli::cli_abort(
        c(
          "Workload identity federation requires a service principal.",
          "i" = "Set the {.envvar DATABRICKS_CLIENT_ID} environment variable to
                 the service principal's UUID."
        ),
        call = quote(DBI::dbConnect())
      )
    }
    id_token <- databricks_read_id_token(auth_type)
    access_token <- databricks_token_exchange(host, id_token, client_id)
    list(
      authMech = 11,
      auth_flow = 0,
      auth_accesstoken = access_token
    )
  } else if (nchar(client_id) != 0) {
    # Next up are explicit OAuth2 M2M credentials.
    list(
      authMech = 11,
      auth_flow = 1,
      auth_client_id = client_id,
      auth_client_secret = client_secret
    )
  } else if (!is.null(wb_token)) {
    # Next up are Workbench-provided credentials.
    list(
      authMech = 11,
      auth_flow = 0,
      auth_accesstoken = wb_token
    )
  } else if (!is_hosted_session() && nchar(Sys.which(cli_path)) != 0) {
    # When on desktop, try using the Databricks CLI for auth.
    output <- suppressWarnings(
      system2(
        cli_path,
        c("auth", "token", "--host", host),
        stdout = TRUE,
        stderr = TRUE
      )
    )
    output <- paste(output, collapse = "\n")
    # If we don't get an error message, try to extract the token from the JSON-
    # formatted output.
    if (grepl("access_token", output, fixed = TRUE)) {
      token <- gsub(".*access_token\":\\s?\"([^\"]+).*", "\\1", output)
      list(
        authMech = 11,
        auth_flow = 0,
        auth_accesstoken = token
      )
    }
  }
}

databricks_read_id_token <- function(auth_type) {
  if (auth_type == "env-oidc") {
    token_env_name <- Sys.getenv("DATABRICKS_OIDC_TOKEN_ENV")
    if (nchar(token_env_name) == 0) {
      token_env_name <- "DATABRICKS_OIDC_TOKEN"
    }
    token <- Sys.getenv(token_env_name)
    if (nchar(token) == 0) {
      cli::cli_abort(
        c(
          "Workload identity federation is enabled
           ({.code DATABRICKS_AUTH_TYPE=env-oidc}) but no OIDC token was found.",
          "i" = "Set the {.envvar {token_env_name}} environment variable to the
                 IdP JWT."
        ),
        call = quote(DBI::dbConnect())
      )
    }
    token
  } else {
    path <- Sys.getenv("DATABRICKS_OIDC_TOKEN_FILEPATH")
    if (nchar(path) == 0) {
      cli::cli_abort(
        c(
          "Workload identity federation is enabled
           ({.code DATABRICKS_AUTH_TYPE=file-oidc}) but
           {.envvar DATABRICKS_OIDC_TOKEN_FILEPATH} is not set.",
          "i" = "Set {.envvar DATABRICKS_OIDC_TOKEN_FILEPATH} to the path of the
                 file containing the IdP JWT."
        ),
        call = quote(DBI::dbConnect())
      )
    }
    if (!file.exists(path)) {
      cli::cli_abort(
        c(
          "OIDC token file {.file {path}} does not exist.",
          "i" = "Verify {.envvar DATABRICKS_OIDC_TOKEN_FILEPATH} points to a
                 readable file."
        ),
        call = quote(DBI::dbConnect())
      )
    }
    token <- trimws(paste(readLines(path, warn = FALSE), collapse = ""))
    if (nchar(token) == 0) {
      cli::cli_abort(
        "OIDC token file {.file {path}} is empty.",
        call = quote(DBI::dbConnect())
      )
    }
    token
  }
}

databricks_token_exchange <- function(host, id_token, client_id) {
  check_installed("httr2", "for Databricks workload identity federation")

  url <- paste0("https://", host, "/oidc/v1/token")
  body <- list(
    grant_type = "urn:ietf:params:oauth:grant-type:token-exchange",
    subject_token_type = "urn:ietf:params:oauth:token-type:jwt",
    subject_token = id_token,
    scope = "all-apis",
    client_id = client_id
  )

  req <- httr2::request(url)
  req <- httr2::req_body_form(req, !!!body)
  req <- httr2::req_error(req, is_error = function(resp) FALSE)

  resp <- try_fetch(
    httr2::req_perform(req),
    error = function(cnd) {
      cli::cli_abort(
        c(
          "Failed to reach the Databricks OIDC token endpoint at {.url {url}}.",
          "i" = "Check network connectivity and that {.arg workspace} is correct."
        ),
        parent = cnd,
        call = quote(DBI::dbConnect())
      )
    }
  )

  if (httr2::resp_is_error(resp)) {
    status <- httr2::resp_status(resp)
    resp_body <- tryCatch(
      httr2::resp_body_string(resp),
      error = function(e) "(unreadable)"
    )
    cli::cli_abort(
      c(
        "Databricks OIDC token exchange failed with HTTP {status}.",
        "i" = "Response: {resp_body}"
      ),
      call = quote(DBI::dbConnect())
    )
  }

  parsed <- httr2::resp_body_json(resp)
  access_token <- parsed$access_token
  if (is.null(access_token) || !nzchar(access_token)) {
    cli::cli_abort(
      c(
        "Databricks OIDC token exchange response did not contain an access
         token.",
        "i" = "Verify the federation policy is configured correctly in
               Databricks."
      ),
      call = quote(DBI::dbConnect())
    )
  }
  access_token
}

# Try to determine whether we can redirect the user's browser to a server on
# localhost, which isn't possible if we are running on a hosted platform.
#
# This is based on the strategy pioneered by the {gargle} package and {httr2}.
is_hosted_session <- function() {
  # If RStudio Server or Posit Workbench is running locally (which is possible,
  # though unusual), it's not acting as a hosted environment.
  Sys.getenv("RSTUDIO_PROGRAM_MODE") == "server" &&
    !grepl("localhost", Sys.getenv("RSTUDIO_HTTP_REFERER"), fixed = TRUE)
}


is_camel_case <- function(x) {
  grepl("^[a-z]+([A-Z_][a-z]+)*$", x)
}

dot_names <- function(...) names(substitute(...()))

# Reads Posit Workbench-managed Databricks credentials from a
# $DATABRICKS_CONFIG_FILE. The generated file will look as follows:
#
# [workbench]
# host = some-host
# token = some-token
workbench_databricks_token <- function(host, cfg_file) {
  cfg <- readLines(cfg_file)
  # We don't attempt a full parse of the INI syntax supported by Databricks
  # config files, instead relying on the fact that this particular file will
  # always contain only one section.
  if (!any(grepl(host, cfg, fixed = TRUE))) {
    # The configuration doesn't actually apply to this host.
    return(NULL)
  }
  line <- grepl("token = ", cfg, fixed = TRUE)
  token <- gsub("token = ", "", cfg[line])
  if (nchar(token) == 0) {
    return(NULL)
  }
  token
}

# p. 44 https://downloads.datastax.com/odbc/2.6.5.1005/Simba%20Spark%20ODBC%20Install%20and%20Configuration%20Guide.pdf
spark_simba_config <- function(driver) {
  spark_env <- Sys.getenv("SIMBASPARKINI")
  URL <- "https://www.databricks.com/spark/odbc-drivers-download"
  if (!identical(spark_env, "")) {
    return(list(path = spark_env, url = URL))
  }
  # Attempt first the exact driver directory
  res <- list.files(
    driver_dir(driver),
    pattern = "simba\\.sparkodbc\\.ini$|databricks\\.databricksodbc\\.ini$",
    full.names = TRUE
  )
  if (!length(res)) {
    common_dirs <- unique(c(
      "/Library/simba/spark/lib",
      "/Library/databricks/databricksodbc/lib",
      "/etc",
      getwd(),
      Sys.getenv("HOME")
    ))
    res <- list.files(
      common_dirs,
      pattern = "simba\\.sparkodbc\\.ini$|databricks\\.databricksodbc\\.ini$",
      full.names = TRUE
    )
  }
  return(list(path = res, url = URL))
}

test_that("databricks arguments use camelcase", {
  withr::local_envvar(DATABRICKS_TOKEN = "", DATABRICKS_AUTH_TYPE = "")

  args <- databricks_args(
    "foo",
    "bar",
    pwd = "pwd",
    uid = "user",
    driver = "driver"
  )
  expect_true(all(is_camel_case(names(args))))
})

test_that("manually supplied arguments override automatic", {
  withr::local_envvar(DATABRICKS_TOKEN = "abc")
  args <- databricks_args("x", "y", driver = "driver")
  expect_equal(args$authMech, 3)

  args <- databricks_args("x", "y", driver = "driver", authMech = 123)
  expect_equal(args$authMech, 123)
})

test_that("fallbacks to driver name", {
  local_mocked_bindings(
    databricks_default_driver_paths = function() character(),
    odbcListDrivers = function() list(name = c("bar", "Databricks"))
  )
  expect_equal(databricks_default_driver(), "Databricks")
})

test_that("errors if can't find driver", {
  local_mocked_bindings(
    databricks_default_driver_paths = function() character(),
    odbcListDrivers = function() list()
  )
  expect_snapshot(databricks_default_driver(), error = TRUE)
})

test_that("databricks host validates inputs", {
  expect_equal(databricks_host("https://my-host.com"), "my-host.com")
  expect_equal(databricks_host("https://my-host.com/"), "my-host.com")
  expect_equal(databricks_host("my-host.com/"), "my-host.com")
  expect_snapshot(databricks_host(""), error = TRUE)
})

test_that("user agent respects envvar", {
  withr::local_envvar(SPARK_CONNECT_USER_AGENT = NULL)
  local_mocked_bindings(packageVersion = function(...) "1.0.0")
  expect_equal(databricks_user_agent(), "r-odbc/1.0.0")

  withr::local_envvar(SPARK_CONNECT_USER_AGENT = "my-odbc/1.0.0")
  expect_equal(databricks_user_agent(), "my-odbc/1.0.0")
})

test_that("errors if auth fails", {
  withr::local_envvar(
    DATABRICKS_TOKEN = "",
    DATABRICKS_AUTH_TYPE = "",
    DATABRICKS_CONFIG_FILE = NULL
  )

  databricks_args1 <- function(...) {
    databricks_args("path", "host", driver = "driver", ...)
  }

  expect_snapshot(. <- databricks_args1(), error = TRUE)

  expect_silent(databricks_args1(uid = "uid", pwd = "pwd"))
  expect_silent(databricks_args1(authMech = 10))
})

test_that("uid and pwd suppress automated auth", {
  auth <- databricks_auth_args("host", uid = "uid", pwd = "pwd")
  expect_equal(auth, list(uid = "uid", pwd = "pwd", authMech = 3))
})

test_that("must supply both uid and pwd", {
  expect_snapshot(databricks_auth_args("host", uid = "uid"), error = TRUE)
})

test_that("supports PAT in env var", {
  withr::local_envvar(DATABRICKS_TOKEN = "abc")
  expect_equal(databricks_auth_args("host")$pwd, "abc")
})

test_that("supports OAuth M2M in env var", {
  withr::local_envvar(
    DATABRICKS_TOKEN = "",
    DATABRICKS_CLIENT_ID = "abc",
    DATABRICKS_CLIENT_SECRET = "def"
  )

  auth <- databricks_auth_args("host")
  expect_equal(auth$auth_client_id, "abc")
  expect_equal(auth$auth_client_secret, "def")
})

test_that("dbConnect method handles httpPath aliases (#787)", {
  local_mocked_bindings(
    databricks_args = function(...) stop("made it"),
    configure_simba = function(...) TRUE
  )

  expect_error(dbConnect(databricks(), HTTPPath = "boop"), "made it")
  expect_error(dbConnect(databricks(), httpPath = "boop"), "made it")
})

test_that("dbConnect method errors informatively re: httpPath (#787)", {
  local_mocked_bindings(configure_simba = function(...) TRUE)

  expect_snapshot(
    error = TRUE,
    dbConnect(databricks(), httpPath = "boop", HTTPPath = "bop")
  )

  expect_snapshot(error = TRUE, dbConnect(databricks(), HTTPPath = 1L))
  expect_snapshot(error = TRUE, dbConnect(databricks(), httpPath = 1L))
})

test_that("Workbench-managed credentials are detected correctly", {
  # Emulate the databricks.cfg file written by Workbench.
  db_home <- tempfile("posit-workbench")
  dir.create(db_home)
  writeLines(
    c(
      '[workbench]',
      'host = some-host',
      'token = token'
    ),
    file.path(db_home, "databricks.cfg")
  )
  withr::local_envvar(
    DATABRICKS_CONFIG_FILE = file.path(db_home, "databricks.cfg")
  )
  args <- databricks_auth_args(host = "some-host")
  expect_equal(args, list(authMech = 11, auth_flow = 0, auth_accesstoken = "token"))
})

test_that("Workbench-managed credentials are ignored for other hosts", {
  # Emulate the databricks.cfg file written by Workbench.
  db_home <- tempfile("posit-workbench")
  dir.create(db_home)
  writeLines(
    c(
      '[workbench]',
      'host = nonmatching',
      'token = token'
    ),
    file.path(db_home, "databricks.cfg")
  )
  withr::local_envvar(
    DATABRICKS_CONFIG_FILE = file.path(db_home, "databricks.cfg")
  )
  expect_equal(databricks_auth_args(host = "some-host"), NULL)
})

test_that("we hint viewer-based and service principal credentials on Connect", {
  local_mocked_bindings(
    running_on_connect = function() TRUE
  )
  expect_snapshot(
    databricks_args(
      workspace = "workspace",
      httpPath = "path",
      driver = "driver"
    ),
    error = TRUE
  )
})

test_that("tokens can be requested from a Connect server", {
  skip_if_not_installed("connectcreds")

  connectcreds::local_mocked_connect_responses(token = "token")
  expect_equal(
    databricks_auth_args("example.cloud.databricks.com"),
    list(authMech = 11, auth_flow = 0, auth_accesstoken = "token")
  )
})

test_that("supports workload identity env-oidc with default token env var", {
  withr::local_envvar(
    DATABRICKS_TOKEN = "",
    DATABRICKS_CLIENT_ID = "my-sp",
    DATABRICKS_AUTH_TYPE = "env-oidc",
    DATABRICKS_OIDC_TOKEN = "my-jwt",
    DATABRICKS_OIDC_TOKEN_ENV = ""
  )
  local_mocked_bindings(
    databricks_token_exchange = function(host, id_token, client_id) {
      expect_equal(id_token, "my-jwt")
      expect_equal(client_id, "my-sp")
      "exchanged-access-token"
    }
  )
  auth <- databricks_auth_args("my.cloud.databricks.com")
  expect_equal(auth$authMech, 11)
  expect_equal(auth$auth_flow, 0)
  expect_equal(auth$auth_accesstoken, "exchanged-access-token")
})

test_that("supports workload identity env-oidc with custom token env var name", {
  withr::local_envvar(
    DATABRICKS_TOKEN = "",
    DATABRICKS_CLIENT_ID = "sp",
    DATABRICKS_AUTH_TYPE = "env-oidc",
    DATABRICKS_OIDC_TOKEN_ENV = "MY_CUSTOM_JWT_VAR",
    MY_CUSTOM_JWT_VAR = "custom-jwt"
  )
  local_mocked_bindings(
    databricks_token_exchange = function(host, id_token, client_id) {
      expect_equal(id_token, "custom-jwt")
      "token"
    }
  )
  auth <- databricks_auth_args("host")
  expect_equal(auth$auth_accesstoken, "token")
})

test_that("workload identity env-oidc falls back when OIDC_TOKEN_ENV is empty", {
  withr::local_envvar(
    DATABRICKS_TOKEN = "",
    DATABRICKS_CLIENT_ID = "sp",
    DATABRICKS_AUTH_TYPE = "env-oidc",
    DATABRICKS_OIDC_TOKEN_ENV = "",
    DATABRICKS_OIDC_TOKEN = "fallback-jwt"
  )
  local_mocked_bindings(
    databricks_token_exchange = function(host, id_token, client_id) {
      expect_equal(id_token, "fallback-jwt")
      "token"
    }
  )
  auth <- databricks_auth_args("host")
  expect_equal(auth$auth_accesstoken, "token")
})

test_that("supports workload identity file-oidc", {
  token_file <- withr::local_tempfile()
  writeLines("file-jwt", token_file)
  withr::local_envvar(
    DATABRICKS_TOKEN = "",
    DATABRICKS_CLIENT_ID = "sp",
    DATABRICKS_AUTH_TYPE = "file-oidc",
    DATABRICKS_OIDC_TOKEN = "",
    DATABRICKS_OIDC_TOKEN_ENV = "",
    DATABRICKS_OIDC_TOKEN_FILEPATH = token_file
  )
  local_mocked_bindings(
    databricks_token_exchange = function(host, id_token, client_id) {
      expect_equal(id_token, "file-jwt")
      "exchanged-token"
    }
  )
  auth <- databricks_auth_args("host")
  expect_equal(auth$auth_accesstoken, "exchanged-token")
})

test_that("workload identity errors when DATABRICKS_CLIENT_ID is not set", {
  withr::local_envvar(
    DATABRICKS_TOKEN = "",
    DATABRICKS_CLIENT_ID = "",
    DATABRICKS_AUTH_TYPE = "env-oidc",
    DATABRICKS_OIDC_TOKEN = "jwt"
  )
  expect_snapshot(databricks_auth_args("host"), error = TRUE)
})

test_that("workload identity env-oidc errors when JWT env var is empty", {
  withr::local_envvar(
    DATABRICKS_TOKEN = "",
    DATABRICKS_CLIENT_ID = "sp",
    DATABRICKS_AUTH_TYPE = "env-oidc",
    DATABRICKS_OIDC_TOKEN = "",
    DATABRICKS_OIDC_TOKEN_ENV = ""
  )
  expect_snapshot(databricks_auth_args("host"), error = TRUE)
})

test_that("workload identity file-oidc errors when filepath is not set", {
  withr::local_envvar(
    DATABRICKS_TOKEN = "",
    DATABRICKS_CLIENT_ID = "sp",
    DATABRICKS_AUTH_TYPE = "file-oidc",
    DATABRICKS_OIDC_TOKEN = "",
    DATABRICKS_OIDC_TOKEN_ENV = "",
    DATABRICKS_OIDC_TOKEN_FILEPATH = ""
  )
  expect_snapshot(databricks_auth_args("host"), error = TRUE)
})

test_that("workload identity file-oidc errors when file does not exist", {
  withr::local_envvar(
    DATABRICKS_TOKEN = "",
    DATABRICKS_CLIENT_ID = "sp",
    DATABRICKS_AUTH_TYPE = "file-oidc",
    DATABRICKS_OIDC_TOKEN = "",
    DATABRICKS_OIDC_TOKEN_ENV = "",
    DATABRICKS_OIDC_TOKEN_FILEPATH = "/nonexistent/path/to/jwt"
  )
  expect_snapshot(databricks_auth_args("host"), error = TRUE)
})

test_that("workload identity file-oidc errors when file is empty", {
  token_file <- withr::local_tempfile()
  writeLines("", token_file)
  withr::local_envvar(
    DATABRICKS_TOKEN = "",
    DATABRICKS_CLIENT_ID = "sp",
    DATABRICKS_AUTH_TYPE = "file-oidc",
    DATABRICKS_OIDC_TOKEN = "",
    DATABRICKS_OIDC_TOKEN_ENV = "",
    DATABRICKS_OIDC_TOKEN_FILEPATH = token_file
  )
  expect_snapshot(
    databricks_auth_args("host"),
    error = TRUE,
    transform = function(x) gsub(token_file, "<tempfile>", x, fixed = TRUE)
  )
})

test_that("DATABRICKS_TOKEN takes precedence over workload identity", {
  withr::local_envvar(
    DATABRICKS_TOKEN = "my-pat",
    DATABRICKS_CLIENT_ID = "sp",
    DATABRICKS_AUTH_TYPE = "env-oidc",
    DATABRICKS_OIDC_TOKEN = "my-jwt",
    DATABRICKS_OIDC_TOKEN_ENV = ""
  )
  auth <- databricks_auth_args("host")
  expect_equal(auth$pwd, "my-pat")
  expect_equal(auth$authMech, 3)
})

test_that("databricks_token_exchange extracts access_token from response", {
  skip_if_not_installed("httr2")
  httr2::local_mocked_responses(function(req) {
    httr2::response(
      status_code = 200L,
      headers = list(`content-type` = "application/json"),
      body = charToRaw(
        '{"access_token":"tok123","token_type":"Bearer","expires_in":3600}'
      )
    )
  })
  result <- databricks_token_exchange("host.example.com", "jwt", "sp-id")
  expect_equal(result, "tok123")
})

test_that("databricks_token_exchange errors on HTTP failure", {
  skip_if_not_installed("httr2")
  httr2::local_mocked_responses(function(req) {
    httr2::response(
      status_code = 400L,
      headers = list(`content-type` = "application/json"),
      body = charToRaw('{"error":"invalid_grant"}')
    )
  })
  expect_snapshot(
    databricks_token_exchange("host.example.com", "bad-jwt", "sp-id"),
    error = TRUE
  )
})

test_that("databricks_token_exchange errors when access_token is missing", {
  skip_if_not_installed("httr2")
  httr2::local_mocked_responses(function(req) {
    httr2::response(
      status_code = 200L,
      headers = list(`content-type` = "application/json"),
      body = charToRaw('{"token_type":"Bearer"}')
    )
  })
  expect_snapshot(
    databricks_token_exchange("host.example.com", "jwt", "sp-id"),
    error = TRUE
  )
})

test_that("DATABRICKS_CLIENT_ID still triggers M2M when AUTH_TYPE is not workload identity", {
  withr::local_envvar(
    DATABRICKS_TOKEN = "",
    DATABRICKS_CLIENT_ID = "abc",
    DATABRICKS_CLIENT_SECRET = "def",
    DATABRICKS_AUTH_TYPE = "",
    DATABRICKS_OIDC_TOKEN = "",
    DATABRICKS_OIDC_TOKEN_ENV = ""
  )
  auth <- databricks_auth_args("host")
  expect_equal(auth$auth_client_id, "abc")
  expect_equal(auth$auth_client_secret, "def")
  expect_equal(auth$auth_flow, 1)
})

# Microsoft SQL Server ---------------------------------------------------------

#' SQL Server
#'
#' Details of SQL Server methods for odbc and DBI generics.
#'
#' Note on binding \href{https://learn.microsoft.com/en-us/sql/relational-databases/tables/use-table-valued-parameters-database-engine}{TVPs}:
#' You can bind `data.frame`s to SQL Server TVPs when executing stored procedures with the following caveats:
#' * All non-df parameters must be of length 1; and
#' * The `batch_rows` parameter (to `dbBind`, for example) should be set to 1.
#' @examples
#' \dontrun{
#' # Bind `data.frame` to a TVP when executing a
#' # stored procedure that takes three parameters:
#' # a TVP, an INT and a VARCHAR(max).
#' res <- dbGetQuery(conn, "{ CALL example_sproc(?, ?, ?) }",
#'   params = list(df.data, 100, "Lorem ipsum dolor sit amet"))
#' }
#' @rdname SQLServer
#' @export
setClass("Microsoft SQL Server", contains = "OdbcConnection")

#' @description
#' ## `dbUnquoteIdentifier()`
#'
#' `conn@quote` returns the quotation mark, but quotation marks and square
#' brackets can be used interchangeably for delimited identifiers.
#' (<https://learn.microsoft.com/en-us/sql/relational-databases/databases/database-identifiers>).
#' This function strips the brackets first and then calls the DBI method to
#' strip the quotation marks.
#' @rdname SQLServer
#' @docType methods
#' @usage NULL
#' @keywords internal
setMethod("dbUnquoteIdentifier", c("Microsoft SQL Server", "SQL"),
  function(conn, x, ...) {
    x <- gsub("(\\[)([^\\.]+?)(\\])", "\\2", x)
    callNextMethod(conn, x, ...)
  }
)


#' @description
#' ## `isTempTable()`
#'
#' Local temp tables are stored as
#' `[tempdb].[dbo].[#name]_____[numeric identifier]`, so `isTempTable()`
#' returns `TRUE` if `catalog_name` is `"tempdb"` or `"%"`, or the
#' name starts with `"#"`.
#' @rdname SQLServer
#' @usage NULL
setMethod("isTempTable", c("Microsoft SQL Server", "character"),
  function(conn, name, catalog_name = NULL, schema_name = NULL, ...) {
    if (!is.null(catalog_name) &&
      catalog_name != "%" &&
      length(catalog_name) > 0 &&
      catalog_name != "tempdb") {
      return(FALSE)
    }

    grepl("^[#][^#]", name)
  }
)

#' @rdname SQLServer
#' @usage NULL
setMethod("isTempTable", c("Microsoft SQL Server", "SQL"),
  function(conn, name, ...) {
    isTempTable(conn, dbUnquoteIdentifier(conn, name)[[1]], ...)
  }
)


#' @description
#' ## `dbExistsTable()`
#' The default implementation reports temporary tables as non-existent
#' since they live in a different catalog. This method provides a special
#' case for temporary tables, as identified by `isTempTable()`.
#' @rdname SQLServer
#' @docType methods
#' @usage NULL
setMethod("dbExistsTable", c("Microsoft SQL Server", "character"),
  function(conn, name, ...) {
    check_string(name)
    if (isTempTable(conn, name, ...)) {
      query <- paste0("SELECT OBJECT_ID('tempdb..", name, "')")
      !is.na(dbGetQuery(conn, query)[[1]])
    } else {
      df <- odbcConnectionTables(conn, name = name, ...)
      NROW(df) > 0
    }
  }
)

#' @description
#' ## `dbListTables()`
#' The default implementation reports temporary tables as non-existent
#' when a `catalog_name` isn't supplied since they live in a different catalog.
#' This method provides a special case for temporary tables.
#' @rdname SQLServer
#' @usage NULL
setMethod("dbListTables", "Microsoft SQL Server",
  function(conn,
           catalog_name = NULL,
           schema_name = NULL,
           table_name = NULL,
           table_type = NULL,
           ...) {
    check_string(catalog_name, allow_null = TRUE)
    check_string(schema_name, allow_null = TRUE)
    check_string(table_name, allow_null = TRUE)
    check_string(table_type, allow_null = TRUE)

    res <- callNextMethod()

    if (is.null(catalog_name) && is.null(schema_name)) {
      res_temp <- callNextMethod(
        conn = conn,
        catalog_name = "tempdb",
        schema_name = "dbo"
      )

      res <- c(res, res_temp)
    }

    res
  }
)

#' @rdname SQLServer
#' @usage NULL
setMethod("dbExistsTable", c("Microsoft SQL Server", "Id"),
  function(conn, name, ...) {
    dbExistsTable(
      conn,
      name = id_field(name, "table"),
      catalog_name = id_field(name, "catalog"),
      schema_name = id_field(name, "schema")
    )
  }
)

#' @rdname SQLServer
#' @usage NULL
setMethod("dbExistsTable", c("Microsoft SQL Server", "SQL"),
  function(conn, name, ...) {
    dbExistsTable(conn, dbUnquoteIdentifier(conn, name)[[1]], ...)
  }
)

#' @description
#' ## `odbcConnectionSchemas`
#'
#' Method for an internal function. Calls catalog-specific `sp_tables` to make
#' sure we list the schemas in the appropriate database/catalog.
#' @rdname SQLServer
#' @usage NULL
setMethod(
  "odbcConnectionSchemas", "Microsoft SQL Server",
  function(conn, catalog_name = NULL) {

    if (is.null(catalog_name) || !nchar(catalog_name)) {
      return(callNextMethod())
    }
    sproc <- paste(
      dbQuoteIdentifier(conn, catalog_name), "dbo.sp_tables", sep = ".")

    res <- dbGetQuery(conn, paste0(
        "EXEC ", sproc, " ",
        "@table_name = '', ",
        "@table_owner = '%', ",
        "@table_qualifier = ''"
    ))
    res$TABLE_OWNER
  }
)

#' @rdname SQLServer
#' @description
#' ## `sqlCreateTable()`
#'
#' Warns if `temporary = TRUE` but the `name` does not conform to temp table
#' naming conventions (i.e. it doesn't start with `#`).
#' @usage NULL
setMethod("sqlCreateTable", "Microsoft SQL Server",
  function(con,
           table,
           fields,
           row.names = NA,
           temporary = FALSE,
           ...,
           field.types = NULL) {
    check_bool(temporary)
    check_row.names(row.names)
    check_field.types(field.types)
    if (temporary && !isTempTable(con, table)) {
      cli::cli_warn(
        "{.arg temporary} is {.code TRUE}, but table name doesn't use # prefix."
      )
    }
    temporary <- FALSE
    callNextMethod()
  }
)

#' @export
#' @rdname odbcDataType
#' @usage NULL
setMethod("odbcDataType", "Microsoft SQL Server",
  function(con, obj, ...) {
    switch_type(
      obj,
      factor = varchar(obj),
      datetime = "DATETIME",
      date = "DATE",
      time = "TIME",
      binary = varbinary(obj),
      integer = "INT",
      int64 = "BIGINT",
      double = "FLOAT",
      character = varchar(obj),
      logical = "BIT",
      list = varchar(obj),
      stop("Unsupported type", call. = FALSE)
    )
  }
)

#' @description
#' ## `odbcConnectionColumns()`
#'
#' If temp table, query the database for the
#' actual table name.
#' @rdname SQLServer
#' @usage NULL
setMethod("odbcConnectionColumns", c("Microsoft SQL Server", "character"),
  function(conn,
           name,
           ...,
           catalog_name = NULL,
           schema_name = NULL,
           column_name = NULL,
           exact = FALSE) {
    if (exact &&
      isTempTable(conn, name, catalog_name, schema_name, column_name, exact)) {
      catalog_name <- "tempdb"
      schema_name <- "dbo"
      query <- paste0("SELECT name FROM tempdb.sys.tables WHERE ",
        "object_id = OBJECT_ID('tempdb..", name, "')")
      name <- dbGetQuery(conn, query)[[1]]
    }

    callNextMethod(
      conn = conn,
      name = name,
      ...,
      catalog_name = catalog_name,
      schema_name = schema_name,
      column_name = column_name,
      exact = exact
    )
  }
)

#' @description
#' ## `odbcConnectionColumns()`
#'
#' Copied over from odbc-connection to avoid S4 dispatch NOTEs.
#' @rdname SQLServer
#' @usage NULL
setMethod("odbcConnectionColumns", c("Microsoft SQL Server", "SQL"),
  function(conn, name, ..., exact = FALSE) {
    odbcConnectionColumns(conn, dbUnquoteIdentifier(conn, name)[[1]], ..., exact = exact)
  }
)

#' @include dbi-connection.R
NULL

#' @export
#' @rdname DBI-classes
setClass("BigQuery", contains = "OdbcConnection")

#' @export
#' @rdname odbcDataType
#' @usage NULL
setMethod("odbcDataType", "BigQuery",
  function(con, obj, ...) {
    switch_type(
      obj,
      factor = "STRING",
      datetime = "TIMESTAMP",
      time = "TIME",
      date = "DATE",
      binary = "BYTES",
      integer = "INT64",
      int64 = "INT64",
      double = "FLOAT64",
      character = "STRING",
      logical = "BOOL",
      stop("Unsupported type", call. = FALSE)
    )
  }
)

#' @include dbi-connection.R
NULL

#' Odbc Result Methods
#'
#' Implementations of pure virtual functions defined in the `DBI` package
#' for OdbcResult objects.
#' @name OdbcResult
#' @docType methods
NULL

OdbcResult <- function(connection, statement, params = NULL, immediate = FALSE) {
  if (nzchar(connection@encoding)) {
    statement <- enc2iconv(statement, connection@encoding)
  }
  ptr <- new_result(
    p = connection@ptr,
    sql = statement, immediate = immediate
  )
  res <- new(
    "OdbcResult",
    connection = connection,
    statement = statement,
    ptr = ptr
  )

  if (!is.null(params)) {
    on.exit(dbClearResult(res))
    dbBind(res, params = params)
    on.exit(NULL)
  }
  res
}

#' @rdname OdbcResult
#' @keywords internal
#' @export
setClass(
  "OdbcResult",
  contains = "DBIResult",
  slots = list(
    connection = "OdbcConnection",
    statement = "character",
    ptr = "externalptr"
  )
)

#' @rdname OdbcResult
#' @param res An object inheriting from [DBI::DBIResult-class].
#' @inheritParams DBI::dbClearResult
#' @export
setMethod("dbClearResult", "OdbcResult",
  function(res, ...) {
    if (!dbIsValid(res)) {
      warning("Result already cleared")
    }
    result_release(res@ptr)
    invisible(TRUE)
  }
)

#' @rdname OdbcResult
#' @param res An object inheriting from [DBI::DBIResult-class].
#' @inheritParams DBI::dbFetch
#' @export
setMethod("dbFetch", "OdbcResult",
  function(res, n = -1, ...) {
    check_number_whole(n, min = -1, allow_infinite = TRUE)
    if (is.infinite(n)) n <- -1
    result_fetch(res@ptr, n)
  }
)

#' @rdname OdbcResult
#' @param res An object inheriting from [DBI::DBIResult-class].
#' @inheritParams DBI::dbHasCompleted
#' @export
setMethod("dbHasCompleted", "OdbcResult",
  function(res, ...) {
    result_completed(res@ptr)
  }
)

#' @rdname OdbcResult
#' @param dbObj An object inheriting from `DBIObject`, i.e. `DBIDriver`,
#' `DBIConnection`, or a `DBIResult`.
#' @inheritParams DBI::dbIsValid
#' @export
setMethod("dbIsValid", "OdbcResult",
  function(dbObj, ...) {
    result_active(dbObj@ptr)
  }
)

#' @rdname OdbcResult
#' @param res An object inheriting from [DBI::DBIResult-class].
#' @inheritParams DBI::dbGetStatement
#' @export
setMethod("dbGetStatement", "OdbcResult",
  function(res, ...) {
    if (!dbIsValid(res)) {
      cli::cli_abort("Result already cleared.")
    }
    res@statement
  }
)

#' @rdname OdbcResult
#' @param res An object inheriting from [DBI::DBIResult-class].
#' @inheritParams DBI::dbColumnInfo
#' @export
setMethod("dbColumnInfo", "OdbcResult",
  function(res, ...) {
    result_column_info(res@ptr)
  }
)

#' @rdname OdbcResult
#' @param res An object inheriting from [DBI::DBIResult-class].
#' @inheritParams DBI::dbGetRowCount
#' @export
setMethod("dbGetRowCount", "OdbcResult",
  function(res, ...) {
    result_row_count(res@ptr)
  }
)

#' @rdname OdbcResult
#' @param res An object inheriting from [DBI::DBIResult-class].
#' @inheritParams DBI::dbGetRowsAffected
#' @export
setMethod("dbGetRowsAffected", "OdbcResult",
  function(res, ...) {
    result_rows_affected(res@ptr)
  }
)

#' @rdname OdbcResult
#' @param res An object inheriting from [DBI::DBIResult-class].
#' @inheritParams DBI::dbBind
#' @inheritParams DBI-tables
#' @export
setMethod("dbBind", "OdbcResult",
  function(res, params, ..., batch_rows = getOption("odbc.batch_rows", NA)) {
    params <- as.list(params)
    if (length(params) == 0) {
      return(invisible(res))
    }

    paramDfs <- sapply(params, is.data.frame)
    if (any(paramDfs)) {
      # When binding table-valued-parameters verify that:
      # * The `batch_rows` parameter is either unset or equal to one.
      # * All non-tvp parameters are of length 1.
      if (!is.na(batch_rows) && batch_rows > 1) {
        cli::cli_warn(c("Since some parameters are data frames, {.arg batch_rows} will be overwritten to `1`.", 
                       "Set {.code batch_rows = 1} to quiet this warning."))
      }
      batch_rows <- 1
      paramLengths <- sapply(params[!paramDfs], length)
      if (any(paramLengths > 1))
      {
        cli::cli_abort("When mixing data.frame(s) with other parameter types,
                       all non-df parameters must be of length one")
      }
    } else {
      params[!paramDfs] <- vctrs::vec_recycle_common(!!!params[!paramDfs], .arg = "params")
      if (is.na(batch_rows)) {
        batch_rows <- length(params[[1]])
      }
    }

    batch_rows <- parse_size(batch_rows)

    result_bind(res@ptr, params, batch_rows)
    invisible(res)
  }
)

#' @include dbi-connection.R
NULL

#' @export
#' @rdname DBI-classes
setClass("Snowflake", contains = "OdbcConnection")

#' getCatalogSchema
#'
#' Internal, Snowflake specific method that will return a list
#' where `catalog_name` and `schema_name` are either the arguments
#' passed by the user ( if not null ), or if null, the CURRENT
#' database ( or schema, respectively ).
#' @noRd
getCatalogSchema <- function(conn, catalog_name = NULL, schema_name = NULL) {
  if(is.null(catalog_name) || is.null(schema_name)) {
    res <- dbGetQuery(conn, "SELECT CURRENT_DATABASE() AS CAT, CURRENT_SCHEMA() AS SCH")
    if(is.null(catalog_name) && !is.na(res$CAT)) {
      catalog_name <- res$CAT
    }
    if(is.null(schema_name) && !is.na(res$SCH)) {
      schema_name <- res$SCH
    }
  }
  return(list(catalog_name = catalog_name, schema_name = schema_name))
}

#' Connecting to Snowflake via ODBC
#'
#' @description
#' ## `odbcConnectionColumns()`
#'
#' If the catalog, or the schema arguments are NULL, attempt to infer
#' by querying for CURRENT_DATABASE() and CURRENT_SCHEMA().  We do this
#' to aid with performance, as the SQLColumns method is more performant
#' when restricted to a particular DB/schema.
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#' `dbConnect()`.
#' @inheritParams DBI::dbListFields
#' @param catalog_name,schema_name Catalog and schema names.
#' @rdname driver-Snowflake
#' @usage NULL
setMethod("odbcConnectionColumns", c("Snowflake", "character"),
  function(conn,
           name,
           ...,
           catalog_name = NULL,
           schema_name = NULL,
           column_name = NULL,
           exact = FALSE) {
    catSchema <- getCatalogSchema(conn, catalog_name, schema_name)

    callNextMethod(conn = conn, name = name, ..., catalog_name = catSchema$catalog_name,
      schema_name = catSchema$schema_name, column_name = column_name, exact = exact)
  }
)

#' @rdname driver-Snowflake
setMethod("dbExistsTableForWrite", c("Snowflake", "character"),
  function(conn, name, ...,
           catalog_name = NULL, schema_name = NULL) {
    catSchema <- getCatalogSchema(conn, catalog_name, schema_name)
    catalog_name <- catSchema$catalog_name
    schema_name <- catSchema$schema_name

    callNextMethod(conn = conn, name = name, ...,
      catalog_name = catalog_name, schema_name = schema_name)
  }
)

#' @export
#' @rdname odbcDataType
#' @usage NULL
setMethod("odbcDataType", "Snowflake",
  function(con, obj, ...) {
    switch_type(
      obj,
      factor = "VARCHAR",
      datetime = "TIMESTAMP",
      date = "DATE",
      time = "TIME",
      binary = "BINARY",
      integer = "INTEGER",
      int64 = "INTEGER",
      double = "FLOAT",
      character = "VARCHAR",
      logical = "BOOLEAN",
      list = "VARCHAR",
      stop("Unsupported type", call. = FALSE)
    )
  }
)

#' Helper for connecting to Snowflake via ODBC
#'
#' @description
#'
#' Connect to a Snowflake account via the [Snowflake ODBC
#' driver](https://docs.snowflake.com/en/developer-guide/odbc/odbc).
#'
#' In particular, the custom `dbConnect()` method for the Snowflake ODBC driver
#' supports the `connections.toml` and `config.toml` files used by the Snowflake
#' Connector for Python and the Snowflake CLI via the \pkg{snowflakeauth}
#' package. It also detects ambient OAuth credentials on platforms like Snowpark
#' Container Services or Posit Workbench. Finally, it can detect viewer-based
#' credentials on Posit Connect if the \pkg{connectcreds} package is installed.
#'
#' In addition, on macOS platforms, the `dbConnect` method will check and warn
#' if it detects irregularities with how the driver is configured, unless the
#' `odbc.no_config_override` environment variable is set.
#'
#' @param drv an object that inherits from [DBI::DBIDriver-class],
#' or an existing [DBI::DBIConnection-class]
#' object (in order to clone an existing connection).
#' @param account A Snowflake [account
#'   identifier](https://docs.snowflake.com/en/user-guide/admin-account-identifier),
#'   e.g. `"testorg-test_account"`.
#' @param driver The name of the Snowflake ODBC driver, or `NULL` to use the
#'   default name.
#' @param warehouse The name of a Snowflake compute warehouse, or `NULL` to use
#'   the default.
#' @param database The name of a Snowflake database, or `NULL` to use the
#'   default.
#' @param schema The name of a Snowflake database schema, or `NULL` to use the
#'   default.
#' @param uid,pwd Manually specify a username and password for authentication.
#'   Specifying these options will disable ambient credential discovery.
#' @param connection_name The name of a connection defined in your Snowflake
#'   `connections.toml` file, or `NULL` to use the default connection.
#' @param ... Further arguments passed on to [`dbConnect()`].
#'
#' @returns An `OdbcConnection` object with an active connection to a Snowflake
#'   account.
#'
#' @examples
#' \dontrun{
#' # Use ambient credentials.
#' DBI::dbConnect(odbc::snowflake())
#'
#' # Use browser-based SSO (if configured). Only works on desktop.
#' DBI::dbConnect(
#'   odbc::snowflake(),
#'   account = "testorg-test_account",
#'   authenticator = "externalbrowser"
#' )
#'
#' # Use a traditional username & password.
#' DBI::dbConnect(
#'   odbc::snowflake(),
#'   account = "testorg-test_account",
#'   uid = "me",
#'   pwd = rstudioapi::askForPassword()
#' )
#'
#' # Use a named connection from a connections.toml file.
#' DBI::dbConnect(odbc::snowflake(), connection_name = "test")
#'
#' # Use credentials from the viewer (when possible) in a Shiny app
#' # deployed to Posit Connect.
#' library(connectcreds)
#' server <- function(input, output, session) {
#'   conn <- DBI::dbConnect(odbc::snowflake())
#' }
#' }
#' @export
snowflake <- function() {
  new("SnowflakeOdbcDriver")
}

#' @rdname snowflake
#' @export
setClass("SnowflakeOdbcDriver", contains = "OdbcDriver")

#' @rdname snowflake
#' @export
setMethod(
  "dbConnect", "SnowflakeOdbcDriver",
  function(drv,
           account = Sys.getenv("SNOWFLAKE_ACCOUNT"),
           driver = NULL,
           warehouse = NULL,
           database = NULL,
           schema = NULL,
           uid = NULL,
           pwd = NULL,
           connection_name = NULL,
           ...) {
    call <- caller_env()
    check_string(account, call = call)
    check_string(driver, allow_null = TRUE, call = call)
    check_string(warehouse, allow_null = TRUE, call = call)
    check_string(database, allow_null = TRUE, call = call)
    check_string(uid, allow_null = TRUE, call = call)
    check_string(pwd, allow_null = TRUE, call = call)
    check_string(connection_name, allow_null = TRUE, call = call)
    args <- snowflake_args(
      account = account,
      driver = driver,
      warehouse = warehouse,
      database = database,
      schema = schema,
      uid = uid,
      pwd = pwd,
      connection_name = connection_name,
      ...
    )
    # Perform some sanity checks on MacOS
    configure_simba(snowflake_simba_config(args$driver),
      action = "modify")
    inject(dbConnect(odbc(), !!!args))
  }
)

snowflake_args <- function(account = Sys.getenv("SNOWFLAKE_ACCOUNT"),
                           driver = NULL,
                           ...) {
  auth <- snowflake_auth_args(account, ...)
  host <- auth$.host
  auth$.host <- NULL
  if (!is.null(auth$.account)) {
    account <- auth$.account
  }
  auth$.account <- NULL

  args <- list(
    driver = driver %||% snowflake_default_driver(),
    account = account,
    server = snowflake_server(account, host = host),
    # Connections to Snowflake are always over HTTPS.
    port = 443
  )
  dots <- list(...)
  dots$connection_name <- NULL
  dots <- compact(dots)
  all <- utils::modifyList(c(args, auth), dots)

  # Set application value and respect the Snowflake Partner environment variable, if present.
  if (is.null(all$application)) {
    if (nchar(Sys.getenv("SF_PARTNER")) != 0 ){
      all$application <- Sys.getenv("SF_PARTNER")
    } else {
      all$application <- "r-odbc"
    }
  }

  arg_names <- tolower(names(all))
  if (!"authenticator" %in% arg_names && !all(c("uid", "pwd") %in% arg_names)) {
    msg <- c(
      "Failed to detect ambient Snowflake credentials.",
      "i" = "Supply {.arg uid} and {.arg pwd} to authenticate manually."
    )
    if (running_on_connect()) {
      msg <- c(
        msg,
        "i" = "Or consider enabling Posit Connect's Snowflake integration \
              for viewer-based credentials. See {.url \
              https://docs.posit.co/connect/user/oauth-integrations/#adding-oauth-integrations-to-deployed-content}
              for details."
      )
    }
    cli::cli_abort(msg, call = quote(DBI::dbConnect()))
  }

  all
}

# Returns a sensible driver name even if odbc.ini and odbcinst.ini do not
# contain an entry for the Snowflake ODBC driver. For Linux and macOS we
# default to known shared library paths used by the official installers.
# On Windows we use the official driver name.
snowflake_default_driver <- function() {
  find_default_driver(
    snowflake_default_driver_paths(),
    fallbacks = c("Snowflake", "SnowflakeDSIIDriver"),
    label = "Snowflake",
    call = quote(DBI::dbConnect())
  )
}

snowflake_default_driver_paths <- function() {
  if (Sys.info()["sysname"] == "Linux") {
    paths <- c(
      "/opt/rstudio-drivers/snowflake/bin/lib/libsnowflakeodbc_sb64.so",
      "/usr/lib/snowflake/odbc/lib/libSnowflake.so"
    )
  } else if (Sys.info()["sysname"] == "Darwin") {
    paths <- c(
      "/opt/snowflake/snowflakeodbc/lib/universal/libSnowflake.dylib",
      "/opt/snowflake-osx-x64/bin/lib/libsnowflakeodbc_sb64-universal.dylib"
    )
  } else {
    paths <- character()
  }
  paths
}

snowflake_simba_config <- function(driver) {
  snowflake_env <- Sys.getenv("SIMBASNOWFLAKEINI")
  URL <- "https://docs.snowflake.com/en/developer-guide/odbc/odbc-download"
  if (!identical(snowflake_env, "")) {
    return(list(path = snowflake_env, url = URL))
  }
  # Posit configuration is likely at:
  # /opt/snowflake-osx-x64/bin/lib/rstudio.snowflakeodbc.ini
  # OEM configuration is likely at:
  # /opt/snowflake/snowflakeodbc/lib/universal/simba.snowflake.ini
  common_dirs <- c(
    driver_dir(driver),
    "/opt/snowflake-osx-x64/bin/lib/",
    "/opt/snowflake/snowflakeodbc/lib/universal/",
    getwd(),
    Sys.getenv("HOME")
  )
  return(list(
    path = list.files(
      common_dirs,
      pattern = "snowflake(odbc)?\\.ini$",
      full.names = TRUE),
    url = URL
  ))
}

snowflake_server <- function(account, host = NULL) {
  if (!is.null(host) && nzchar(host)) {
    return(host)
  }
  if (nchar(account) == 0) {
    abort(
      c(
        "No Snowflake account ID provided.",
        i = "Either supply `account` argument or set env var `SNOWFLAKE_ACCOUNT`."
      ),
      call = quote(DBI::dbConnect())
    )
  }
  paste0(account, ".snowflakecomputing.com")
}

snowflake_auth_args <- function(account,
                                uid = NULL,
                                pwd = NULL,
                                authenticator = NULL,
                                connection_name = NULL,
                                ...) {
  check_string(authenticator, allow_null = TRUE)
  # Detect viewer-based credentials from Posit Connect.
  url <- paste0("https://", account, ".snowflakecomputing.com")
  if (is_installed("connectcreds") && connectcreds::has_viewer_token(url)) {
    token <- connectcreds::connect_viewer_token(url)
    return(list(authenticator = "oauth", token = token$access_token))
  }

  if (!is.null(uid) &&
      # allow for uid without pwd for alt auth (#817, #889)
      (!is.null(pwd) ||
       isTRUE(authenticator %in% c("externalbrowser", "SNOWFLAKE_JWT")))) {
    return(list(uid = uid, pwd = pwd))
  } else if (xor(is.null(uid), is.null(pwd))) {
    abort(
      c(
        "Both `uid` and `pwd` must be specified to authenticate.",
        i = "Or leave both unset to use ambient Snowflake credentials."
      ),
      call = quote(DBI::dbConnect())
    )
  }

  # Manual override of the authentication method.
  if (!is.null(authenticator)) {
    return(list())
  }

  check_installed("snowflakeauth")
  sf_args <- list(name = connection_name)
  dots <- list(...)
  sf_params <- c(
    "user", "role", "schema", "database", "warehouse", "authenticator",
    "private_key", "private_key_file", "private_key_path",
    "private_key_file_pwd", "token", "token_file_path", "password", "host"
  )
  dots <- compact(dots)
  sf_args <- c(sf_args, dots[intersect(names(dots), sf_params)])
  if (file.exists("/snowflake/session/token") &&
      is.null(sf_args$token_file_path) &&
      is.null(sf_args$token)) {
    sf_args$token_file_path <- "/snowflake/session/token"
  }
  # Don't pass `account` so we can compare it to what snowflakeauth resolves.
  conn <- inject(snowflakeauth::snowflake_connection(!!!sf_args))
  if (is.null(connection_name) &&
      nzchar(account) && !is.null(conn$account) &&
      !identical(account, conn$account)) {
    cli::cli_abort(
      c(
        "Snowflake account mismatch: requested {.val {account}} but
         connection {.val {conn$name %||% 'default'}} resolved
         {.val {conn$account}}.",
        "i" = "Check your {.file connections.toml} or pass the correct
              {.arg account}."
      ),
      call = quote(DBI::dbConnect())
    )
  }
  snowflake_connection_to_odbc_args(conn)
}

snowflake_connection_to_odbc_args <- function(conn) {
  args <- list()

  if (!is.null(conn$user)) {
    args$uid <- conn$user
  }
  if (!is.null(conn$password)) {
    args$pwd <- unclass(conn$password)
  }
  if (!is.null(conn$token)) {
    args$token <- unclass(conn$token)
  }
  if (!is.null(conn$token_file_path)) {
    args$token_file_path <- conn$token_file_path
  }
  if (!is.null(conn$authenticator) && conn$authenticator != "snowflake") {
    args$authenticator <- conn$authenticator
  }

  priv_key <- conn$private_key_file %||%
    conn$private_key %||%
    conn$private_key_path
  if (!is.null(priv_key)) {
    args$priv_key_file <- priv_key
  }
  if (!is.null(conn$private_key_file_pwd)) {
    args$priv_key_file_pwd <- unclass(conn$private_key_file_pwd)
  }

  if (!is.null(conn$warehouse)) args$warehouse <- conn$warehouse
  if (!is.null(conn$database)) args$database <- conn$database
  if (!is.null(conn$schema)) args$schema <- conn$schema
  if (!is.null(conn$role)) args$role <- conn$role
  if (!is.null(conn$workload_identity_provider)) {
    args$workload_identity_provider <- conn$workload_identity_provider
  }

  # Pass non-ODBC fields for snowflake_args() to use.
  if (!is.null(conn$account) && nzchar(conn$account)) {
    args$.account <- conn$account
  }
  if (!is.null(conn$host) && nzchar(conn$host)) {
    args$.host <- conn$host
  }

  args
}

test_that("can connect to snowflake", {
  key_exists <- Sys.getenv("SNOWFLAKE_PRIVATE_KEY_EXISTS")
  if (nchar(key_exists) == 0) {
    skip("Secret SNOWFLAKE_PRIVATE_KEY not available.")
  }

  con <- test_con("SNOWFLAKE")
})

test_that("an account ID is required", {
  sf_home <- tempfile()
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "",
    SNOWFLAKE_HOME = sf_home
  )
  expect_snapshot(
    snowflake_args(driver = "driver"),
    error = TRUE,
    transform = function(x) gsub(sf_home, "<tmp>", x, fixed = TRUE)
  )
})

test_that("environment variables are handled correctly", {
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "testorg-test_account",
    SF_PARTNER = "test_partner"
  )
  args <- snowflake_args(
    pwd = "pwd",
    uid = "user",
    driver = "driver"
  )
  expect_equal(args$account, "testorg-test_account")
  expect_equal(args$server, "testorg-test_account.snowflakecomputing.com")
  expect_equal(args$application, "test_partner")
})

test_that("environment variables can be overridden with parameters", {
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "account",
    SF_PARTNER = "test_partner"
  )
  args <- snowflake_args(
    account = "testorg-test_account",
    pwd = "pwd",
    uid = "user",
    driver = "driver",
    application = "myapp"
  )
  expect_equal(args$account, "testorg-test_account")
  expect_equal(args$server, "testorg-test_account.snowflakecomputing.com")
  expect_equal(args$application, "myapp")
})

test_that("the 'uid' and 'pwd' arguments suppress automated auth", {
  args <- snowflake_args(
    account = "testorg-test_account",
    driver = "driver",
    uid = "uid",
    pwd = "pwd"
  )
  expect_equal(args$uid, "uid")
  expect_equal(args$pwd, "pwd")
  expect_equal(args$authenticator, NULL)
})

test_that("both 'uid' and 'pwd' are required when present", {
  expect_snapshot(
    snowflake_args(
      account = "testorg-test_account",
      driver = "driver",
      uid = "uid",
    ),
    error = TRUE
  )
})

test_that("alternative authenticators are supported", {
  args <- snowflake_args(
    account = "testorg-test_account",
    driver = "driver",
    authenticator = "externalbrowser"
  )
  expect_equal(args$authenticator, "externalbrowser")
})

test_that("can pass only `uid` with non-NULL `authenticator` (#817, #889)", {
  args <- snowflake_args(
    account = "testorg-test_account",
    driver = "driver",
    authenticator = "externalbrowser",
    uid = "boop"
  )
  expect_equal(args$uid, "boop")
  expect_equal(args$authenticator, "externalbrowser")

  args <- snowflake_args(
    account = "testorg-test_account",
    driver = "driver",
    authenticator = "SNOWFLAKE_JWT",
    uid = "boop"
  )
  expect_equal(args$uid, "boop")
  expect_equal(args$authenticator, "SNOWFLAKE_JWT")
})

test_that("we error if we can't find ambient credentials", {
  withr::local_envvar(SF_PARTNER = "")
  local_mocked_bindings(
    snowflake_auth_args = function(...) list()
  )
  expect_snapshot(
    snowflake_args(account = "testorg-test_account", driver = "driver"),
    error = TRUE
  )
})

test_that("we hint viewer-based credentials on Connect", {
  withr::local_envvar(SF_PARTNER = "")
  local_mocked_bindings(
    snowflake_auth_args = function(...) list(),
    running_on_connect = function() TRUE
  )
  expect_snapshot(
    snowflake_args(account = "testorg-test_account", driver = "driver"),
    error = TRUE
  )
})

test_that("tokens can be requested from a Connect server", {
  skip_if_not_installed("connectcreds")

  connectcreds::local_mocked_connect_responses(token = "token")
  expect_equal(
    snowflake_auth_args("testorg-test_account"),
    list(authenticator = "oauth", token = "token")
  )
})

test_that("the default driver falls back to a known driver name", {
  local_mocked_bindings(
    snowflake_default_driver_paths = function() character(),
    odbcListDrivers = function() list(name = c("bar", "Snowflake"))
  )
  expect_equal(snowflake_default_driver(), "Snowflake")
})

test_that("we error if we can't find the driver", {
  local_mocked_bindings(
    snowflake_default_driver_paths = function() character(),
    odbcListDrivers = function() list()
  )
  expect_snapshot(snowflake_default_driver(), error = TRUE)
})

test_that("Workbench-managed credentials are detected correctly", {
  # Emulate the connections.toml file written by Workbench.
  sf_home <- tempfile("posit-workbench")
  dir.create(sf_home)
  writeLines(
    c(
      '[workbench]',
      'account = "testorg-test_account"',
      'token = "token"',
      'authenticator = "oauth"'
    ),
    file.path(sf_home, "connections.toml")
  )
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "testorg-test_account",
    SNOWFLAKE_DEFAULT_CONNECTION_NAME = "workbench",
    SNOWFLAKE_HOME = sf_home
  )
  args <- snowflake_args(driver = "driver")
  expect_equal(args$token, "token")
  expect_equal(args$authenticator, "oauth")
})

test_that("Workbench-managed credentials are rejected for other accounts", {
  # Emulate the connections.toml file written by Workbench.
  sf_home <- tempfile("posit-workbench")
  dir.create(sf_home)
  writeLines(
    c(
      '[workbench]',
      'account = "nonmatching"',
      'token = "token"',
      'authenticator = "oauth"'
    ),
    file.path(sf_home, "connections.toml")
  )
  withr::local_envvar(
    SNOWFLAKE_HOME = sf_home,
    SNOWFLAKE_DEFAULT_CONNECTION_NAME = "workbench",
    SF_PARTNER = "test_partner"
  )
  expect_snapshot(
    snowflake_args(account = "testorg-test_account", driver = "driver"),
    error = TRUE
  )
})

test_that("snowflake_connection_to_odbc_args maps fields correctly", {
  conn <- structure(
    list(
      name = "test",
      account = "testorg-test_account",
      user = "myuser",
      authenticator = "oauth",
      token = structure("mytoken", class = c("snowflake_redacted", "character")),
      warehouse = "mywh",
      database = "mydb",
      schema = "myschema",
      role = "myrole"
    ),
    class = c("snowflake_connection", "list")
  )
  args <- snowflake_connection_to_odbc_args(conn)
  expect_equal(args$uid, "myuser")
  expect_equal(args$token, "mytoken")
  expect_false(inherits(args$token, "snowflake_redacted"))
  expect_equal(args$authenticator, "oauth")
  expect_equal(args$warehouse, "mywh")
  expect_equal(args$database, "mydb")
  expect_equal(args$schema, "myschema")
  expect_equal(args$role, "myrole")
  expect_null(args$account)
  expect_null(args$name)
})

test_that("snowflake_connection_to_odbc_args maps private key fields", {
  conn <- structure(
    list(
      name = "test",
      account = "testorg-test_account",
      user = "myuser",
      authenticator = "SNOWFLAKE_JWT",
      private_key_file = "/path/to/rsa_key.p8",
      private_key_file_pwd = structure("secret", class = c("snowflake_redacted", "character"))
    ),
    class = c("snowflake_connection", "list")
  )
  args <- snowflake_connection_to_odbc_args(conn)
  expect_equal(args$priv_key_file, "/path/to/rsa_key.p8")
  expect_equal(args$priv_key_file_pwd, "secret")
  expect_false(inherits(args$priv_key_file_pwd, "snowflake_redacted"))
  expect_equal(args$authenticator, "SNOWFLAKE_JWT")
})

test_that("snowflake_connection_to_odbc_args omits default authenticator", {
  conn <- structure(
    list(
      name = "test",
      account = "testorg-test_account",
      user = "myuser",
      password = structure("mypwd", class = c("snowflake_redacted", "character")),
      authenticator = "snowflake"
    ),
    class = c("snowflake_connection", "list")
  )
  args <- snowflake_connection_to_odbc_args(conn)
  expect_null(args$authenticator)
  expect_equal(args$uid, "myuser")
  expect_equal(args$pwd, "mypwd")
  expect_false(inherits(args$pwd, "snowflake_redacted"))
})

test_that("snowflake_connection_to_odbc_args passes through token_file_path", {
  conn <- structure(
    list(
      name = "test",
      account = "testorg-test_account",
      authenticator = "oauth",
      token_file_path = "/snowflake/session/token"
    ),
    class = c("snowflake_connection", "list")
  )
  args <- snowflake_connection_to_odbc_args(conn)
  expect_equal(args$token_file_path, "/snowflake/session/token")
  expect_equal(args$authenticator, "oauth")
})

test_that("host field overrides server in snowflake_args", {
  sf_home <- tempfile("snowflake-host-test")
  dir.create(sf_home)
  writeLines(
    c(
      '[default]',
      'account = "testorg-test_account"',
      'host = "custom.snowflake.example.com"',
      'token = "token"',
      'authenticator = "oauth"'
    ),
    file.path(sf_home, "connections.toml")
  )
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "testorg-test_account",
    SNOWFLAKE_HOME = sf_home
  )
  args <- snowflake_args(driver = "driver")
  expect_equal(args$server, "custom.snowflake.example.com")
})

test_that("snowflake_server uses host when provided", {
  expect_equal(
    snowflake_server("testorg-test_account", host = "custom.example.com"),
    "custom.example.com"
  )
  expect_equal(
    snowflake_server("testorg-test_account", host = NULL),
    "testorg-test_account.snowflakecomputing.com"
  )
  expect_equal(
    snowflake_server("testorg-test_account", host = ""),
    "testorg-test_account.snowflakecomputing.com"
  )
})

test_that("explicit account works without connections.toml", {
  sf_home <- tempfile()
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "",
    SNOWFLAKE_HOME = sf_home
  )
  args <- snowflake_args(
    account = "testorg-test_account",
    driver = "driver",
    uid = "user",
    pwd = "pwd"
  )
  expect_equal(args$account, "testorg-test_account")
  expect_equal(args$server, "testorg-test_account.snowflakecomputing.com")
})

test_that("connection_name selects a named connection", {
  sf_home <- tempfile("snowflake-named")
  dir.create(sf_home)
  writeLines(
    c(
      '[default]',
      'account = "default-account"',
      'token = "default-token"',
      'authenticator = "oauth"',
      '',
      '[production]',
      'account = "prod-account"',
      'token = "prod-token"',
      'authenticator = "oauth"'
    ),
    file.path(sf_home, "connections.toml")
  )
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "",
    SNOWFLAKE_HOME = sf_home
  )
  args <- snowflake_args(driver = "driver", connection_name = "production")
  expect_equal(args$account, "prod-account")
  expect_equal(args$token, "prod-token")
  expect_equal(args$server, "prod-account.snowflakecomputing.com")
  expect_null(args$connection_name)
})

test_that("account is resolved from connections.toml when SNOWFLAKE_ACCOUNT is unset", {
  sf_home <- tempfile("snowflake-no-account")
  dir.create(sf_home)
  writeLines(
    c(
      '[default]',
      'account = "testorg-test_account"',
      'token = "token"',
      'authenticator = "oauth"'
    ),
    file.path(sf_home, "connections.toml")
  )
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "",
    SNOWFLAKE_HOME = sf_home
  )
  args <- snowflake_args(driver = "driver")
  expect_equal(args$account, "testorg-test_account")
  expect_equal(args$server, "testorg-test_account.snowflakecomputing.com")
  expect_equal(args$token, "token")
})

test_that("NULL caller args don't clobber config-resolved values", {
  sf_home <- tempfile("snowflake-null-args")
  dir.create(sf_home)
  writeLines(
    c(
      '[default]',
      'account = "testorg-test_account"',
      'user = "me@example.com"',
      'authenticator = "externalbrowser"',
      'warehouse = "mywh"',
      'database = "mydb"',
      'schema = "myschema"'
    ),
    file.path(sf_home, "connections.toml")
  )
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "testorg-test_account",
    SNOWFLAKE_HOME = sf_home
  )
  args <- snowflake_args(
    driver = "driver",
    uid = NULL,
    pwd = NULL,
    warehouse = NULL,
    database = NULL,
    schema = NULL
  )
  expect_equal(args$uid, "me@example.com")
  expect_equal(args$authenticator, "externalbrowser")
  expect_equal(args$warehouse, "mywh")
  expect_equal(args$database, "mydb")
  expect_equal(args$schema, "myschema")
})

test_that("connection_name bypasses account mismatch check", {
  sf_home <- tempfile("snowflake-bypass")
  dir.create(sf_home)
  writeLines(
    c(
      '[myconn]',
      'account = "other-account"',
      'token = "tok"',
      'authenticator = "oauth"'
    ),
    file.path(sf_home, "connections.toml")
  )
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "testorg-test_account",
    SNOWFLAKE_HOME = sf_home
  )
  args <- snowflake_args(driver = "driver", connection_name = "myconn")
  expect_equal(args$account, "other-account")
  expect_equal(args$token, "tok")
  expect_equal(args$server, "other-account.snowflakecomputing.com")
})

test_that("workload_identity params are passed through", {
  conn <- structure(
    list(
      name = "test",
      account = "testorg-test_account",
      authenticator = "workload_identity",
      workload_identity_provider = "OIDC",
      token_file_path = "/var/run/secrets/token"
    ),
    class = c("snowflake_connection", "list")
  )
  args <- snowflake_connection_to_odbc_args(conn)
  expect_equal(args$authenticator, "workload_identity")
  expect_equal(args$workload_identity_provider, "OIDC")
  expect_equal(args$token_file_path, "/var/run/secrets/token")
})

#' List Configured Data Source Names
#'
#' @description
#'
#' Collect information about the available data source names (DSNs). A DSN must
#' be both installed and configured with the driver manager to be included in
#' this list. Configuring a DSN just sets up a lookup table (e.g. in
#' `odbc.ini`) to allow users to pass only the DSN to [dbConnect()].
#'
#' DSNs that are not configured with the driver manager can still be
#' connected to with [dbConnect()] by providing DSN metadata directly.
#'
#' @return A data frame with two columns:
#' \describe{
#'   \item{name}{Name of the data source. The entries in this column can be
#'     passed to the `dsn` argument of [dbConnect()].}
#'   \item{description}{Data source description.}
#' }
#'
#' @section Configuration:
#'
#' This function interfaces with the driver manager to collect information
#' about the available data source names.
#'
#' For **MacOS and Linux**, the odbc package supports the unixODBC driver
#' manager. unixODBC looks to the `odbc.ini` _configuration file_ for information
#' on DSNs. Find the location(s) of your `odbc.ini` file(s) with `odbcinst -j`.
#'
#' In this example `odbc.ini` file:
#'
#' ```
#' [MySQL]
#' Driver = MySQL Driver
#' Database = test
#' Server = 127.0.0.1
#' User = root
#' password = root
#' Port = 3306
#' ```
#'
#' ...the data source name is `MySQL`, which will appear in the `name`
#' column of this function's output. To pass the DSN as the `dsn` argument to
#' [dbConnect()], pass it as a string, like `"MySQL"`.
#' `Driver = MySQL Driver` references the driver `name` in [odbcListDrivers()]
#' output.
#'
#' **Windows** is [bundled](https://learn.microsoft.com/en-us/sql/odbc/admin/odbc-data-source-administrator)
#' with an ODBC driver manager.
#'
#' When a DSN is configured with a driver manager, information on the DSN will
#' be automatically passed on to [dbConnect()] when its `dsn` argument is set.
#'
#' For example, with the `MySQL` data source name configured, and the driver
#' name `MySQL Driver` appearing in [odbcListDrivers()] output, the code:
#'
#' ```
#' con <-
#'   dbConnect(
#'     odbc::odbc(),
#'     Driver = "MySQL Driver",
#'     Database = "test",
#'     Server = "127.0.0.1",
#'     User = "root",
#'     password = "root",
#'     Port = 3306
#'   )
#' ```
#'
#' ...can be written:
#'
#' ```
#' con <- dbConnect(odbc::odbc(), dsn = "MySQL")
#' ```
#'
#' In this case, `dbConnect()` will look up the information defined for `MySQL`
#' in the driver manager (in our example, `odbc.ini`) and automatically
#' pass the needed arguments.
#'
#' @seealso
#'
#' [odbcListDrivers()]
#'
#' @export
odbcListDataSources <- function() {
  list_data_sources_()
}

test_that("odbcListDataSources() returns available data sources", {
  skip_on_cran()
  skip_if_no_drivers()

  res <- odbcListDataSources()
  if (nrow(res) == 0) {
    skip("No odbc data-sources configured")
  }
  expect_identical(names(res), c("name", "description"))
  expect_true(nrow(res) >= 1)
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

list_drivers_ <- function() {
    .Call(`_odbc_list_drivers_`)
}

list_data_sources_ <- function() {
    .Call(`_odbc_list_data_sources_`)
}

odbc_connect <- function(connection_string, timezone = "", timezone_out = "", encoding = "", name_encoding = "", bigint = 0L, timeout = 0L, r_attributes = NULL, interruptible_execution = TRUE) {
    .Call(`_odbc_odbc_connect`, connection_string, timezone, timezone_out, encoding, name_encoding, bigint, timeout, r_attributes, interruptible_execution)
}

has_result <- function(p) {
    .Call(`_odbc_has_result`, p)
}

connection_info <- function(p) {
    .Call(`_odbc_connection_info`, p)
}

connection_quote <- function(p) {
    .Call(`_odbc_connection_quote`, p)
}

connection_release <- function(p) {
    invisible(.Call(`_odbc_connection_release`, p))
}

connection_begin <- function(p) {
    invisible(.Call(`_odbc_connection_begin`, p))
}

connection_commit <- function(p) {
    invisible(.Call(`_odbc_connection_commit`, p))
}

connection_rollback <- function(p) {
    invisible(.Call(`_odbc_connection_rollback`, p))
}

connection_valid <- function(p) {
    .Call(`_odbc_connection_valid`, p)
}

connection_sql_tables <- function(p, catalog_name = NULL, schema_name = NULL, table_name = NULL, table_type = NULL) {
    .Call(`_odbc_connection_sql_tables`, p, catalog_name, schema_name, table_name, table_type)
}

connection_sql_catalogs <- function(p) {
    .Call(`_odbc_connection_sql_catalogs`, p)
}

connection_sql_schemas <- function(p) {
    .Call(`_odbc_connection_sql_schemas`, p)
}

connection_sql_table_types <- function(p) {
    .Call(`_odbc_connection_sql_table_types`, p)
}

connection_sql_columns <- function(p, column_name = NULL, catalog_name = NULL, schema_name = NULL, table_name = NULL) {
    .Call(`_odbc_connection_sql_columns`, p, column_name, catalog_name, schema_name, table_name)
}

transactionLevels <- function() {
    .Call(`_odbc_transactionLevels`)
}

set_transaction_isolation <- function(p, level) {
    invisible(.Call(`_odbc_set_transaction_isolation`, p, level))
}

bigint_mappings <- function() {
    .Call(`_odbc_bigint_mappings`)
}

result_release <- function(r) {
    invisible(.Call(`_odbc_result_release`, r))
}

result_active <- function(r) {
    .Call(`_odbc_result_active`, r)
}

result_completed <- function(r) {
    .Call(`_odbc_result_completed`, r)
}

new_result <- function(p, sql, immediate) {
    .Call(`_odbc_new_result`, p, sql, immediate)
}

result_fetch <- function(r, n_max = -1L) {
    .Call(`_odbc_result_fetch`, r, n_max)
}

result_column_info <- function(r) {
    .Call(`_odbc_result_column_info`, r)
}

result_bind <- function(r, params, batch_rows) {
    invisible(.Call(`_odbc_result_bind`, r, params, batch_rows))
}

result_insert_dataframe <- function(r, df, batch_rows) {
    invisible(.Call(`_odbc_result_insert_dataframe`, r, df, batch_rows))
}

result_describe_parameters <- function(r, df) {
    invisible(.Call(`_odbc_result_describe_parameters`, r, df))
}

result_rows_affected <- function(r) {
    .Call(`_odbc_result_rows_affected`, r)
}

result_row_count <- function(r) {
    .Call(`_odbc_result_row_count`, r)
}

column_types <- function(df) {
    invisible(.Call(`_odbc_column_types`, df))
}

#' @export
#' @rdname DBI-classes
setClass("NetezzaSQL", contains = "OdbcConnection")

#' @export
#' @rdname odbcDataType
#' @usage NULL
setMethod("odbcDataType", "NetezzaSQL",
  function(con, obj, ...) {
    switch_type(
      obj,
      character = varchar(obj),
      logical = "BOOL",
      callNextMethod(con, obj, ...)
    )
  }
)

#' @include dbi-connection.R
#' @include odbc-connection.R
NULL

#' Oracle
#'
#' Details of Oracle methods for odbc and DBI generics.
#'
#' @export
#' @rdname Oracle
#' @keywords internal
setClass("Oracle", contains = "OdbcConnection")

#' @rdname Oracle
setMethod("sqlCreateTable", "Oracle",
  function(con,
           table,
           fields,
           row.names = NA,
           temporary = FALSE,
           ...,
           field.types = NULL) {
    check_bool(temporary)
    check_row.names(row.names)
    check_field.types(field.types)
    table <- dbQuoteIdentifier(con, table)
    fields <- createFields(con, fields, field.types, row.names)

    SQL(paste0(
      "CREATE ", if (temporary) " GLOBAL TEMPORARY ", "TABLE ", table, " (\n",
      "  ", paste(fields, collapse = ",\n  "), "\n)\n",
      if (temporary) " ON COMMIT PRESERVE ROWS"
    ))
  }
)

#' @rdname Oracle
#' @description
#' ## `odbcConnectionTables()`
#'
#' Method for an internal function that otherwise relies on the `SQLTables`
#' ODBC API. While this method is much faster than the OEM implementation, it
#' does not look through synonyms.
setMethod("odbcConnectionTables", c("Oracle", "character"),
  function(conn,
           name,
           catalog_name = NULL,
           schema_name = NULL,
           table_type = NULL,
           exact = FALSE) {
    qTable <- getSelector("object_name", name, exact)
    if (is.null(schema_name)) {
      query <- paste0(
        " SELECT null AS \"table_catalog\", USER AS \"table_schema\", object_name AS \"table_name\", object_type AS \"table_type\", null AS \"table_remarks\"",
        " FROM user_objects ",
        " WHERE 1 = 1 ", qTable,
        " AND ( object_type = 'TABLE' OR object_type = 'VIEW' ) "
      )
    } else {
      qSchema <- getSelector("owner", schema_name, exact)
      query <- paste0(
        " SELECT null AS \"table_catalog\", owner AS \"table_schema\", object_name AS \"table_name\", object_type AS \"table_type\", null AS \"table_remarks\"",
        " FROM all_objects ",
        " WHERE 1 = 1 ", qSchema, qTable,
        " AND ( object_type = 'TABLE' OR object_type = 'VIEW' ) "
      )
    }

    dbGetQuery(conn, query)
  }
)

#' @rdname Oracle
#' @description
#' ## `odbcConnectionColumns()`
#'
#' Query, rather than use `SQLColumns` ODBC API, since we bind a `BIGINT` to
#' one of the column results and Oracle's OEM driver can't handle it.
#' @usage NULL
setMethod("odbcConnectionColumns", c("Oracle", "character"),
  function(conn,
           name,
           catalog_name = NULL,
           schema_name = NULL,
           column_name = NULL,
           exact = FALSE) {
    query <- ""
    baseSelect <- paste0(
      "SELECT
      column_name AS \"name\",
      data_type AS \"field.type\",
      table_name AS \"table_name\",
      owner AS \"schema_name\",
      null AS \"catalog_name\",
      null AS \"data_type\",
      decode(data_type,'CHAR',char_length,'NCHAR',char_length, 'VARCHAR2',char_length,'NVARCHAR2',char_length, 'RAW',data_length, 'ROWID', data_length, 'UNDEFINED', 0, data_precision) AS \"column_size\",
      decode(data_type, 'DATE',16,'FLOAT',8,'BINARY_FLOAT',4,'BINARY_DOUBLE',8,'LONG RAW',2147483647,'LONG',2147483647,'CLOB',2147483647,'NCLOB',2147483647,'BLOB',2147483647,'BFILE',2147483647,'CHAR',data_length,'NCHAR',data_length,'VARCHAR2',data_length,'NVARCHAR2',data_length,'NUMBER',NVL(data_precision+2,40),data_length) AS \"buffer_length\",
      data_scale AS \"decimal_digits\",
      null AS \"numeric_precision_radix\",
      null AS \"remarks\",
      null AS \"column_default\",
      null AS \"sql_data_type\",
      null AS \"sql_datetime_subtype\",
      decode(data_type,'CHAR',data_length,'VARCHAR2',data_length,'NVARCHAR2',data_length,'NCHAR',data_length, 0) AS \"char_octet_length\",
      column_id AS \"ordinal_position\",
      decode(nullable, 'Y', 1, 'N', 0) AS \"nullable\""
    )
    qTable <- getSelector("table_name", name, exact)
    if (is.null(schema_name)) {
      baseSelect <- gsub("owner AS \"schema_name\"", paste0("'", conn@info$username, "' AS \"schema_name\""), baseSelect)
      query <- paste0(
        baseSelect,
        " FROM user_tab_columns ",
        " WHERE 1 = 1 ", qTable
      )
    } else {
      qSchema <- getSelector("owner", schema_name, exact)
      query <- paste0(
        baseSelect,
        " FROM all_tab_columns ",
        " WHERE 1 = 1 ", qSchema, qTable
      )
    }

    res <- dbGetQuery(conn, query)

    # SQLColumns does not correctly describe buffer sizes for date/time
    # fields ( nor does our custom query above ).  We use these data
    # as a fallback for SQLDescribeParam ( also incorrectly describes
    # parameters for ORACLE ) when writing using prepared statements ( see
    # `dbAppendTable` ).  Fix them here.  91 and 93 are SQL_TYPE_DATE,
    # and SQL_TYPE_TIMESTAMP, respectively in sql.h.  Six and sixteen
    # are the sizes of nanodbc::date and nanodbc::timestmp in bytes
    # (#349, #350, #391).
    res$data_type <- as.numeric(res$data_type)
    isDate <- res$field.type == "DATE"
    res$data_type[isDate] <- 93
    res$column_size[isDate] <- 16
    isTimestamp <- grepl("TIMESTAMP", res$field.type)
    res$data_type[isTimestamp] <- 93
    res$column_size[isTimestamp] <- 16
    res
  }
)

#' @export
#' @rdname odbcDataType
#' @usage NULL
setMethod("odbcDataType", "Oracle",
  function(con, obj, ...) {
    switch_type(
      obj,
      factor = "VARCHAR2(255)",

      # No native oracle type for time
      time = "VARCHAR2(255)",
      date = "DATE",
      datetime = "TIMESTAMP",
      binary = "BLOB",
      integer = "INTEGER",
      int64 = "INTEGER",
      double = "BINARY_DOUBLE",
      character = "VARCHAR2(255)",
      logical = "DECIMAL",
      list = "VARCHAR2(255)",
      stop("Unsupported type", call. = FALSE)
    )
  }
)

#' List locations of ODBC configuration files
#'
#' @description
#' On MacOS and Linux, odbc uses the unixODBC driver manager to manage
#' information about driver and data sources. This helper returns the filepaths
#' where the driver manager will look for that information.
#'
#' `odbcListConfig()` is a wrapper around the command line call `odbcinst -j`.
#' The [odbcEditDrivers()], [odbcEditSystemDSN()], and
#' [odbcEditUserDSN()] helpers provide a shorthand for
#' `file.edit(odbcListConfig()[[i]])`.
#'
#' Windows does not use `.ini` configuration files; on Windows,
#' `odbcListConfig()` will return a 0-length vector and `odbcEdit*()`
#' will raise an error.
#'
#' @seealso
#' The [odbcListDrivers()] and [odbcListDataSources()] helpers return
#' information on the contents of `odbcinst.ini` and `odbc.ini` files,
#' respectively. [odbcListDataSources()] collates entries from both the
#' System and User `odbc.ini` files.
#'
#' Learn more about unixODBC and the `odbcinst` utility
#' [here](https://www.unixodbc.org/odbcinst.html).
#'
#' @examplesIf FALSE
#' configs <- odbcListConfig()
#'
#' configs
#'
#' # shorthand for file.edit(configs[[1]])
#' odbcEditDrivers()
#' @export
odbcListConfig <- function() {
  if (is_windows()) {
    return(character(0))
  }

  if (!has_unixodbc()) {
    abort(
      c("The unixODBC driver manager is not available. ",
        "Please install and try again.")
    )
  }

  res <- system("odbcinst -j", intern = TRUE)
  res <- res[grepl("\\.ini", res)]
  res <- strsplit(res, "\\:")

  if (!identical(vapply(res, length, numeric(1)), c(2, 2, 2))) {
    abort("Failed to parse output from odbcinst.", .internal = TRUE)
  }

  res <- vapply(res, `[[`, character(1), 2)
  res <- trimws(res)
  names(res) <- c("drivers", "system_dsn", "user_dsn")

  res
}

#' @rdname odbcListConfig
#' @export
odbcEditDrivers <- function() {
  open_config(1)
}

#' @rdname odbcListConfig
#' @export
odbcEditSystemDSN <- function() {
  open_config(2)
}

#' @rdname odbcListConfig
#' @export
odbcEditUserDSN <- function() {
  open_config(3)
}

open_config <- function(idx, call = caller_env()) {
  config <- odbcListConfig()

  if (length(config) == 0) {
    cli::cli_abort(
      c(
        "Configuration files are only available on macOS and Linux.",
        "i" = "Use the ODBC Data Source Administrator to configure connections on Windows."
      ),
      call = call
    )
  }

  utils::file.edit(config[[idx]])
}

system <- NULL

has_unixodbc <- function() {
  !identical(unname(Sys.which("odbcinst")), "")
}

test_that("odbcListConfig returns appropriate result", {
  skip_on_os(c("windows", "solaris"))
  skip_if(!has_unixodbc(), "odbcinst not available.")

  res <- odbcListConfig()

  expect_type(res, "character")
  expect_length(res, 3)
  expect_named(res, c("drivers", "system_dsn", "user_dsn"))
  expect_match(res, "\\.ini")
})

test_that("odbcListConfig returns an empty vector on Windows", {
  local_mocked_bindings(is_windows = function() {TRUE})

  res <- odbcListConfig()

  expect_equal(res, character(0))
})

test_that("odbcListConfig errors informatively without unixODBC", {
  local_mocked_bindings(is_windows = function() {FALSE},
                        has_unixodbc = function() {FALSE})

  expect_snapshot(error = TRUE, odbcListConfig())
})

test_that("odbcListConfig errors informatively with unexpected odbcinst output", {
  local_mocked_bindings(is_windows = function() {FALSE},
                        has_unixodbc = function() {TRUE})

  local_mocked_bindings(system = function(...) {c("beep", "bop")})
  expect_snapshot(error = TRUE, odbcListConfig())

  local_mocked_bindings(system = function(...) {""})
  expect_snapshot(error = TRUE, odbcListConfig())
})

test_that("odbcEdit*() errors informatively on Windows", {
  local_mocked_bindings(is_windows = function() {TRUE})

  expect_snapshot(error = TRUE, odbcEditDrivers())
})

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom bit64 integer64
#' @importFrom blob blob
#' @importFrom hms hms
#' @importFrom lifecycle deprecated
#' @importFrom Rcpp sourceCpp
#' @useDynLib odbc, .registration = TRUE
## usethis namespace: end
NULL

utils::globalVariables(c("attribute", "name"))

#' @include dbi-connection.R
NULL

#' @export
#' @rdname DBI-classes
setClass("Impala", contains = "OdbcConnection")

# TODO: Revisit binary type (Impala)
#' @export
#' @rdname odbcDataType
#' @usage NULL
setMethod("odbcDataType", "Impala",
  function(con, obj, ...) {
    switch_type(
      obj,
      factor = "STRING",
      datetime = "STRING",
      date = "VARCHAR(10)",
      integer = "INT",
      int64 = "INT",
      double = "DOUBLE",
      character = "STRING",
      logical = "BOOLEAN",
      list = "STRING",
      time = ,
      stop("Unsupported type", call. = FALSE)
    )
  }
)

#' @export
#' @rdname DBI-classes
setClass("Redshift", contains = "OdbcConnection")

#' @export
#' @rdname odbcDataType
#' @usage NULL
setMethod("odbcDataType", "Redshift", function(con, obj, ...) {
  switch_type(
    obj,
    factor = "VARCHAR(255)",
    datetime = "TIMESTAMP",
    date = "DATE",
    integer = "INTEGER",
    int64 = "INTEGER",
    double = "DOUBLE PRECISION",
    character = "VARCHAR(255)",
    logical = "BOOLEAN",
    list = "VARCHAR(255)",
    time = ,
    binary = ,
    stop("Unsupported type", call. = FALSE)
  )
})

#' Helper for Connecting to Redshift via ODBC
#'
#' @description
#'
#' Connect to Redshift clusters via ODBC.
#'
#' In particular, the custom `dbConnect()` method for Redshift ODBC drivers
#' automatically determines whether IAM-based credentials are available, much
#' like other AWS SDKs and tools. This requires the \pkg{paws.common} package.
#'
#' @param drv an object that inherits from [DBI::DBIDriver-class],
#' or an existing [DBI::DBIConnection-class]
#' object (in order to clone an existing connection).
#' @param clusterId The Redshift cluster identifier. Only one of `clusterId` or
#'   the more verbose `server` is required.
#' @param region The AWS region the Redshift cluster runs in. Ignored when
#'   `server` is provided. Defaults to the value of the environment variable
#'   `AWS_REGION`, then `AWS_REGION`, or uses `us-east-1` if both are unset.
#' @param database The name of the Redshift database to connect to.
#' @param dbUser The Redshift database account.
#' @param server The full hostname of the Redshift cluster.
#' @param driver The name of or path to a Redshift ODBC driver, or `NULL` to
#'   locate one automatically.
#' @param uid,pwd Disable IAM credentials and manually specify a username and
#'   password for authentication.
#' @param ... Further arguments passed on to [`dbConnect()`].
#'
#' @returns An `OdbcConnection` object with an active connection to a Redshift
#'   cluster or SQL warehouse.
#'
#' @examples
#' \dontrun{
#' # Connect to Redshift using IAM credentials.
#' DBI::dbConnect(
#'   odbc::redshift(),
#'   clusterId = "my-testing-cluster",
#'   database = "dev",
#'   dbUser = "me"
#' )
#' }
#' @export
redshift <- function() {
  new("RedshiftOdbcDriver")
}

#' @rdname redshift
#' @export
setClass("RedshiftOdbcDriver", contains = "OdbcDriver")

#' @rdname redshift
#' @export
setMethod(
  "dbConnect",
  "RedshiftOdbcDriver",
  function(
    drv,
    clusterId,
    server,
    database,
    region = NULL,
    driver = NULL,
    uid = NULL,
    pwd = NULL,
    dbUser = uid,
    ...
  ) {
    call <- caller_env()
    switch(
      check_exclusive(clusterId, server, .call = call),
      clusterId = check_string(clusterId, allow_empty = FALSE, call = call),
      server = check_string(server, allow_empty = FALSE, call = call)
    )
    region <- region %||% default_aws_region()
    check_string(region, allow_empty = FALSE, call = call)
    check_string(database, allow_empty = FALSE, call = call)
    check_string(driver, allow_null = TRUE, call = call)
    check_string(uid, allow_null = TRUE, call = call)
    check_string(pwd, allow_null = TRUE, call = call)
    args <- redshift_args(
      driver = driver,
      uid = uid,
      pwd = pwd,
      clusterId = if (!missing(clusterId)) clusterId,
      region = if (!missing(clusterId)) region,
      server = if (!missing(server)) server,
      database = database,
      dbUser = dbUser,
      ...
    )
    inject(dbConnect(odbc(), !!!args))
  }
)

redshift_args <- function(driver = NULL, uid = NULL, pwd = NULL, ...) {
  default_args <- list(
    driver = driver %||% redshift_default_driver(),
    port = 5439L,
    endpointUrl = getenv2("AWS_ENDPOINT_URL_REDSHIFT")
  )
  auth_args <- redshift_auth_args(uid = uid, pwd = pwd)
  utils::modifyList(c(default_args, auth_args), list(...))
}

redshift_default_driver <- function() {
  find_default_driver(
    redshift_default_driver_paths(),
    fallbacks = c("Redshift", "Amazon Redshift (x64)"),
    label = "Redshift",
    call = quote(DBI::dbConnect())
  )
}

redshift_default_driver_paths <- function() {
  if (Sys.info()["sysname"] == "Linux") {
    c(
      "/opt/rstudio-drivers/redshift/bin/lib/libamazonredshiftodbc_sb64.so",
      "/opt/amazon/redshiftodbc/lib/64/libamazonredshiftodbc64.so",
      "/opt/amazon/redshiftodbcx64/librsodbc64.so"
    )
  } else if (Sys.info()["sysname"] == "Darwin") {
    "/opt/amazon/redshift/lib/amazonredshiftodbc.dylib"
  } else {
    character()
  }
}

redshift_auth_args <- function(uid = NULL, pwd = NULL) {
  if (!is.null(uid) && !is.null(pwd)) {
    return(list(uid = uid, pwd = pwd))
  }
  if (is.null(uid) && !is.null(pwd)) {
    cli::cli_abort(
      c(
        "Both {.arg uid} and {.arg pwd} must be specified for manual \
         authentication to Redshift.",
        i = "Or leave {.arg pwd} unset to use IAM credentials."
      ),
      call = quote(DBI::dbConnect())
    )
  }
  check_installed("paws.common", "to use IAM credentials for Redshift")
  try_fetch(
    {
      creds <- paws.common::locate_credentials()
      list(
        iam = 1L,
        accessKeyId = creds$access_key_id,
        secretAccessKey = creds$secret_access_key,
        sessionToken = creds$session_token
      )
    },
    error = function(cnd) {
      cli::cli_abort(
        c(
          "No IAM credentials found.",
          i = "Set {.arg uid} and {.arg pwd} for manual authentication to \
              Redshift."
        ),
        parent = cnd,
        call = quote(DBI::dbConnect())
      )
    }
  )
}

default_aws_region <- function() {
  # Check the preferred AWS_REGION environment variable but respect the ancient
  # AWS_DEFAULT_REGION one when set, too.
  getenv2("AWS_REGION", "AWS_DEFAULT_REGION", .unset = "us-east-1")
}

test_that("the 'uid' and 'pwd' arguments suppress IAM auth", {
  args <- redshift_args(driver = "driver", uid = "uid", pwd = "pwd")
  expect_equal(args$uid, "uid")
  expect_equal(args$pwd, "pwd")
  expect_equal(args$iam, NULL)
})

test_that("IAM credentials in environment variables are handled correctly", {
  withr::local_envvar(
    AWS_ACCESS_KEY_ID = "access-key-id",
    AWS_SECRET_ACCESS_KEY = "secret-access-key",
    AWS_SESSION_TOKEN = "session-token"
  )
  args <- redshift_args(driver = "driver")
  expect_equal(args$accessKeyId, "access-key-id")
  expect_equal(args$secretAccessKey, "secret-access-key")
  expect_equal(args$sessionToken, "session-token")
})

#' Odbc Driver Methods
#'
#' Implementations of pure virtual functions defined in the `DBI` package
#' for OdbcDriver objects.
#' @name OdbcDriver
#' @keywords internal
NULL

#' @rdname dbConnect-OdbcDriver-method
#' @export
#' @import methods DBI
odbc <- function() {
  new("OdbcDriver")
}

#' @rdname OdbcDriver
#' @export
setClass("OdbcDriver", contains = "DBIDriver")

#' @rdname OdbcDriver
#' @inheritParams methods::show
#' @export
setMethod("show", "OdbcDriver",
  function(object) {
    cat("<OdbcDriver>\n")
    # TODO: Print more details
  }
)

#' Connect to a database via an ODBC driver
#'
#' The `dbConnect()` method documented here is invoked when [DBI::dbConnect()]
#' is called with the first argument `odbc()`. Connecting to a database via
#' an ODBC driver is likely the first step in analyzing data using the odbc
#' package; for an overview of package concepts, see the *Overview* section
#' below.
#'
#' @param drv An `OdbcDriver`, from `odbc()`.
#' @param dsn The data source name. For currently available options, see the
#'   `name` column of [odbcListDataSources()] output.
#' @param timezone The server time zone. Useful if the database has an internal
#'   timezone that is _not_ 'UTC'. If the database is in your local timezone,
#'   set this argument to [Sys.timezone()]. See [OlsonNames()] for a complete
#'   list of available time zones on your system.
#'   Note, if the datatype itself carries timezone information, as is the case,
#'   for example, with SQL Server::DATETIMEOFFSET, `package:odbc` will make
#'   an effort to respect the timezone declared therein.  In those cases,
#'   this parameter is not used - the timezone that is part of the datatype
#'   takes precedence.
#' @param timezone_out The time zone returned to R. If you want to display
#'   datetime values in the local timezone, set to [Sys.timezone()].
#' @param encoding The text encoding used on the Database. If the database is
#'   not using UTF-8 you will need to set the encoding to get accurate
#'   re-encoding. See [iconvlist()] for a complete list of available encodings
#'   on your system. Note strings are always returned `UTF-8` encoded.
#' @param name_encoding The text encoding for column names used on the
#'   Database.  May be different than the `encoding` argument.  Defaults to
#'   empty string which is equivalent to returning the column names without
#'   performing any conversion.
#' @param driver The ODBC driver name or a path to a driver. For currently
#'   available options, see the `name` column of [odbcListDrivers()] output.
#' @param server The server hostname. Some drivers use `Servername` as the name
#'   for this argument. Not required when configured for the supplied `dsn`.
#' @param database The database on the server. Not required when configured for
#'   the supplied `dsn`.
#' @param uid The user identifier. Some drivers use `username` as the name
#'   for this argument. Not required when configured for the supplied `dsn`.
#' @param pwd The password. Some drivers use `password` as the name
#'   for this argument. Not required when configured for the supplied `dsn`.
#' @param dbms.name The database management system name. This should normally
#'   be queried automatically by the ODBC driver. This name is used as the class
#'   name for the OdbcConnection object returned from [dbConnect()]. However, if
#'   the driver does not return a valid value, it can be set manually with this
#'   parameter.
#' @param attributes A list of connection attributes that are passed
#'   prior to the connection being established. See [ConnectionAttributes].
#' @param interruptible Logical.  If `TRUE` calls to `SQLExecute` and
#'   `SQLExecuteDirect` can be interrupted when the user sends SIGINT ( ctrl-c ).
#'   Otherwise, they block.  Defaults to `TRUE` in interactive sessions, and
#'   `FALSE` otherwise.  It can be set explicitly either by manipulating this
#'   argument, or by setting the global option `odbc.interruptible`.
#' @param ... Additional ODBC keywords. These will be joined with the other
#'   arguments to form the final connection string.
#'
#'   Note that ODBC parameter names are case-insensitive so that (e.g.) `DRV`
#'   and `drv` are equivalent. Since this is different to R and a possible
#'   source of confusion, odbc will error if you supply multiple arguments that
#'   have the same name when case is ignored.
#'
#'   Any values containing a leading or trailing space, a `=`, `;`, `{`,
#'   or `}` are likely to require quoting. Use [quote_value()] for a fairly
#'   standard approach or see your driver documentation for specifics.
#' @param .connection_string A complete connection string, useful if you are
#'   copy pasting it from another source. If this argument is used, any
#'   additional arguments will be appended to this string.
#' @param bigint The R type that `SQL_BIGINT` types should be mapped to.
#'   Default is [bit64::integer64], which allows the full range of 64 bit
#'   integers.
#' @param timeout Time in seconds to timeout the connection attempt. Setting a
#'   timeout of `Inf` or `NA` indicates no timeout. Defaults to 10 seconds.
#'
#' @section Connection strings:
#'
#' Internally, `dbConnect()` creates a connection string using the supplied
#' arguments. Connection string keywords are driver-dependent; the arguments
#' documented here are common, but some drivers may not accept them.
#'
#' Alternatively to configuring DSNs and driver names with the driver manager,
#' you can pass a complete connection string directly as the
#' `.connection_string` argument.
#' [The Connection Strings Reference](https://www.connectionstrings.com) is a
#' useful resource that has example connection strings for a large variety of
#' databases.
#'
#' @section Overview:
#'
#' The odbc package is one piece of the R interface to databases with support
#' for ODBC:
#'
#' \figure{whole-game.png}
#'
#' The package supports any **Database Management System (DBMS)** with ODBC
#' support. Support for a given DBMS is provided by an **ODBC driver**, which
#' defines how to interact with that DBMS using the standardized syntax of ODBC
#' and SQL. Drivers can be downloaded from the DBMS vendor or, if you're a Posit
#' customer, using the [professional drivers](https://docs.posit.co/pro-drivers/).
#' To manage information about each driver and the data sources they provide
#' access to, our computers use a **driver manager**. Windows is bundled with
#' a driver manager, while MacOS and Linux require installation of one; this
#' package supports the [unixODBC](https://www.unixodbc.org/) driver manager.
#'
#' In the **R interface**, the [DBI package](https://dbi.r-dbi.org/) provides a
#' front-end while odbc implements a back-end to communicate with the driver
#' manager. The odbc package is built on top of the
#' [nanodbc](https://nanodbc.github.io/nanodbc/) C++ library.
#'
#' Interfacing with DBMSs using R and odbc involves three high-level steps:
#'
#' 1) *Configure drivers and data sources*: the functions [odbcListDrivers()],
#'   [odbcListDataSources()], and [odbcListConfig()] help to interface with the
#'   driver manager.
#' 2) *Connect to a database*: The [dbConnect()] function, called with the
#'   first argument odbc(), connects to a database using the specified ODBC
#'   driver to create a connection object.
#' 3) *Interface with connections*: The resulting connection object can be
#'   passed to various functions to retrieve information on database
#'   structure ([DBI::dbListTables()][]), iteratively develop queries
#'   ([DBI::dbSendQuery()], [DBI::dbColumnInfo()]), and query data objects
#'   ([DBI::dbFetch()]).
#'
#' @aliases dbConnect odbc
#'
#' @section Learn more:
#'
#' To learn more about databases:
#'
#' * ["Best Practices in Working with Databases"](https://solutions.posit.co/connections/db/)
#'   documents how to use the odbc package with various popular databases.
#' * [The pyodbc "Drivers and Driver Managers" Wiki](https://github.com/mkleehammer/pyodbc/wiki/Drivers-and-Driver-Managers)
#'   provides further context on drivers and driver managers.
#' * [Microsoft's "Introduction to ODBC"](https://learn.microsoft.com/en-us/sql/odbc/reference)
#'   is a thorough resource on the ODBC interface.
#'
#' @import rlang
#' @export
setMethod("dbConnect", "OdbcDriver",
  function(
      drv,
      dsn = NULL,
      ...,
      timezone = "UTC",
      timezone_out = "UTC",
      encoding = "",
      name_encoding = "",
      bigint = c("integer64", "integer", "numeric", "character"),
      timeout = 10,
      driver = NULL,
      server = NULL,
      database = NULL,
      uid = NULL,
      pwd = NULL,
      dbms.name = NULL,
      attributes = NULL,
      interruptible = getOption("odbc.interruptible", interactive()),
      .connection_string = NULL) {
    check_string(dsn, allow_null = TRUE)
    check_string(timezone)
    check_string(timezone_out)
    check_string(encoding)
    check_string(name_encoding)
    arg_match(bigint)
    check_number_decimal(timeout, allow_na = TRUE)
    check_string(driver, allow_null = TRUE)
    check_string(server, allow_null = TRUE)
    check_string(database, allow_null = TRUE)
    check_string(uid, allow_null = TRUE)
    check_string(pwd, allow_null = TRUE)
    check_string(dbms.name, allow_null = TRUE)
    check_bool(interruptible)

    if (!is_windows() && length(locate_install_unixodbc()) == 0) {
      error_install_unixodbc(call = caller_env())
    }

    con <- OdbcConnection(
      dsn = dsn,
      ...,
      timezone = timezone,
      timezone_out = timezone_out,
      encoding = encoding,
      name_encoding = name_encoding,
      bigint = bigint,
      timeout = timeout,
      driver = driver,
      server = server,
      database = database,
      uid = uid,
      pwd = pwd,
      dbms.name = dbms.name,
      attributes = attributes,
      interruptible = interruptible,
      .connection_string = .connection_string
    )

    # perform the connection notification at the top level, to ensure that it's had
    # a chance to get its external pointer connected, and so we can capture the
    # expression that created it
    if (!is.null(getOption("connectionObserver"))) { # nocov start
      addTaskCallback(function(expr, ...) {
        tryCatch(
          {
            if (rlang::is_call(x = expr, name = c("<-", "=")) &&
              "dbConnect" %in% as.character(expr[[3]][[1]])) {
              # notify if this is an assignment we can replay
              on_connection_opened(eval(expr[[2]]), paste(
                c("library(DBI)", deparse(expr)),
                collapse = "\n"
              ))
            }
          },
          error = function(e) {
            warning("Could not notify connection observer. ", e$message, call. = FALSE)
          }
        )

        # always return false so the task callback is run at most once
        FALSE
      })
    } # nocov end

    con
  }
)

#' @rdname OdbcDriver
#' @param dbObj A object inheriting from [DBI::DBIDriver-class]
#' or [DBI::DBIConnection-class]
#' @inheritParams DBI::dbDataType
#' @export
setMethod("dbDataType", "OdbcDriver",
  function(dbObj, obj, ...) {
    odbcDataType(dbObj, obj, ...)
  }
)

#' @rdname OdbcDriver
#' @param dbObj A object inheriting from [DBI::DBIDriver-class]
#' or [DBI::DBIConnection-class]
#' @inheritParams DBI::dbDataType
#' @export
setMethod("dbDataType", c("OdbcDriver", "list"),
  function(dbObj, obj, ...) {
    odbcDataType(dbObj, obj, ...)
  }
)

odbc_data_type_df <- function(dbObj, obj, ...) {
  res <- character(NCOL(obj))
  nms <- names(obj)
  for (i in seq_along(obj)) {
    withCallingHandlers(
      res[[i]] <- odbcDataType(con = dbObj, obj[[i]]),
      error = function(err) {
        cli::cli_abort("Can't determine type for column {nms[[i]]}.", parent = err, call = quote(odbcDataType()))
      }
    )
  }
  names(res) <- nms
  res
}

#' @rdname OdbcDriver
#' @param dbObj A object inheriting from [DBI::DBIDriver-class]
#' or [DBI::DBIConnection-class]
#' @inheritParams DBI::dbDataType
#' @export
setMethod("dbDataType", c("OdbcDriver", "data.frame"), odbc_data_type_df)

#' @rdname OdbcDriver
#' @param dbObj A object inheriting from [DBI::DBIDriver-class]
#' or [DBI::DBIConnection-class]
#' @inheritParams DBI::dbIsValid
#' @export
setMethod("dbIsValid", "OdbcDriver",
  function(dbObj, ...) {
    TRUE
  }
)

#' @rdname OdbcDriver
#' @param dbObj A object inheriting from [DBI::DBIDriver-class]
#' or [DBI::DBIConnection-class]
#' @inheritParams DBI::dbGetInfo
#' @export
setMethod("dbGetInfo", "OdbcDriver",
  function(dbObj, ...) {
    list(max.connections = NULL, driver.version = NULL, client.version = NULL)
  }
)

skip_if_no_unixodbc()

test_that("Oracle", {
  DBItest::make_context(
    odbc(),
    test_connection_string("ORACLE"),
    tweaks = DBItest::tweaks(),
    name = "Oracle"
  )

  DBItest::test_getting_started(c(
    "package_name", # Not an error
    NULL
  ))
  DBItest::test_driver(c(
    "connect_bigint_integer",
    "connect_bigint_numeric",
    "connect_bigint_character",
    "connect_bigint_integer64"
  ))
  DBItest::test_connection()
  DBItest::test_result(c(
    "send_query.*",
    "fetch.*",
    "clear_result_return_query",
    "clear_result_return_statement",
    "clear_result_return_query_arrow",
    "cannot_clear_result_twice_query",
    "cannot_clear_result_twice_statement",
    "get_query_atomic",
    "cannot_clear_result_twice_query_arrow",
    "get_query.*",
    "send_statement.*",
    "execute_atomic",
    "execute_params",
    "execute_immediate",
    "data.*"
  ))
  DBItest::test_sql(c(
    "quote_string.*",
    "quote_literal.*",
    "quote_identifier.*",
    "read_table.*",
    "create_table.*",
    "create_table_visible.*",
    "create_roundtrip_quotes",
    "append_roundtrip_quotes_column_names",
    "append_roundtrip.*",
    "write_table.*",
    "overwrite_table.*",
    "append_table.*",
    "temporary_table.*",
    "table_visible.*",
    "roundtrip.*",
    "remove_table.*",
    "list_objects_features",
    "list_fields_wrong_table"
  ))
  DBItest::test_meta(c(
    "bind.*",
    "is_valid_result_query",
    "has_completed.*",
    "get_statement.*",
    "column_info.*",
    "row_count.*",
    "get_row_count_error",
    "rows_affected.*",
    "get_info_result"
  ))
  DBItest::test_transaction("begin_write_commit.*")
  DBItest::test_compliance("reexport")
})

test_that("can round columns", {
  con <- test_con("ORACLE")
  # - Date does not roundtrip correctly since
  #   their DATE data-type contains hour/min/seconds.
  #   To avoid loss of precision we read it in as
  #   POSIXct.
  # - There also looks like there are issues related
  #   to binary elements of size zero.
  # - Finally, no boolean in Oracle prior to 23
  test_roundtrip(con, columns = c("time", "date", "binary", "logical"))
})

test_that("can detect existence of table", {
  con <- test_con("ORACLE")

  tbl1 <- local_table(con, "mtcarstest", mtcars)
  expect_true(dbExistsTable(con, tbl1))

  tbl2 <- local_table(con, "mtcars_test", mtcars)
  expect_true(dbExistsTable(con, tbl2))
})

test_that("Writing date/datetime with batch size > 1", {
  # See #349, #350, #391
  con <- test_con("ORACLE")
  tbl <- "test_batched_write_w_dates"

  values <- data.frame(
    datetime = as.POSIXct(as.numeric(iris$Petal.Length * 10), origin = "2024-01-01", tz = "UTC"),
    date    = as.Date(as.numeric(iris$Petal.Length * 10), origin = "2024-01-01", tz = "UTC"),
    integer = as.integer(iris$Petal.Width * 100),
    double = iris$Sepal.Length,
    varchar = iris$Species)
  tbl <- local_table(con, tbl, values)
  res <- dbReadTable(con, tbl)
  expect_true(nrow(res) == nrow(values))
  # Now write character dates and datetimes to these targets
  # Issue #959
  valuesMod <- values
  valuesMod$date <- as.character(values$date)
  dbAppendTable(con, tbl, valuesMod)
  res <- dbReadTable(con, tbl)
  expect_true(nrow(res) == 2 * nrow(values))
  valuesMod$date <- as.character(values$datetime)
  dbAppendTable(con, tbl, valuesMod)
  res <- dbReadTable(con, tbl)
  expect_true(nrow(res) == 3 * nrow(values))
  valuesMod$datetime <- as.character(values$datetime)
  dbAppendTable(con, tbl, valuesMod)
  res <- dbReadTable(con, tbl)
  expect_true(nrow(res) == 4 * nrow(values))
  # Finally write dates to datetime targets
  # Issue #959, continued
  valuesMod$datetime <- values$date
  dbAppendTable(con, tbl, valuesMod)
  res <- dbReadTable(con, tbl)
  expect_true(nrow(res) == 5 * nrow(values))

})

#' List Configured ODBC Drivers
#'
#' @description
#'
#' Collect information about the configured driver names. A driver must be both
#' installed and configured with the driver manager to be included in this list.
#' Configuring a driver name just sets up a lookup table (e.g. in
#' `odbcinst.ini`) to allow users to pass only the driver name to [dbConnect()].
#'
#' Driver names that are not configured with the driver manager (and thus
#' do not appear in this function's output) can still be
#' used in [dbConnect()] by providing a path to a driver directly.
#'
#' @param keep,filter A character vector of driver names to keep in or remove
#'   from the results, respectively. If `NULL`, all driver names will be kept,
#'   or none will be removed, respectively. The `odbc.drivers_keep` and
#'   `odbc.drivers_filter` options control the argument defaults.
#'
#'   Driver names are first processed with `keep`, then `filter`. Thus, if a
#'   driver name is in both `keep` and `filter`, it won't appear in output.
#'
#' @return A data frame with three columns.
#'
#' \describe{
#'   \item{name}{Name of the driver. The entries in this column can be
#'     passed to the `driver` argument of [dbConnect()] (as long as the
#'     driver accepts the argument).}
#'   \item{attribute}{Driver attribute name.}
#'   \item{value}{Driver attribute value.}
#' }
#'
#' If a driver has multiple attributes, there will be one row per attribute,
#' each with the same driver `name`. If a given driver name does not have any
#' attributes, the function will return one row with the driver `name`, but
#' the last two columns will be `NA`.
#'
#' @section Configuration:
#'
#' This function interfaces with the driver manager to collect information
#' about the available driver names.
#'
#' For **MacOS and Linux**, the odbc package supports the unixODBC driver
#' manager. unixODBC looks to the `odbcinst.ini` _configuration file_ for
#' information on driver names. Find the location(s) of your `odbcinst.ini`
#' file(s) with `odbcinst -j`.
#'
#' In this example `odbcinst.ini` file:
#'
#' ```
#' [MySQL Driver]
#' Driver=/opt/homebrew/Cellar/mysql/8.2.0_1/lib/libmysqlclient.dylib
#' ```
#'
#' Then the driver name is `MySQL Driver`, which will appear in the `name`
#' column of this function's output. To pass the driver name as the `driver`
#' argument to [dbConnect()], pass it as a string, like `"MySQL Driver"`.
#'
#' **Windows** is [bundled](https://learn.microsoft.com/en-us/sql/odbc/admin/odbc-data-source-administrator)
#' with an ODBC driver manager.
#'
#' In this example, function output would include 1 row: the `name` column
#' would read `"MySQL Driver"`, `attribute` would be `"Driver"`, and `value`
#' would give the file path to the driver. Additional key-value pairs
#' under the driver name would add additional rows with the same `name` entry.
#'
#' When a driver is configured with a driver manager, information on the driver
#' will be automatically passed on to [dbConnect()] when its `driver` argument
#' is set. For an example, see the same section in the [odbcListDataSources()]
#' help-file. Instead of configuring driver information with a driver manager,
#' it is also possible to provide a path to a driver directly to [dbConnect()].
#'
#' @seealso
#' [odbcListDataSources()]
#'
#' @examplesIf FALSE
#' odbcListDrivers()
#'
#' @export
odbcListDrivers <- function(keep = getOption("odbc.drivers_keep"), filter = getOption("odbc.drivers_filter")) {
  res <- list_drivers_()

  if (nrow(res) > 0) {
    res[res == ""] <- NA_character_

    if (!is.null(keep)) {
      res <- res[res[["name"]] %in% keep, ]
    }

    if (!is.null(filter)) {
      res <- res[!res[["name"]] %in% filter, ]
    }
  }

  res
}

test_that("odbcListDrivers() returns available drivers", {
  skip_on_cran()
  skip_if_no_drivers()

  res <- odbcListDrivers()
  expect_identical(names(res), c("name", "attribute", "value"))
  expect_true(nrow(res) >= 1)
})

test_that("odbcListDrivers() keep and filter work", {
  skip_on_cran()
  skip_if_no_drivers()

  current_drivers <- odbcListDrivers()[["name"]]
  res <- odbcListDrivers(filter = current_drivers)
  expect_true(nrow(res) == 0)

  res <- odbcListDrivers(keep = current_drivers[[1]])
  expect_true(unique(res$name) == current_drivers[[1]])
})

#' @include dbi-connection.R
NULL

#' @export
#' @rdname DBI-classes
setClass("HDB", contains = "OdbcConnection")

#' @rdname DBI-methods
setMethod("sqlCreateTable", "HDB",
  function(con,
           table,
           fields,
           row.names = NA,
           temporary = FALSE,
           ...,
           field.types = NULL) {
    check_bool(temporary)
    check_row.names(row.names)
    check_field.types(field.types)
    table <- dbQuoteIdentifier(con, table)
    fields <- createFields(con, fields, field.types, row.names)

    SQL(paste0(
      "CREATE ", if (temporary) "LOCAL TEMPORARY COLUMN ",
      "TABLE ", table, " (\n",
      "  ", paste(fields, collapse = ",\n  "),
      "\n)\n"
    ))
  }
)

#' @include dbi-connection.R
NULL

#' @export
#' @rdname DBI-classes
setClass("Teradata", contains = "OdbcConnection")

#' @rdname DBI-methods
setMethod("sqlCreateTable", "Teradata",
  function(con,
           table,
           fields,
           row.names = NA,
           temporary = FALSE,
           ...,
           field.types = NULL) {
    check_bool(temporary)
    check_row.names(row.names)
    check_field.types(field.types)
    table <- dbQuoteIdentifier(con, table)
    fields <- createFields(con, fields, field.types, row.names)

    SQL(paste0(
      "CREATE ", if (temporary) " MULTISET VOLATILE ", "TABLE ", table, " (\n",
      "  ", paste(fields, collapse = ",\n  "), "\n)\n", if (temporary) " ON COMMIT PRESERVE ROWS"
    ))
  }
)


setMethod("odbcConnectionTables", c("Teradata", "character"),
  function(conn,
           name,
           catalog_name = NULL,
           schema_name = NULL,
           table_type = NULL,
           exact = FALSE) {
    res <- callNextMethod()
    if (!is.null(schema_name)) {
      return(res)
    }

    # Also look through volatile tables if
    # not querying in a specific schema
    tempTableNames <- dbGetQuery(conn, "HELP VOLATILE TABLE")[["Table SQL Name"]]
    # If a name argument is supplied, subset the temp table names vector
    # to either an exact match, or if pattern value, to an approximate match
    if (!is.null(name)) {
      if (!exact && isPatternValue(name)) {
        name <- convertWildCards(name)
      } else {
        name <- escapePattern(name)
        name <- paste0("^", name, "$")
      }
      tempTableNames <- tempTableNames[grepl(name, tempTableNames)]
    }

    navec <- rep(NA_character_, length(tempTableNames))
    rbind(
      res,
      data.frame(
        table_catalog = navec,
        table_schema = navec,
        table_name = tempTableNames,
        table_remarks = navec,
        stringsAsFactors = FALSE
      )
    )
  }
)


#' @export
#' @rdname odbcDataType
#' @usage NULL
setMethod("odbcDataType", "Teradata",
  function(con, obj, ...) {
    switch_type(
      obj,
      factor = "VARCHAR(255)",
      datetime = "TIMESTAMP",
      date = "DATE",
      time = "TIME",
      binary = "BLOB",
      integer = "INTEGER",
      int64 = "INTEGER",
      double = "FLOAT",
      character = "VARCHAR(255)",
      logical = "BYTEINT",
      list = "VARCHAR(255)",
      stop("Unsupported type", call. = FALSE)
    )
  }
)

test_that("show method prints only host if no user is defined", {
  con <- new("OdbcConnection")
  local_mocked_bindings(
    dbGetInfo = function(x) c(servername = "localhost", username = "", dbname = "", dbms.name = "", db.version = ""),
    dbIsValid = function(x) TRUE
  )

  expect_snapshot(con)
})

test_that("show method prints DISCONNECTED if not valid", {
  con <- new("OdbcConnection")
  local_mocked_bindings(
    dbGetInfo = function(x) c(servername = "localhost", username = "", dbname = "", dbms.name = "", db.version = ""),
    dbIsValid = function(x) FALSE
  )
  expect_snapshot(con)
})

test_that("show method does not print server if it is not available", {
  con <- new("OdbcConnection")
  local_mocked_bindings(
    dbGetInfo = function(x) c(servername = "", username = "", dbname = "", dbms.name = "", db.version = ""),
    dbIsValid = function(x) TRUE
  )
  expect_snapshot(con)
})

test_that("dbQuoteIdentifier() errors informatively", {
  skip_if_no_unixodbc()
  con <- test_con("SQLITE")

  expect_snapshot(
    error = TRUE,
    dbQuoteIdentifier(con, NA_character_)
  )
})

test_that("dbConnect() errors informatively without unixODBC (#782)", {
  skip_on_os("windows")
  local_mocked_bindings(
    locate_install_unixodbc = function(...) {character()}
  )
  expect_snapshot(error = TRUE, test_con("SQLITE"), variant = Sys.info()[["sysname"]])
})

#!/bin/bash
set -e
PATH=~/.cargo/bin/:$PATH
git fetch
git checkout master
git reset --hard origin/master
cargo clean
cargo doc
TMP_DIR=`mktemp -d`
mv ./target/doc/* "$TMP_DIR/"
git checkout gh-pages
rm -rf ./*
mv "$TMP_DIR"/* ./
git add .
git commit -am "Update docs gh-pages"
git checkout master
rm -rf "$TMP_DIR"

---
title: "Installing and Configuring Drivers"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Installing and Configuring Drivers}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette outlines how to install common ODBC drivers and configure them with a driver manager.

## Installation

For Unix and MacOS, ODBC drivers should be compiled against [unixODBC](https://www.unixodbc.org/). Drivers compiled against [iODBC](https://www.iodbc.org/) _may_ also work, but are not fully supported.

After installation of the driver manager and driver, you will have to register the driver in a [odbcinst.ini](#dsn-configuration-files) file for it to appear in `odbc::odbcListDrivers()`.

### Windows

Windows is bundled with ODBC libraries. However, drivers for each database need to be installed separately. Windows ODBC drivers typically include an installer that needs to be run and will install the driver to the proper locations.

### MacOS

[homebrew](https://brew.sh/) can be used to easily install database drivers on MacOS.

The UnixODBC driver manager is required for all databases:

```shell
## Install the unixODBC library
brew install unixodbc
```

Most common database drivers can be installed with homebrew. Some drivers are only available from the database vendor.

```shell
## Microsoft SQL Server ODBC drivers
brew tap microsoft/mssql-release https://github.com/Microsoft/homebrew-mssql-release
brew update
brew install msodbcsql17 mssql-tools

## SQL Server ODBC drivers (Free TDS)
brew install freetds

## PostgreSQL ODBC drivers
brew install psqlodbc

## MySQL ODBC drivers (and database)
brew install mysql
brew install mariadb-connector-odbc

## SQLite ODBC drivers
brew install sqliteodbc
```

Some driver installations will not automatically configure the driver with the driver manager. See the ["Connecting to a Database"](#connecting) section below to learn more about configuring drivers and data sources.

### Linux - Debian / Ubuntu

[apt-get](https://wiki.debian.org/Apt) can be used to easily install database
drivers on Linux distributions which support it, such as Debian and Ubuntu.

The UnixODBC driver manager is required for all databases:

```shell
## Install the unixODBC library
apt-get install unixodbc unixodbc-dev
```

For SQL Server on Linux it is recommended you use the [Microsoft Drivers](https://learn.microsoft.com/en-us/sql/connect/odbc/linux-mac/installing-the-microsoft-odbc-driver-for-sql-server?view=sql-server-ver15#ubuntu18). The code below will install some common drivers:

```shell
## SQL Server ODBC Drivers (Free TDS)
apt-get install tdsodbc

## PostgreSQL ODBC ODBC Driversz
apt-get install odbc-postgresql

## MySQL ODBC Drivers
apt-get install libmyodbc

## SQLite ODBC Drivers
apt-get install libsqliteodbc
```

## Driver configuration

On Windows, driver config is managed by the operating system, and you don't generally need to edit any configuration files directly. On MacOS and Linux, however, the driver config is managed by unixODBC, and depending on how the driver is installed, it may or may not be automatically configured. 

You can see which drivers are configured by running `odbcListDrivers()`. If the driver you want is not installed, you will need to edit `odbcinst.ini`. You can use `odbcinst -j` (provided by unixODBC) to find out where this file lives.

The basic form of this file is something like this:

```ini
[PostgreSQL Driver]
Driver          = /usr/local/lib/psqlodbcw.so

[SQLite Driver]
Driver          = /usr/local/lib/libsqlite3odbc.dylib
```

i.e., a mapping between driver names, and the location of the driver file. The driver name is what you pass to `odbc::odbc()` when you connect, e,g:

```{r}
#| eval: false

library(DBI)
con1 <- dbConnect(odbc::odbc(), driver = "PostgreSQL Driver")
con2 <- dbConnect(odbc::odbc(), driver = "SQLite Driver")
```

If installing the driver did not automatically update this file, you'll need to update it yourself. You'll need to figure out where the driver library was installed, using something like `brew list {drivername}` or `dpkg-query -L {drivername}`.

## Data source configuration

It's also possible to configure named data sources, so instead of typing this:

```{r}
#| eval: false

library(DBI)
con <- dbConnect(
  odbc::odbc(),
  driver = "PostgreSQL Driver",
  database = "test_db",
  username = "postgres",
  password = Sys.getenv("PASSWORD"),
  host = "localhost",
  port = 5432
)
```

You can type:

```{r}
#| eval: false

con <- dbConnect(odbc::odbc(), dsn = "test_db")
```

On Windows, you can set up a data source by using the [ODBC Data Source Administrator](https://learn.microsoft.com/en-us/sql/odbc/admin/odbc-data-source-administrator) application is used to manage ODBC data sources on Windows. On Linux and Mac, you'll need to edit `odbc.ini` (again finding the location with `odbcinst -j`). Each data source is defined in a separate section (defined by `[name]`]), with each paramter going on its own line, e.g.:

```ini
[test_db]
driver     = PostgreSQL Driver
database   = test_db
username   = postgres
password   = password
host       = localhost
port       = 5432
```

The `driver` entry represents the name of the driver defined in `odbcinst.ini`. You can see all currently defined data sources by running `odbcListDataSources()`.

## Debugging driver and data source configurations

The odbc package may have trouble locating your driver and data source configurations. If you find that `odbcListDrivers()` and `odbcListDataSources()` report no configurations, check the output of `odbcListConfig()` next. While `odbcListDrivers()` and `odbcListConfig()` interface with the unixodbc driver manager through nanodbc—the tool that the odbc package uses to interface with the ODBC API—`odbcListConfig()` interfaces with unixODBC directly. `odbcListConfig()` will show where unixODBC is looking for your configurations.

* If the files listed in `odbcListConfig()` are at a location other than where you've configured drivers and data sources, either 1) configure the listed files as instructed above, or 2) change the folder where ODBC should look for configurations using the `ODBCSYSINI` environmental variable. See [the following section](#odbcsysini) on setting `ODBCSYSINI`.

* If the files listed in `odbcListConfig()` are where you expected them to be and appear complete, the odbc package may have had trouble interfacing with the unixODBC driver manager through nanodbc. To remedy this, first run `dirname(odbcListConfig()[1])` and note its output. Then, follow the instructions in [the section below](#odbcsysini), replacing `"some/folder"` with the noted output. If this doesn't resolve the issue, try to build the odbc package from source, e.g. with `devtools::install_github("r-dbi/odbc")`.

### Setting `ODBCSYSINI` {#odbcsysini}

The `ODBCSYSINI` environmental variable controls the path where unixODBC and the odbc package will look for configuration files. Setting `Sys.setenv(ODBCSYSINI = "some/folder")` means that your configuration files should be located at `"/my/folder/odbc.ini"` and `"/my/folder/odbcinst.ini"`. Be sure to set `ODBCSYSINI` _before_ loading the odbc package. That is:

* Restart R
* Run `Sys.setenv(ODBCSYSINI = "some/folder")`
* Load the odbc package with `library(odbc)`
* Test `odbcListDrivers()` and `odbcListDataSources()` again

If setting `ODBCSYSINI` resolves the issue you've noted, you may want to set that environmental variable every time you start R. To do so, add an entry like `ODBCSYSINI = some/folder` to your `~/.Renviron` file, which contains environmental variables that are set every time an R session is started. To open that file using R, run `usethis::edit_r_environ()`.

Once you've edited `~/.Renviron`, you'll need to save the file and restart R for changes to take effect.

Thanks for using this project.

#### If you have an error or support request

- Read the error message and documentation.
- Search [existing issues](https://github.com/Koka/odbc-rs/issues), [closed issues](https://github.com/Koka/odbc-rs/issues?page=1&state=closed) first.
- If the issue is with a dependency of this project, post on the dependency's repo.
- If you can fix the issue, submit a PR.
- If the issue persists, post on the issue tracker. Include any information that could help others reproduce and fix.

#### If you have a feature proposal or want to contribute

- Post your proposal on the issue tracker so we can review it together. Some proposals may not be a good fit for the project.
- Contribute a PR (don't forget to write all necessary tests, please).
- Respond to code review.
- Watch the PR be merged.

---
title: "Developing odbc"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Developing odbc}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette is intended to help developers of the R package install needed dependencies.

While the odbc package contains some documentation on how to install and configure database drivers in `vignette("setup")`, the documentation assumes that users are connecting to databases that have already been set up. In order to test package functionality, though, odbc sets up small example database deployments.

For the most part, this vignette assumes a MacOS system with aarch64 (e.g. M1 or M2) architecture. For Linux example code, see [`.github/workflows/db.yaml`](https://github.com/r-dbi/odbc/blob/main/.github/workflows/db.yaml), and for Windows, see [`.github/workflows/db-windows.yml`](https://github.com/r-dbi/odbc/blob/main/.github/workflows/db-windows.yml).

## Posit Professional Drivers

Posit employees have access to the [Posit Professional drivers](https://docs.posit.co/pro-drivers/) and the infrastructure used to build and test them in the rstudio/pro-drivers GitHub repository. The Posit Professional drivers are a set of drivers vendored from Magnitude Simba that support many of the most popular DBMS, including SQL Server, Oracle, Redshift, Databricks, Snowflake, etc. The repository they're developed in contains tooling to spin up a number of databases in docker containers to test against.

Note that Athena, Hive, Impala, MongoDB, and Oracle drivers are [not available for macOS aarch64](https://github.com/oracle/python-cx_Oracle/issues/617) (M1, M2, etc) at the time of writing.

### Drivers

The only documented installation method for these drivers on MacOS is via RStudio Desktop Pro. The Posit confluence page titled "\[INTERNAL\] Posit License Files for Employee Use" contains instructions for installing RStudio Desktop Pro. Once RStudio Pro is installed, install individual drivers as documented [here](https://docs.posit.co/ide/desktop-pro/database_drivers/install_database_drivers.html).

### Databases

Among other things, the rstudio/pro-drivers GitHub repository defines a `MAKE` tool for setting up and tearing down databases in docker containers. Ensure that you have a docker daemon running (i.e. Docker Desktop open) and, if you're on macOS aarch64, have `Settings > Use Rosetta for x86_64/amd64 emulation on Apple Silicon` enabled. To start a container for a given `dbms`, run `MAKE dist=none DB=dbms up`, and **tear it down** with `MAKE dist=none DB=db down`. To see available `dbms` options, see the names of `.yml` files in the `docker-compose` directory. Find connection details for each database in `docker/shared/odbc.ini`.

DBMS-specific notes:

* `sqlserver`: Be sure to pass `uid` and `pwd` as arguments to `dbConnect()` explicitly, even though they're in the `odbc.ini` as well.

On macOS aarch64, you will see a `requested image's platform (linux/amd64) does not match the detected host platform (linux/arm64/v8)` warning when you start any of these containers. Feel free to ignore. :)

## PostgreSQL

On MacOS, install PostgreSQL with:

```shell
brew install postgresql@14
```

You'll also need to install and configure the PostgreSQL driver `psqlodbc`; see `vignette("setup")` to learn more.

To launch a PostgreSQL server locally, run:

```
brew services start postgresql@14
```

Next, create a database called "test" (or by whatever name is in the entry `Database` in your `odbc.ini` file):

```shell
createdb test
```

At this point, you should be able to connect to PostgreSQL through the R interface. Connect with:

```r
postgres <- dbConnect(odbc(), "PostgreSQL")
```

where `"PostgreSQL"` is replaced with whatever DSN you've configured.


## MySQL

First, installing MySQL with Homebrew:

```shell
brew install mysql@8.2
```

MariaDB drivers are compatible with MySQL and are more easily installed than MySQL drivers themselves in most cases. To install the MariaDB driver:

```shell
brew install mariadb-connector-odbc
```

Then, link the MariaDB driver with your MySQL data source name. That is, with the driver name `[MariaDB]` configuring your MariaDB install in `odbcinst.ini`, the first lines of your `odbc.ini` entry should look something like:

```ini
[MySQL]
Driver = MariaDB
```

After running `brew services start mysql` if needed, and confirming that the database is running with `brew services info mysql`, you should be able to:

```
library(odbc)
dbConnect(odbc(), "MySQL")
```

The second argument `"MySQL"` refers to the data source name configured above.

## SQL Server test setup

To run Microsoft SQL Server on **aarch64 (e.g. M1 or M2) MacOS**, you will need: 

* Docker 4.16 or higher
* MacOS 13 Ventura (or higher)

If needed, install Docker with:

```shell
brew install --cask docker
```

The Docker Desktop app provides a GUI to monitor deployed Docker containers and lives in `Docker.app > Show Package Contents > Contents > MacOS > Docker Desktop.app`.

To [install the SQL Server ODBC driver and (optional) command line tool](https://learn.microsoft.com/en-us/sql/connect/odbc/linux-mac/install-microsoft-odbc-driver-sql-server-macos?view=sql-server-ver15), use:

```shell
brew tap microsoft/mssql-release https://github.com/Microsoft/homebrew-mssql-release
brew install microsoft/mssql-release/msodbcsql18 microsoft/mssql-release/mssql-tools18
```

The `odbc.ini` entry should look something like:

```ini
[MicrosoftSQLServer]
driver = ODBC Driver 18 for SQL Server
Server = 127.0.0.1
port = 1433
Encrypt = no
```

With the driver name in `odbcinst.ini` being `[ODBC Driver 18 for SQL Server]`.

With docker and the needed driver installed, deploy the container with:

```shell
sudo docker run \
  --platform linux/amd64 \
  -e "ACCEPT_EULA=Y" \
  -e 'MSSQL_SA_PASSWORD=BoopBop123!' \
  -p 1433:1433 \
   --name sql1 \
   --hostname sql1 \
   -d mcr.microsoft.com/mssql/server:2022-latest
```

The `--platform` tag is correct for M1; if you see `Status: Exited (1)` in Docker Desktop or a warning about incompatible architectures, navigate to `Settings > General` and ensure that `Use Rosetta for x86/amd64 emulation on Apple Silicon` is checked.

To connect via odbc, we need to pass the UID and PWD via the connection string; configuring those arguments via `odbc.ini` is [not permitted](https://stackoverflow.com/questions/42387084/sql-server-odbc-driver-linux-username). With the container deployed as above, the connection arguments would be:

```r
con <- dbConnect(
  odbc::odbc(), 
  dsn = "MicrosoftSQLServer", 
  uid = "SA", 
  pwd = "BoopBop123!"
)
```

Then do some configuration of the server to add a testuser and create the test database

To configure a server to add a testing user and create a test database:

```r
# Add a test user, but currently unused
dbExecute(con, "USE test")
dbExecute(con, "EXEC sp_configure 'contained database authentication', 1")
dbExecute(con, "RECONFIGURE")
dbExecute(con, "alter database test set containment = partial")
dbExecute(con, "CREATE USER testuser with password = 'BoopBop123!'")
dbExecute(con, "GRANT CONTROL TO testuser")
dbExecute(con, "DROP USER testuser")

# Create a test database
dbExecute(con, "CREATE DATABASE test")
```

On **Linux**, create a docker container with:

```shell
docker run -v "$(pwd)":"/opt/$(basename $(pwd))":delegated --security-opt=seccomp:unconfined --link sql1 -it rstudio/r-base:3.6.1-bionic /bin/bash
```

Then run:

```shell
curl https://packages.microsoft.com/keys/microsoft.asc | apt-key add -
#Ubuntu 18.04
curl https://packages.microsoft.com/config/ubuntu/18.04/prod.list > /etc/apt/sources.list.d/mssql-release.list

apt-get update
ACCEPT_EULA=Y apt-get install -y msodbcsql17
apt-get install -y unixodbc-dev
```

The resulting `odbc.ini` file will look something like:

```ini
[MicrosoftSQLServer]
driver = ODBC Driver 17 for SQL Server
Server = sql1
port = 1433
Database = test
```

## SQLite

MacOS provides SQLite natively. With the SQLite odbc driver installed (see `vignette("setup")` if needed), run:

```
library(odbc)
dbConnect(odbc(), "SQLite")
```

The above example assumes the configured SQLite DSN is called `"SQLite"`.

## Oracle

A huge pain.

### On MacOS

Getting an Oracle database booted up in a Docker container on MacOS is a bit challenging. First, you'll build the container image yourself from the official Oracle source using the instructions [here](https://simonpcouch.com/blog/2024-03-14-oracle/). Once you've finished building the image, run it with:

``` shell
docker run -d --name oracle -p 1521:1521 -e ORACLE_PWD=odbcTest321 oracle/database:19.3.0-ee
```

Note that the `run` command here is a bit different than the one in the linked blog post.

Next, you'll install the Basic, ODBC, and SQLPlus Instant Client libraries from the Oracle website (`https://www.oracle.com/database/technologies/instant-client/macos-arm64-downloads.html`) by doing the following:

* Download the Basic, ODBC, and SQLPlus `.dmg`s.
* Create a folder for the instant client to live in. I made mine at `/Users/simoncouch/instantclient_23_3`. (This folder will show up many times in the rest of these instructions--always substitute it with your own from here on out.)
* For each of the `dmg`s, open up the image and copy and paste its contents into the folder you made in the previous step.
* Symlink your `libodbcinst.2.dylib` (unixODBC library) into the folder. I used `ln -s /opt/homebrew/Cellar/unixodbc/2.3.12/lib/libodbcinst.2.dylib /Users/simoncouch/instantclient_23_3/libodbcinst.2.dylib` to do so for myself; use `odbc_config --lib-prefix` to help locate yours.

Make a text file at `/Users/simoncouch/instantclient_23_3/network/admin/tnsnames.ora`. Open that file (maybe with `file.edit("/Users/simoncouch/instantclient_23_3/network/admin/tnsnames.ora")`) and add the following:

``` ora
db=
   (DESCRIPTION=
    (ADDRESS=(PROTOCOL=tcp)(HOST=0.0.0.0)(PORT=1521))
    (CONNECT_DATA=(SERVICE_NAME=orclpdb1)))
```

Now, set some paths to help R and the Instant Client communicate. In `usethis::edit_r_profile()`, add:

``` r
Sys.setenv(PATH = paste0("/Users/simoncouch/instantclient_23_3:", Sys.getenv("PATH")))
Sys.setenv(LD_LIBRARY_PATH = "/Users/simoncouch/instantclient_23_3")
Sys.setenv(ORACLE_HOME = "/Users/simoncouch/instantclient_23_3")
# this one is the default, but just to prevent gotchas:
Sys.setenv(TNS_ADMIN = "/Users/simoncouch/instantclient_23_3/network/admin")
```

(These are indeed all environmental variables that could be set in `.Renviron`. Be sure to substitute `<$PATH>` carefully if you choose to do so there.)

Note that there's a [MacOS-specific bug](https://github.com/rstudio/rstudio/issues/9815) where RStudio overwrites `PATH` when set in Rprofile/Renviron before the 2024.04.0 "Chocolate Cosmos" release, so be sure to update your RStudio if needed.

In the terminal, test your Instant Client installation and database connection with:

``` shell
sqlplus sys/odbcTest321@//localhost:1521/ORCLPDB1 as sysdba
```

If you're able to connect, make a test user with:

``` sql
CREATE USER test IDENTIFIED BY odbc;
GRANT DBA TO test;
exit;
```

Test connecting with that new user with:

```
sqlplus test/odbc@//localhost:1521/ORCLPDB1
```

If that succeeds, configure a DSN in your `odbc.ini` file with:

```ini
[Oracle]
Driver = /Users/simoncouch/instantclient_23_3/libsqora.dylib.23.1
Port = 1521
Host = 0.0.0.0
```

Then, connect in R with:

``` r
library(odbc)

con <- dbConnect(
  odbc(), 
  dsn = "Oracle", 
  DBQ = "db", 
  UID = "test", 
  PWD = "odbc"
)
```

Potential gotchas:

* Setting those DBQ, UID, and PWD parameters in my `.ini` file instead of the call to `dbConnect()` introduces errors for me. 
* Note that, as of July 2024, there are no Posit Pro Drivers for Oracle on ARM Macs. 
* If you see a "file not found" error for a driver that does exist, it's because some dependency for the driver doesn't exist; use `otool -L /yada/yada/drivername.dylib` to find the dependencies and then symlink them from elsewhere on your computer into where the driver looks for them.

### On Linux

The process looks a bit different on Linux. These instructions are for a machine running Ubuntu 22.04. 

Two steps here--1) booting up an Oracle database in a docker container and 2) installing and configuring the Oracle Instant Client and driver to connect to that database.

If you haven't already, you'll need to install Docker Desktop, unixODBC, and `libaio1`. You can get the last with:

``` shell
sudo apt-get install libaio1
```

First, to boot up an Oracle database in a Docker container, run:

``` shell
docker run -d \
  --name oracle2 \
  -p 1521:1521 \
  -e ORACLE_RANDOM_PASSWORD=true \
  -e ORACLE_DATABASE=test \
  -e APP_USER=RodbcR \
  -e APP_USER_PASSWORD=Password12 \
  gvenzl/oracle-xe:21.3.0
```

Some documentation on this Docker image [here](https://hub.docker.com/r/gvenzl/oracle-xe).

Now, as for installing and configuring the Oracle Instant Client and driver, the PRs [implementing support for unit testing Oracle](https://github.com/r-dbi/odbc/commit/257f041a3aec8348000a89208da5ed0eb059fe81) and [adding support for the Oracle Pro Driver](https://github.com/r-dbi/odbc/commit/4090f0a7916db7ab22e4088040aa2492751140dd) for odbc are a great guide here. The "Installation Instructions" at the bottom of Oracle's Instant Client docs (`https://www.oracle.com/database/technologies/instant-client/linux-x86-64-downloads.html`) are thorough--you shouldn't need them here, but a good backup.

I ultimately wasn't able to make a connection with the driver that Oracle ships, but was able to do so with the Posit Pro Drivers.

Run the lines in `.github/odbc/install-oracle-driver.sh` from the first linked PR, which installs both the Instant Client and Oracle's driver.

Add the following lines to your .Rprofile (with `usethis::edit_r_profile()`): 

``` r
Sys.setenv(LD_LIBRARY_PATH = paste0("/opt/oracle/instantclient_21_12:", Sys.getenv("LD_LIBRARY_PATH")))
Sys.setenv(PATH = paste0("/opt/oracle/instantclient_21_12:", Sys.getenv("PATH")))
```

Create a text file called `/opt/oracle/instant_client_21_12/tnsnames.ora` and paste the following inside of it:

```
TEST=
   (DESCRIPTION=
    (ADDRESS=(PROTOCOL=tcp)(HOST=0.0.0.0)(PORT=1521))
    (CONNECT_DATA=(SERVICE_NAME=test)))
```

Install the RStudio Pro Oracle driver, either using the [installation code from the linked PR](https://github.com/r-dbi/odbc/commit/4090f0a7916db7ab22e4088040aa2492751140dd#diff-bc1733bfd46bbe6f651eee14b11879e4b7e0428af876e0322768bdbf138ef238R84-R87) (see the repository secret for the driver URL or contact Simon Couch or Hadley Wickham if you don't have access) or through RStudio Pro (if you have credentials, you can generate yourself a license key here: `https://connect.posit.it/support/license_management/`).

Add the following lines to your `.Renviron` (with `usethis::edit_r_environ()`). Adjust the value for `ODBCSYSINI` to the directory prefixing `.ini` files in `odbcinst -j`.

```
TNS_ADMIN=/opt/oracle/instantclient_21_12
ODBCSYSINI=/etc
ORACLE_HOME=/opt/oracle
```

Symlink your Instant Client install into the Pro Driver install directory with `sudo ln -s /opt/oracle/instantclient_21_12/* /etc/pro_drivers/oracle/bin/lib/`.

Add the following to your `odbc.ini`:

``` ini
[ProOracle]
Driver = /etc/pro_drivers/oracle/bin/lib/liboracleodbc_sb64.so
Port = 1521
Host = localhost
SVC = FREEPDB1
UID = RodbcR
PWD = Password12
DBQ = test
```

Now, putting the pieces together, connect to the Oracle Database docker container from R with:

``` r
library(odbc)
dbConnect(odbc(), dsn = "ProOracle")
```

## Snowflake

To obtain a Snowflake account:

* If you're a Posit employee, reach out to James Blair.
* If you're not affiliated with Posit, sign up for a [free trial](https://signup.snowflake.com/).

Install drivers using the [instructions on the Snowflake documentation](https://docs.snowflake.com/en/developer-guide/odbc/odbc). For MacOS, feel free to ignore sections noting that iODBC is required.

On MacOS and Linux:

* Locate your `simba.snowflake.ini` file, likely in `opt/snowflake/`. You will need `sudo` access to edit it.
  * Change `DriverManagerEncoding` to `UTF-16`.
  * Change the uncommented `ODBCInstLib` entry to the [path of your unixODBC install](https://community.snowflake.com/s/article/R-Studio-to-Snowflake-fails-with-error-Unable-to-locate-SQLGetPrivateProfileString-function).
* In the Snowflake UI, navigate to Worksheets, make a new worksheet using the blue "+" button in the top right-hand corner, and run the line `SELECT SYSTEM$ALLOWLIST()`. In the output, copy the URL corresponding to `"type":"SNOWFLAKE_DEPLOYMENT"`. This Deployment URL should look like a random string, an AWS region, and the snowflake API URL, e.g. `wmc39401.us-east-1.snowflakecomputing.com`.

The `odbc.ini` file should look like:

```ini
[Snowflake]
Driver      = Snowflake
Description =
uid         = <uid>
server      = <deployment-url>
```

The above assumes `Driver = Snowflake` is configured in `odbcinst.ini`. Then, connect by supplying the password as an argument:

```
con <- dbConnect(odbc(), dsn = "Snowflake", pwd = Sys.getenv("snowflakePass"))
```

In the above, the password has been set as `Sys.setenv(snowflakePass = "actualPass")`.

## Amazon Redshift

If you're a Posit employee, you should have access to the rstudio/warehouse GitHub repository. Follow the instructions there to get access to the internal Redshift cluster. Access to the cluster is only enabled through Posit's internal Workbench instance, where the professional drivers will already be installed.

Note that Redshift is based on a modified version of PostgreSQL.

## RODBC

We need to install the RODBC package for benchmarking in the vignette `vignette("benchmarks")`. The CRAN version of RODBC uses iODBC, so to use unixODBC we need to recompile it from source, specifying the odbc manager explicitly:

```r
install.packages("RODBC", type = "source", INSTALL_opts="--configure-args='--with-odbc-manager=odbc'")
```

---
title: "Benchmarks"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Benchmarks}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, eval = FALSE, include = FALSE}
# to run the timings in this readme, set the following env variable:
Sys.setenv(ODBC_EVAL_BENCHMARKS = "true")
```

```{r, include = FALSE}
eval_timings <- as.logical(Sys.getenv("ODBC_EVAL_BENCHMARKS", "false"))

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = eval_timings
)
```

The odbc package is often much faster than the existing
[RODBC](https://cran.r-project.org/package=RODBC) and DBI compatible
[RODBCDBI](https://cran.r-project.org/package=RODBCDBI) packages. We'll benchmark writing and reading data from the [nycflights13](https://github.com/tidyverse/nycflights13) package using the three packages.

```{r, include = FALSE}
# The CRAN version of RODBC uses iODBC, so to use unixODBC we need to 
# recompile it from source, specifying the odbc manager explicitly:
#
# install.packages("RODBC", type = "source", INSTALL_opts="--configure-args='--with-odbc-manager=odbc'")
#
# see `vignette("develop")` for more details.
```

```{r setup}
library(odbc)
library(RODBC)
library(RODBCDBI)

library(DBI)

library(nycflights13)
```

## Setup

RODBC and RODBCDBI trip up on `flights`' tibble subclass, so we'll convert it to a data frame before passing to each package. Also, the RODBCDBI package does not support writing timestamps, so we remove that column as well.

```{r}
flights$time_hour <- NULL
flights <- as.data.frame(flights)
```

Now, let's configure three driver connections to a Microsoft SQL Server database, one for each package we'll be benchmarking.

```{r, echo = FALSE}
options("connectionObserver" = NULL)
```

```{r}
odbc <- dbConnect(odbc::odbc(), dsn = "MicrosoftSQLServer", uid = "SA", pwd = "BoopBop123!")
rodbc <- RODBC::odbcConnect(dsn = "MicrosoftSQLServer", uid = "SA", pwd = "BoopBop123!")
rodbcdbi <- dbConnect(RODBCDBI::ODBC(), dsn = "MicrosoftSQLServer", user = "SA", password = "BoopBop123!")
```

The above code requires a `dsn` `MicrosoftSQLServer` to have been previously configured; see `vignettes("setup")` to learn more.

## Writing

We'll first benchmark writing the flights dataset, which contains ~300,000 rows and 18 columns, to a database.

The code for odbc and RODBCDBI looks quite similar, as they both use the DBI front-end:

```{r, include = FALSE}
if ("flights" %in% dbListTables(odbc)) {
  dbExecute(odbc, "drop table flights")
}
```

```{r}
odbc_write <- system.time(dbWriteTable(odbc, "flights", flights))

odbc_write
```

```{r, include = !eval_timings}
#>   user  system elapsed 
#>  0.883   0.176   8.108 
```

```{r, include = FALSE}
if ("flights" %in% dbListTables(odbc)) {
  dbExecute(odbc, "drop table flights")
}
```

```{r}
rodbcdbi_write <- system.time(dbWriteTable(rodbcdbi, "flights", flights))

rodbcdbi_write
```

```{r, include = !eval_timings}
#>   user  system elapsed 
#>  8.287  11.107 257.841 
```

```{r, include = FALSE}
if ("flights" %in% dbListTables(odbc)) {
  dbExecute(odbc, "drop table flights")
}
```

The timings for RODBC and RODBCDBI are quite similar, as they both utilize the RODBC back-end:

```{r}
rodbc_write <- system.time(sqlSave(rodbc, flights, "flights"))

rodbc_write
```

```{r, include = !eval_timings}
#>   user  system elapsed 
#>  8.266  11.023 235.825 
```

```{r, include = FALSE}
if ("flights" %in% dbListTables(odbc)) {
  dbExecute(odbc, "drop table flights")
}
```

## Reading

```{r, include = FALSE}
if (!"flights" %in% dbListTables(odbc)) {
  dbWriteTable(odbc, "flights", flights)
}
```

Again, the syntax for odbc and RODBCDBI is identical, while the timings for RODBCDBI and RODBC are quite similar.

```{r}
odbc_read     <- system.time(result <- dbReadTable(odbc, "flights"))
rodbcdbi_read <- system.time(result <- dbReadTable(rodbcdbi, "flights"))
rodbc_read    <- system.time(result <- sqlFetch(rodbc, "flights"))
```

```{r}
odbc_read
```

```{r, include = !eval_timings}
#>   user  system elapsed 
#>  0.515   0.024   0.557 
```

```{r}
rodbcdbi_read
```

```{r, include = !eval_timings}
#>   user  system elapsed 
#>  1.308   0.035   1.356 
```

```{r}
rodbc_read
```

```{r, include = !eval_timings}
#>   user  system elapsed 
#>  1.291   0.033   1.343 
```

# dbWriteTable errors if field.types don't exist (#271)

    Code
      sqlCreateTable(con, "foo", iris, field.types = list(bar = "[int]"))
    Condition
      Warning:
      Some columns in `field.types` not in the input, missing columns:
        - 'bar'
    Output
      <SQL> CREATE TABLE "foo" (
        "Sepal.Length" FLOAT,
        "Sepal.Width" FLOAT,
        "Petal.Length" FLOAT,
        "Petal.Width" FLOAT,
        "Species" varchar(255),
        "bar" [int]
      )
      

# can create / write to temp table

    `temporary` is `TRUE`, but table name doesn't use # prefix.

---

    `temporary` is `TRUE`, but table name doesn't use # prefix.

# odbcListConfig errors informatively without unixODBC

    Code
      odbcListConfig()
    Condition
      Error in `odbcListConfig()`:
      ! The unixODBC driver manager is not available. 
      * Please install and try again.

# odbcListConfig errors informatively with unexpected odbcinst output

    Code
      odbcListConfig()
    Condition
      Error in `odbcListConfig()`:
      ! Failed to parse output from odbcinst.
      i This is an internal error that was detected in the odbc package.
        Please report it at <https://github.com/r-dbi/odbc/issues> with a reprex (<https://tidyverse.org/help/>) and the full backtrace.

---

    Code
      odbcListConfig()
    Condition
      Error in `odbcListConfig()`:
      ! Failed to parse output from odbcinst.
      i This is an internal error that was detected in the odbc package.
        Please report it at <https://github.com/r-dbi/odbc/issues> with a reprex (<https://tidyverse.org/help/>) and the full backtrace.

# odbcEdit*() errors informatively on Windows

    Code
      odbcEditDrivers()
    Condition
      Error in `odbcEditDrivers()`:
      ! Configuration files are only available on macOS and Linux.
      i Use the ODBC Data Source Administrator to configure connections on Windows.

# an account ID is required

    Code
      snowflake_args(driver = "driver")
    Condition
      Error in `snowflakeauth::snowflake_connection()`:
      ! An `account` parameter is required when '<tmp>/connections.toml' is missing or empty.
      i Pass `account` or define a [default] section with an account field in '<tmp>/connections.toml'.

# both 'uid' and 'pwd' are required when present

    Code
      snowflake_args(account = "testorg-test_account", driver = "driver", uid = "uid",
        )
    Condition
      Error in `DBI::dbConnect()`:
      ! Both `uid` and `pwd` must be specified to authenticate.
      i Or leave both unset to use ambient Snowflake credentials.

# we error if we can't find ambient credentials

    Code
      snowflake_args(account = "testorg-test_account", driver = "driver")
    Condition
      Error in `DBI::dbConnect()`:
      ! Failed to detect ambient Snowflake credentials.
      i Supply `uid` and `pwd` to authenticate manually.

# we hint viewer-based credentials on Connect

    Code
      snowflake_args(account = "testorg-test_account", driver = "driver")
    Condition
      Error in `DBI::dbConnect()`:
      ! Failed to detect ambient Snowflake credentials.
      i Supply `uid` and `pwd` to authenticate manually.
      i Or consider enabling Posit Connect's Snowflake integration for viewer-based credentials. See <https://docs.posit.co/connect/user/oauth-integrations/#adding-oauth-integrations-to-deployed-content> for details.

# we error if we can't find the driver

    Code
      snowflake_default_driver()
    Condition
      Error in `DBI::dbConnect()`:
      ! Failed to automatically find the Snowflake ODBC driver.
      i Set `driver` to known driver name or path.

# Workbench-managed credentials are rejected for other accounts

    Code
      snowflake_args(account = "testorg-test_account", driver = "driver")
    Condition
      Error in `DBI::dbConnect()`:
      ! Snowflake account mismatch: requested "testorg-test_account" but connection "workbench" resolved "nonmatching".
      i Check your 'connections.toml' or pass the correct `account`.

# dbConnect() errors informatively without unixODBC (#782)

    Code
      test_con("SQLITE")
    Condition
      Error in `dbConnect()`:
      ! Unable to locate the unixODBC driver manager.
      i Please install unixODBC using Homebrew with `brew install unixodbc`.

# parse_size works

    Code
      parse_size("foo")
    Condition
      Error:
      ! `"foo"` must be a whole number, not the string "foo".

# id_field checks inputs

    Code
      id_field(DBI::Id("a"), "foo")
    Condition
      Error:
      ! `field` must be one of "catalog", "schema", or "table", not "foo".
    Code
      id_field(DBI::Id("a", "b", "c", "d"))
    Condition
      Error:
      ! Identifier must be length 1, 2, or 3.

# errors are rethrown informatively (#643, #788)

    Code
      dbConnect(odbc(), dsn = "does_not_exist_db")
    Condition
      Error in `dbConnect()`:
      ! ODBC failed with error 00000 from [unixODBC][Driver Manager].
      x Data source name not found and no default driver specified
      i See `?odbc::odbcListDataSources()` to learn more.
      i From 'nanodbc/nanodbc.cpp:1184'.

---

    Code
      dbExecute(con, "SELECT * FROM boopbopbopbeep")
    Condition
      Error in `dbExecute()`:
      ! ODBC failed with error 00000 from [SQLite].
      x no such table: boopbopbopbeep (1)
      * <SQL> 'SELECT * FROM boopbopbopbeep'
      i From 'nanodbc/nanodbc.cpp:1802'.

# rethrow_database_error() errors well when parse_database_error() fails

    Code
      rethrow_database_error("boop", call = NULL)
    Condition
      Error:
      ! boop

# parse_database_error() works with messages from the wild

    Code
      rethrow_database_error(msg, call = NULL)
    Condition
      Error:
      ! ODBC failed with error 00000 from [unixODBC][Driver Manager].
      x  Data source name not found and no default driver specified
      i See `?odbc::odbcListDataSources()` to learn more.
      i From 'nanodbc/nanodbc.cpp:1135'.

---

    Code
      rethrow_database_error(msg, call = NULL)
    Condition
      Error:
      ! ODBC failed with error 07002 from [Microsoft][ODBC Driver 17 for SQL Server].
      x  COUNT field incorrect or syntax error
      i From 'nanodbc/nanodbc.cpp:1594'.

---

    Code
      rethrow_database_error(msg, call = NULL)
    Condition
      Error:
      ! ODBC failed with error 07002 from [ODBC Firebird Driver].
      x  COUNT field incorrect
      i From 'nanodbc/nanodbc.cpp:1710'.

---

    Code
      rethrow_database_error(msg, call = NULL)
    Condition
      Error:
      ! ODBC failed with error HYT00 from [Microsoft][SQL Server Native Client 11.0].
      x  Query timeout expired
      i From 'nanodbc/nanodbc.cpp:1655'.

---

    Code
      rethrow_database_error(msg, call = NULL)
    Condition
      Error:
      ! ODBC failed with error 00000 from [Microsoft][ODBC Driver 18 for SQL Server].
      x  Login timeout expired
      *  TCP Provider: Error code 0x2726
      *  A network-related or instance-specific error has occurred while establishing a connection to 127.0.0.1. Server is not found or not accessible. Check if instance name is correct and if SQL Server is configured to allow remote connections. For more information see SQL Server Books Online.
      i From 'nanodbc/nanodbc.cpp:1147'.

---

    Code
      rethrow_database_error(msg, call = NULL)
    Condition
      Error:
      ! ODBC failed with error S1000 from [SAP AG][LIBODBCHDB DLL][HDBODBC][SAP AG][LIBODBCHDB SO][HDBODBC][89013].
      x  General error;403 internal error: Error opening the cursor for the remote database <***.***> Connection not open;-10807 Connection down: Socket closed by peer {***.**.*.**:***** -> ***.**.***.**:***** TenantName:(none) SiteVolumeID:1:3 SiteType:PRIMARY ConnectionID:****** SessionID:************}
      *  <SQL> 'SELECT DISTINCT "po_id", ***CENSORED***'
      i From 'nanodbc/nanodbc.cpp:1722'.

---

    Code
      rethrow_database_error(msg, call = NULL)
    Condition
      Error:
      ! `parse_database_error()` will not {be able to parse this}, but it should still be successfully rethrown as-is.

# check_row.names()

    Code
      dbWriteTable(con, "boopery", data.frame(bop = 1), row.names = c("no", "way"))
    Condition
      Error in `dbWriteTable()`:
      ! `row.names` must be `NULL`, `TRUE`, `FALSE`, `NA`, or a single string, not a character vector.

# check_field.types()

    Code
      dbWriteTable(con, "boopery", data.frame(bop = 1), field.types = "numeric")
    Condition
      Error in `dbWriteTable()`:
      ! `field.types` must be `NULL` or a named vector of field types, not a string.

# check_attributes()

    Code
      con <- test_con("SQLITE", attributes = list(boop = "bop"))
    Condition
      Error in `dbConnect()`:
      ! `attributes` does not support the connection attribute "boop".
      i Allowed connection attributes are "azure_token", "sf_private_key", and "sf_private_key_password".

---

    Code
      con <- test_con("SQLITE", attributes = list(boop = "bop", beep = "boop"))
    Condition
      Error in `dbConnect()`:
      ! `attributes` does not support the connection attributes "boop" and "beep".
      i Allowed connection attributes are "azure_token", "sf_private_key", and "sf_private_key_password".

# configure_simba() errors informatively on failure to install unixODBC

    Code
      configure_simba()
    Condition
      Error:
      ! Unable to locate the unixODBC driver manager.
      i Please install unixODBC using Homebrew with `brew install unixodbc`.

# databricks() errors informatively when spark ini isn't writeable

    Code
      write_simba_lines("", ".", ".", call2("databricks"))
    Condition
      Error in `databricks()`:
      ! Detected needed changes to the driver configuration file at ., but the file was not writeable.
      i Please make the changes outlined at https://solutions.posit.co/connections/db/databases/databricks/#troubleshooting-apple-macos-users.

# configure_unixodbc_simba() writes reasonable entries

    Code
      configure_unixodbc_simba(unixodbc_install = unixodbc_install_path,
        simba_config = spark_config_path, action = "warn")
    Condition
      Warning:
      i Detected potentially unsafe driver settings:
      * Please consider revising the `ODBCInstLib` field in 'simba.sparkodbc.ini' and setting its value to "libodbcinst.dylib"
      * Please consider revising the `DriverManagerEncoding` field in 'simba.sparkodbc.ini' and setting its value to "UTF-16".

---

    Code
      configure_unixodbc_simba(unixodbc_install = unixodbc_install_path,
        simba_config = spark_config_path, action = "warn")
    Condition
      Warning:
      i Detected potentially unsafe driver settings:
      * Please consider revising the `ODBCInstLib` field in 'simba.sparkodbc.ini' and setting its value to "libodbcinst.dylib"
      * Please consider revising the `DriverManagerEncoding` field in 'simba.sparkodbc.ini' and setting its value to "UTF-16".

---

    Code
      configure_unixodbc_simba(unixodbc_install = unixodbc_install_path,
        simba_config = spark_config_path, action = "warn")
    Condition
      Warning:
      i Detected potentially unsafe driver settings:
      * Please consider revising the `DriverManagerEncoding` field in 'simba.sparkodbc.ini' and setting its value to "UTF-16".

---

    Code
      configure_unixodbc_simba(unixodbc_install = unixodbc_install_path,
        simba_config = spark_config_path, action = "warn")
    Condition
      Warning:
      i Detected potentially unsafe driver settings:
      * Please consider revising the `ODBCInstLib` field in 'simba.sparkodbc.ini' and setting its value to "libodbcinst.dylib"
      * Please consider revising the `DriverManagerEncoding` field in 'simba.sparkodbc.ini' and setting its value to "UTF-16".

# errors if unnamed arguments

    Code
      check_args(list(1, 2, 3))
    Condition
      Error in `DBI::dbConnect()`:
      ! All elements of ... must be named.

# errors about case-insensitve arguments

    Code
      check_args(list(xa = 1, xA = 1))
    Condition
      Error in `DBI::dbConnect()`:
      ! After ignoring case, some arguments have the same name:
      * xa, xA
    Code
      check_args(list(xa = 1, xA = 1, XA = 1))
    Condition
      Error in `DBI::dbConnect()`:
      ! After ignoring case, some arguments have the same name:
      * xa, xA, XA
    Code
      check_args(list(xa = 1, xA = 1, xb = 1, xB = 1))
    Condition
      Error in `DBI::dbConnect()`:
      ! After ignoring case, some arguments have the same name:
      * xa, xA
      * xb, xB

# messages if values might need quoting

    Code
      check_quoting(list(foo = "f{oo", bar = "b{ar", baz = "baz"))
    Message
      `foo`, `bar` contains a special character that may need quoting.
      i Wrap the value in `odbc::quote_value()` to use a heuristic that should work for most drivers.
      i If that still doesn't work, consult your driver's documentation.
      i Otherwise, you can suppress this message by wrapping the value in `I()`.

# automatically picks correct quote type

    Code
      quote_value("'\"")
    Condition
      Error in `quote_value()`:
      ! Don't know how to escape a value with both single and double quotes.

# validateObjectName() errors informatively

    Code
      odbcListColumns(con, table = "boop", view = "bop")
    Condition
      Error in `odbcListColumns()`:
      ! Exactly one of `table` or `view` must be supplied.

---

    Code
      odbcListColumns(con)
    Condition
      Error in `odbcListColumns()`:
      ! One of `table` or `view` must be supplied.

# errors if can't find driver

    Code
      databricks_default_driver()
    Condition
      Error in `DBI::dbConnect()`:
      ! Failed to automatically find the Databricks/Spark ODBC driver.
      i Set `driver` to known driver name or path.

# databricks host validates inputs

    Code
      databricks_host("")
    Condition
      Error in `DBI::dbConnect()`:
      ! No Databricks workspace URL provided.
      i Either supply `workspace` argument or set env var `DATABRICKS_HOST`.

# errors if auth fails

    Code
      . <- databricks_args1()
    Condition
      Error in `DBI::dbConnect()`:
      ! Failed to detect ambient Databricks credentials.
      i Supply `uid` and `pwd` to authenticate manually.

# must supply both uid and pwd

    Code
      databricks_auth_args("host", uid = "uid")
    Condition
      Error in `DBI::dbConnect()`:
      ! Both `uid` and `pwd` must be specified for manual authentication.
      i Or leave both unset for automated authentication.

# dbConnect method errors informatively re: httpPath (#787)

    Code
      dbConnect(databricks(), httpPath = "boop", HTTPPath = "bop")
    Condition
      Error in `dbConnect()`:
      ! Exactly one of `httpPath` or `HTTPPath` must be supplied.

---

    Code
      dbConnect(databricks(), HTTPPath = 1L)
    Condition
      Error in `dbConnect()`:
      ! `HTTPPath` must be a single string or `NULL`, not the number 1.

---

    Code
      dbConnect(databricks(), httpPath = 1L)
    Condition
      Error in `dbConnect()`:
      ! `httpPath` must be a single string or `NULL`, not the number 1.

# we hint viewer-based and service principal credentials on Connect

    Code
      databricks_args(workspace = "workspace", httpPath = "path", driver = "driver")
    Condition
      Error in `DBI::dbConnect()`:
      ! Failed to detect ambient Databricks credentials.
      i Supply `uid` and `pwd` to authenticate manually.
      i Or consider enabling Posit Connect's Databricks integration.  For viewer-based credentials, see <https://docs.posit.co/connect/user/oauth-integrations/#viewer-oauth-integrations> for details.  For service principal credentials, see <https://docs.posit.co/connect/user/oauth-integrations/#service-account-oauth-integrations> for details.

# workload identity errors when DATABRICKS_CLIENT_ID is not set

    Code
      databricks_auth_args("host")
    Condition
      Error in `DBI::dbConnect()`:
      ! Workload identity federation requires a service principal.
      i Set the `DATABRICKS_CLIENT_ID` environment variable to the service principal's UUID.

# workload identity env-oidc errors when JWT env var is empty

    Code
      databricks_auth_args("host")
    Condition
      Error in `DBI::dbConnect()`:
      ! Workload identity federation is enabled (`DATABRICKS_AUTH_TYPE=env-oidc`) but no OIDC token was found.
      i Set the `DATABRICKS_OIDC_TOKEN` environment variable to the IdP JWT.

# workload identity file-oidc errors when filepath is not set

    Code
      databricks_auth_args("host")
    Condition
      Error in `DBI::dbConnect()`:
      ! Workload identity federation is enabled (`DATABRICKS_AUTH_TYPE=file-oidc`) but `DATABRICKS_OIDC_TOKEN_FILEPATH` is not set.
      i Set `DATABRICKS_OIDC_TOKEN_FILEPATH` to the path of the file containing the IdP JWT.

# workload identity file-oidc errors when file does not exist

    Code
      databricks_auth_args("host")
    Condition
      Error in `DBI::dbConnect()`:
      ! OIDC token file '/nonexistent/path/to/jwt' does not exist.
      i Verify `DATABRICKS_OIDC_TOKEN_FILEPATH` points to a readable file.

# workload identity file-oidc errors when file is empty

    Code
      databricks_auth_args("host")
    Condition
      Error in `DBI::dbConnect()`:
      ! OIDC token file '<tempfile>' is empty.

# databricks_token_exchange errors on HTTP failure

    Code
      databricks_token_exchange("host.example.com", "bad-jwt", "sp-id")
    Condition
      Error in `DBI::dbConnect()`:
      ! Databricks OIDC token exchange failed with HTTP 400.
      i Response: {"error":"invalid_grant"}

# databricks_token_exchange errors when access_token is missing

    Code
      databricks_token_exchange("host.example.com", "jwt", "sp-id")
    Condition
      Error in `DBI::dbConnect()`:
      ! Databricks OIDC token exchange response did not contain an access token.
      i Verify the federation policy is configured correctly in Databricks.

# show method prints only host if no user is defined

    Code
      con
    Output
      <OdbcConnection> localhost

# show method prints DISCONNECTED if not valid

    Code
      con
    Output
      <OdbcConnection> localhost
        DISCONNECTED

# show method does not print server if it is not available

    Code
      con
    Output
      <OdbcConnection>

# dbQuoteIdentifier() errors informatively

    Code
      dbQuoteIdentifier(con, NA_character_)
    Condition
      Error in `dbQuoteIdentifier()`:
      ! `x` can't be `NA`.

# unsupported types gives informative error

    Code
      dbWriteTable(con, "df", df)
    Condition
      Error in `odbcDataType()`:
      ! Can't determine type for column foo.
      Caused by error:
      ! Unsupported type

# dbWriteTable() with `field.types` with `append = TRUE`

    Code
      dbWriteTable(con, "boopery", data.frame(bop = 1), field.types = c(bop = "numeric"),
      append = TRUE)
    Condition
      Error in `dbWriteTable()`:
      ! Cannot specify `field.types` with `append = TRUE`.

# WriteTable() with `overwrite = TRUE` and `append = TRUE`

    Code
      dbWriteTable(con, "boopery", data.frame(bop = 1), overwrite = TRUE, append = TRUE)
    Condition
      Error in `dbWriteTable()`:
      ! `overwrite` and `append` cannot both be "TRUE".

# dbWriteTable(), existing table, `overwrite = FALSE`, `append = FALSE`

    Code
      dbWriteTable(con, "boopery", data.frame(bop = 1), overwrite = FALSE, append = FALSE)
    Condition
      Error in `dbWriteTable()`:
      ! Table boopery exists in database, and both overwrite and append are `FALSE`.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odbc-config.R
\name{odbcListConfig}
\alias{odbcListConfig}
\alias{odbcEditDrivers}
\alias{odbcEditSystemDSN}
\alias{odbcEditUserDSN}
\title{List locations of ODBC configuration files}
\usage{
odbcListConfig()

odbcEditDrivers()

odbcEditSystemDSN()

odbcEditUserDSN()
}
\description{
On MacOS and Linux, odbc uses the unixODBC driver manager to manage
information about driver and data sources. This helper returns the filepaths
where the driver manager will look for that information.

\code{odbcListConfig()} is a wrapper around the command line call \code{odbcinst -j}.
The \code{\link[=odbcEditDrivers]{odbcEditDrivers()}}, \code{\link[=odbcEditSystemDSN]{odbcEditSystemDSN()}}, and
\code{\link[=odbcEditUserDSN]{odbcEditUserDSN()}} helpers provide a shorthand for
\code{file.edit(odbcListConfig()[[i]])}.

Windows does not use \code{.ini} configuration files; on Windows,
\code{odbcListConfig()} will return a 0-length vector and \verb{odbcEdit*()}
will raise an error.
}
\examples{
\dontshow{if (FALSE) withAutoprint(\{ # examplesIf}
configs <- odbcListConfig()

configs

# shorthand for file.edit(configs[[1]])
odbcEditDrivers()
\dontshow{\}) # examplesIf}
}
\seealso{
The \code{\link[=odbcListDrivers]{odbcListDrivers()}} and \code{\link[=odbcListDataSources]{odbcListDataSources()}} helpers return
information on the contents of \code{odbcinst.ini} and \code{odbc.ini} files,
respectively. \code{\link[=odbcListDataSources]{odbcListDataSources()}} collates entries from both the
System and User \code{odbc.ini} files.

Learn more about unixODBC and the \code{odbcinst} utility
\href{https://www.unixodbc.org/odbcinst.html}{here}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connection-pane.R
\name{odbcListObjects}
\alias{odbcListObjects}
\title{List objects in a connection.}
\usage{
odbcListObjects(connection, ...)
}
\arguments{
\item{connection}{A connection object, as returned by \code{dbConnect()}.}

\item{...}{Attributes to filter by.}
}
\value{
A data frame with \code{name} and \code{type} columns, listing the
objects.
}
\description{
Lists all of the objects in the connection, or all the objects which have
specific attributes.
}
\details{
When used without parameters, this function returns all of the objects known
by the connection. Any parameters passed will filter the list to only objects
which have the given attributes; for instance, passing \code{schema = "foo"}
will return only objects matching the schema \code{foo}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db.R
\name{isTempTable}
\alias{isTempTable}
\alias{isTempTable,OdbcConnection,Id-method}
\alias{isTempTable,OdbcConnection,SQL-method}
\title{Helper method used to determine if a table identifier is that
of a temporary table.}
\usage{
isTempTable(conn, name, ...)

\S4method{isTempTable}{OdbcConnection,Id}(conn, name, ...)

\S4method{isTempTable}{OdbcConnection,SQL}(conn, name, ...)
}
\arguments{
\item{conn}{OdbcConnection}

\item{name}{Table name}

\item{...}{additional parameters to methods}
}
\description{
Currently implemented only for select back-ends where
we have a use for it (SQL Server, for example).  Generic, in case
we develop a broader use case.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa-odbc-data-type.R, R/driver-access.R,
%   R/driver-bigquery.R, R/driver-hive.R, R/driver-impala.R, R/driver-mysql.R,
%   R/driver-netezza.R, R/driver-oracle.R, R/driver-postgres.R,
%   R/driver-redshift.R, R/driver-snowflake.R, R/driver-spark.R,
%   R/driver-sql-server.R, R/driver-sqlite.R, R/driver-teradata.R,
%   R/driver-vertica.R
\name{odbcDataType}
\alias{odbcDataType}
\alias{odbcDataType,ANY-method}
\alias{odbcDataType,ACCESS-method}
\alias{odbcDataType,BigQuery-method}
\alias{odbcDataType,Hive-method}
\alias{odbcDataType,Impala-method}
\alias{odbcDataType,MySQL-method}
\alias{odbcDataType,NetezzaSQL-method}
\alias{odbcDataType,Oracle-method}
\alias{odbcDataType,PostgreSQL-method}
\alias{odbcDataType,Redshift-method}
\alias{odbcDataType,Snowflake-method}
\alias{odbcDataType,Spark SQL-method}
\alias{odbcDataType,Microsoft SQL Server-method}
\alias{odbcDataType,SQLite-method}
\alias{odbcDataType,Teradata-method}
\alias{odbcDataType,Vertica Database-method}
\title{Return the corresponding ODBC data type for an R object}
\usage{
odbcDataType(con, obj, ...)
}
\arguments{
\item{con}{A driver connection object, as returned by \code{dbConnect()}.}

\item{obj}{An R object.}

\item{...}{Additional arguments passed to methods.}
}
\value{
Corresponding SQL type for the \code{obj}.
}
\description{
This is used when creating a new table with \code{dbWriteTable()}.
Databases with default methods defined are:
\itemize{
\item MySQL
\item PostgreSQL
\item SQL Server
\item Oracle
\item SQLite
\item Spark
\item Hive
\item Impala
\item Redshift
\item Vertica
\item BigQuery
\item Teradata
\item Access
\item Snowflake
}
}
\details{
If you are using a different database and \code{dbWriteTable()} fails with a SQL
parsing error the default method is not appropriate, you will need to write
a new method. The object type for your method will be the database name
retrieved by \code{dbGetInfo(con)$dbms.name}. Use the documentation provided with
your database to determine appropriate values for each R data type.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connection-pane.R
\name{odbcConnectionActions}
\alias{odbcConnectionActions}
\title{List the actions supported for the connection}
\usage{
odbcConnectionActions(connection)
}
\arguments{
\item{connection}{A connection object, as returned by \code{dbConnect()}.}
}
\value{
A named list of actions that can be performed on the connection.
}
\description{
Return a list of actions that can be performed on the connection.
}
\details{
The list returned is a named list of actions, where each action has the
following properties:

\describe{
\item{callback}{A function to be invoked to perform the action}
\item{icon}{An optional path to an icon representing the action}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbi-table.R, R/dbi.R, R/driver-db2.R,
%   R/driver-hana.R, R/driver-hive.R, R/driver-teradata.R
\name{sqlData,OdbcConnection-method}
\alias{sqlData,OdbcConnection-method}
\alias{DBI-methods}
\alias{sqlCreateTable,DB2/AIX64-method}
\alias{odbcConnectionTables,DB2/AIX64,character-method}
\alias{sqlCreateTable,HDB-method}
\alias{dbQuoteString,Hive,character-method}
\alias{sqlCreateTable,Teradata-method}
\title{Unimportant DBI methods}
\usage{
\S4method{sqlData}{OdbcConnection}(con, value, row.names = NA, ...)

\S4method{sqlCreateTable}{HDB}(
  con,
  table,
  fields,
  row.names = NA,
  temporary = FALSE,
  ...,
  field.types = NULL
)

\S4method{dbQuoteString}{Hive,character}(conn, x, ...)

\S4method{sqlCreateTable}{Teradata}(
  con,
  table,
  fields,
  row.names = NA,
  temporary = FALSE,
  ...,
  field.types = NULL
)
}
\description{
Unimportant DBI methods
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbi-connection.R, R/odbc-connection.R,
%   R/dbi-table.R
\docType{class}
\name{OdbcConnection-class}
\alias{OdbcConnection-class}
\alias{show,OdbcConnection-method}
\alias{dbIsValid,OdbcConnection-method}
\alias{dbDisconnect,OdbcConnection-method}
\alias{dbSendQuery,OdbcConnection,character-method}
\alias{dbExecute,OdbcConnection,character-method}
\alias{dbSendStatement,OdbcConnection,character-method}
\alias{dbDataType,OdbcConnection,ANY-method}
\alias{dbDataType,OdbcConnection,data.frame-method}
\alias{dbQuoteIdentifier,OdbcConnection,character-method}
\alias{dbQuoteIdentifier,OdbcConnection,SQL-method}
\alias{dbGetInfo,OdbcConnection-method}
\alias{dbGetQuery,OdbcConnection,character-method}
\alias{dbBegin,OdbcConnection-method}
\alias{dbCommit,OdbcConnection-method}
\alias{dbRollback,OdbcConnection-method}
\alias{OdbcConnection}
\alias{dbListFields,OdbcConnection,Id-method}
\alias{dbListFields,OdbcConnection,SQL-method}
\alias{dbListFields,OdbcConnection,character-method}
\alias{dbExistsTable,OdbcConnection,Id-method}
\alias{dbExistsTable,OdbcConnection,SQL-method}
\alias{dbExistsTable,OdbcConnection,character-method}
\alias{dbRemoveTable,OdbcConnection,character-method}
\title{Odbc Connection Methods}
\usage{
\S4method{show}{OdbcConnection}(object)

\S4method{dbIsValid}{OdbcConnection}(dbObj, ...)

\S4method{dbDisconnect}{OdbcConnection}(conn, ...)

\S4method{dbSendQuery}{OdbcConnection,character}(conn, statement, params = NULL, ..., immediate = FALSE)

\S4method{dbExecute}{OdbcConnection,character}(conn, statement, params = NULL, ..., immediate = is.null(params))

\S4method{dbSendStatement}{OdbcConnection,character}(conn, statement, params = NULL, ..., immediate = FALSE)

\S4method{dbDataType}{OdbcConnection,ANY}(dbObj, obj, ...)

\S4method{dbDataType}{OdbcConnection,data.frame}(dbObj, obj, ...)

\S4method{dbQuoteIdentifier}{OdbcConnection,character}(conn, x, ...)

\S4method{dbQuoteIdentifier}{OdbcConnection,SQL}(conn, x, ...)

\S4method{dbGetInfo}{OdbcConnection}(dbObj, ...)

\S4method{dbGetQuery}{OdbcConnection,character}(
  conn,
  statement,
  n = -1,
  params = NULL,
  immediate = is.null(params),
  ...
)

\S4method{dbBegin}{OdbcConnection}(conn, ...)

\S4method{dbCommit}{OdbcConnection}(conn, ...)

\S4method{dbRollback}{OdbcConnection}(conn, ...)

\S4method{dbListFields}{OdbcConnection,Id}(conn, name, ...)

\S4method{dbListFields}{OdbcConnection,SQL}(conn, name, ...)

\S4method{dbListFields}{OdbcConnection,character}(
  conn,
  name,
  catalog_name = NULL,
  schema_name = NULL,
  column_name = NULL,
  ...
)

\S4method{dbExistsTable}{OdbcConnection,Id}(conn, name, ...)

\S4method{dbExistsTable}{OdbcConnection,SQL}(conn, name, ...)

\S4method{dbExistsTable}{OdbcConnection,character}(conn, name, ...)

\S4method{dbRemoveTable}{OdbcConnection,character}(conn, name, ...)
}
\arguments{
\item{object}{Any R object}

\item{dbObj}{An object inheriting from \code{DBIObject}, i.e. \code{DBIDriver},
\code{DBIConnection}, or a \code{DBIResult}.}

\item{...}{Other arguments to methods.}

\item{conn}{A \link[DBI:DBIConnection-class]{DBI::DBIConnection} object, as returned by
\code{dbConnect()}.}

\item{statement}{a character string containing SQL.}

\item{params}{Query parameters to pass to \code{\link[DBI:dbBind]{DBI::dbBind()}}.
See \code{\link[DBI:dbBind]{DBI::dbBind()}} for details.}

\item{immediate}{If \code{TRUE}, SQLExecDirect will be used instead of
SQLPrepare, and the \code{params} argument is ignored}

\item{obj}{An R object whose SQL type we want to determine.}

\item{x}{A character vector, \link[DBI]{SQL} or \link[DBI]{Id} object to quote as identifier.}

\item{n}{maximum number of records to retrieve per fetch. Use \code{n = -1}
or \code{n = Inf}
to retrieve all pending records.  Some implementations may recognize other
special values.}

\item{name}{The table name, passed on to \code{\link[DBI:dbQuoteIdentifier]{dbQuoteIdentifier()}}. Options are:
\itemize{
\item a character string with the unquoted DBMS table name,
e.g. \code{"table_name"},
\item a call to \code{\link[DBI:Id]{Id()}} with components to the fully qualified table name,
e.g. \code{Id(schema = "my_schema", table = "table_name")}
\item a call to \code{\link[DBI:SQL]{SQL()}} with the quoted and fully qualified table name
given verbatim, e.g. \code{SQL('"my_schema"."table_name"')}
}}

\item{catalog_name}{Catalog where table is located.}

\item{schema_name}{Schema where table is located.}

\item{column_name}{The name of the column to return, the default returns all columns.}
}
\description{
Implementations of pure virtual functions defined in the \code{DBI} package
for OdbcConnection objects.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbi.R, R/driver-access.R, R/driver-bigquery.R,
%   R/driver-db2.R, R/driver-hana.R, R/driver-hive.R, R/driver-impala.R,
%   R/driver-mysql.R, R/driver-netezza.R, R/driver-postgres.R,
%   R/driver-redshift.R, R/driver-snowflake.R, R/driver-spark.R,
%   R/driver-sqlite.R, R/driver-teradata.R, R/driver-vertica.R
\docType{class}
\name{DBI-classes}
\alias{DBI-classes}
\alias{ACCESS-class}
\alias{BigQuery-class}
\alias{DB2/AIX64-class}
\alias{HDB-class}
\alias{Hive-class}
\alias{Impala-class}
\alias{MySQL-class}
\alias{NetezzaSQL-class}
\alias{PostgreSQL-class}
\alias{Redshift-class}
\alias{Snowflake-class}
\alias{Spark SQL-class}
\alias{SQLite-class}
\alias{Teradata-class}
\alias{Vertica Database-class}
\title{DBI classes}
\description{
DBI classes
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odbc-data-sources.R
\name{odbcListDataSources}
\alias{odbcListDataSources}
\title{List Configured Data Source Names}
\usage{
odbcListDataSources()
}
\value{
A data frame with two columns:
\describe{
\item{name}{Name of the data source. The entries in this column can be
passed to the \code{dsn} argument of \code{\link[=dbConnect]{dbConnect()}}.}
\item{description}{Data source description.}
}
}
\description{
Collect information about the available data source names (DSNs). A DSN must
be both installed and configured with the driver manager to be included in
this list. Configuring a DSN just sets up a lookup table (e.g. in
\code{odbc.ini}) to allow users to pass only the DSN to \code{\link[=dbConnect]{dbConnect()}}.

DSNs that are not configured with the driver manager can still be
connected to with \code{\link[=dbConnect]{dbConnect()}} by providing DSN metadata directly.
}
\section{Configuration}{


This function interfaces with the driver manager to collect information
about the available data source names.

For \strong{MacOS and Linux}, the odbc package supports the unixODBC driver
manager. unixODBC looks to the \code{odbc.ini} \emph{configuration file} for information
on DSNs. Find the location(s) of your \code{odbc.ini} file(s) with \code{odbcinst -j}.

In this example \code{odbc.ini} file:

\if{html}{\out{<div class="sourceCode">}}\preformatted{[MySQL]
Driver = MySQL Driver
Database = test
Server = 127.0.0.1
User = root
password = root
Port = 3306
}\if{html}{\out{</div>}}

...the data source name is \code{MySQL}, which will appear in the \code{name}
column of this function's output. To pass the DSN as the \code{dsn} argument to
\code{\link[=dbConnect]{dbConnect()}}, pass it as a string, like \code{"MySQL"}.
\verb{Driver = MySQL Driver} references the driver \code{name} in \code{\link[=odbcListDrivers]{odbcListDrivers()}}
output.

\strong{Windows} is \href{https://learn.microsoft.com/en-us/sql/odbc/admin/odbc-data-source-administrator}{bundled}
with an ODBC driver manager.

When a DSN is configured with a driver manager, information on the DSN will
be automatically passed on to \code{\link[=dbConnect]{dbConnect()}} when its \code{dsn} argument is set.

For example, with the \code{MySQL} data source name configured, and the driver
name \verb{MySQL Driver} appearing in \code{\link[=odbcListDrivers]{odbcListDrivers()}} output, the code:

\if{html}{\out{<div class="sourceCode">}}\preformatted{con <-
  dbConnect(
    odbc::odbc(),
    Driver = "MySQL Driver",
    Database = "test",
    Server = "127.0.0.1",
    User = "root",
    password = "root",
    Port = 3306
  )
}\if{html}{\out{</div>}}

...can be written:

\if{html}{\out{<div class="sourceCode">}}\preformatted{con <- dbConnect(odbc::odbc(), dsn = "MySQL")
}\if{html}{\out{</div>}}

In this case, \code{dbConnect()} will look up the information defined for \code{MySQL}
in the driver manager (in our example, \code{odbc.ini}) and automatically
pass the needed arguments.
}

\seealso{
\code{\link[=odbcListDrivers]{odbcListDrivers()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/driver-snowflake.R
\docType{class}
\name{snowflake}
\alias{snowflake}
\alias{SnowflakeOdbcDriver-class}
\alias{dbConnect,SnowflakeOdbcDriver-method}
\title{Helper for connecting to Snowflake via ODBC}
\usage{
snowflake()

\S4method{dbConnect}{SnowflakeOdbcDriver}(
  drv,
  account = Sys.getenv("SNOWFLAKE_ACCOUNT"),
  driver = NULL,
  warehouse = NULL,
  database = NULL,
  schema = NULL,
  uid = NULL,
  pwd = NULL,
  connection_name = NULL,
  ...
)
}
\arguments{
\item{drv}{an object that inherits from \link[DBI:DBIDriver-class]{DBI::DBIDriver},
or an existing \link[DBI:DBIConnection-class]{DBI::DBIConnection}
object (in order to clone an existing connection).}

\item{account}{A Snowflake \href{https://docs.snowflake.com/en/user-guide/admin-account-identifier}{account identifier},
e.g. \code{"testorg-test_account"}.}

\item{driver}{The name of the Snowflake ODBC driver, or \code{NULL} to use the
default name.}

\item{warehouse}{The name of a Snowflake compute warehouse, or \code{NULL} to use
the default.}

\item{database}{The name of a Snowflake database, or \code{NULL} to use the
default.}

\item{schema}{The name of a Snowflake database schema, or \code{NULL} to use the
default.}

\item{uid, pwd}{Manually specify a username and password for authentication.
Specifying these options will disable ambient credential discovery.}

\item{connection_name}{The name of a connection defined in your Snowflake
\code{connections.toml} file, or \code{NULL} to use the default connection.}

\item{...}{Further arguments passed on to \code{\link[=dbConnect]{dbConnect()}}.}
}
\value{
An \code{OdbcConnection} object with an active connection to a Snowflake
account.
}
\description{
Connect to a Snowflake account via the \href{https://docs.snowflake.com/en/developer-guide/odbc/odbc}{Snowflake ODBC driver}.

In particular, the custom \code{dbConnect()} method for the Snowflake ODBC driver
supports the \code{connections.toml} and \code{config.toml} files used by the Snowflake
Connector for Python and the Snowflake CLI via the \pkg{snowflakeauth}
package. It also detects ambient OAuth credentials on platforms like Snowpark
Container Services or Posit Workbench. Finally, it can detect viewer-based
credentials on Posit Connect if the \pkg{connectcreds} package is installed.

In addition, on macOS platforms, the \code{dbConnect} method will check and warn
if it detects irregularities with how the driver is configured, unless the
\code{odbc.no_config_override} environment variable is set.
}
\examples{
\dontrun{
# Use ambient credentials.
DBI::dbConnect(odbc::snowflake())

# Use browser-based SSO (if configured). Only works on desktop.
DBI::dbConnect(
  odbc::snowflake(),
  account = "testorg-test_account",
  authenticator = "externalbrowser"
)

# Use a traditional username & password.
DBI::dbConnect(
  odbc::snowflake(),
  account = "testorg-test_account",
  uid = "me",
  pwd = rstudioapi::askForPassword()
)

# Use a named connection from a connections.toml file.
DBI::dbConnect(odbc::snowflake(), connection_name = "test")

# Use credentials from the viewer (when possible) in a Shiny app
# deployed to Posit Connect.
library(connectcreds)
server <- function(input, output, session) {
  conn <- DBI::dbConnect(odbc::snowflake())
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connection-pane.R
\name{odbcConnectionIcon}
\alias{odbcConnectionIcon}
\title{Get an icon representing a connection.}
\usage{
odbcConnectionIcon(connection)
}
\arguments{
\item{connection}{A connection object, as returned by \code{dbConnect()}.}
}
\value{
The path to an icon file on disk.
}
\description{
Return the path on disk to an icon representing a connection.
}
\details{
The icon returned should be a 32x32 square image file.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odbc-connection.R
\name{odbcConnectionColumns}
\alias{odbcConnectionColumns}
\title{odbcConnectionColumns}
\usage{
odbcConnectionColumns(conn, name, ..., exact = FALSE)
}
\arguments{
\item{conn}{OdbcConnection}

\item{name}{Table identifiers.
By default, are interpreted as a ODBC search pattern where \verb{_} and \verb{\%} are
wild cards. Set \code{exact = TRUE} to match \verb{_} exactly.}

\item{...}{additional parameters to methods}

\item{exact}{Set to \code{TRUE} to escape \verb{_} in identifier names so that it
matches exactly, rather than matching any single character. \verb{\%} always
matches any number of characters as this is unlikely to appear in a
table name.}
}
\value{
data.frame with columns
\itemize{
\item name
\item field.type - equivalent to type_name in SQLColumns output
\item table_name
\item schema_name
\item catalog_name
\item data_type
\item column_size
\item buffer_length
\item decimal_digits
\item numeric_precision_radix
\item remarks
\item column_default
\item sql_data_type
\item sql_datetime_subtype
\item char_octet_length
\item ordinal_position
\item nullable
}
}
\description{
For a given table this function returns detailed information on
all fields / columns.  The expectation is that this is a relatively thin
wrapper around the ODBC \code{SQLColumns} function call, with some of the field
names renamed / re-ordered according to the return specifications below.
}
\details{
In \code{\link[DBI:dbWriteTable]{DBI::dbWriteTable()}} we make a call to this method
to get details on the fields of the table we are writing to.  In particular
the columns \code{data_type}, \code{column_size}, and \code{decimal_digits} are used.  An
implementation is not necessary for \code{\link[DBI:dbWriteTable]{DBI::dbWriteTable()}} to work.

\code{odbcConnectionColumns} is routed through the \code{SQLColumns} ODBC
method.  This function, together with remaining catalog functions
(\code{SQLTables}, etc), by default ( \code{SQL_ATTR_METADATA_ID == false} ) expect
either ordinary arguments (OA) in the case of the catalog, or pattern value
arguments (PV) in the case of schema/table name.  For these, quoted
identifiers do not make sense, so we unquote identifiers prior to the call.
}
\seealso{
The ODBC documentation on
\href{https://learn.microsoft.com/en-us/sql/odbc/reference/syntax/sqlcolumns-function}{SQLColumns}
for further details.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/driver-redshift.R
\docType{class}
\name{redshift}
\alias{redshift}
\alias{RedshiftOdbcDriver-class}
\alias{dbConnect,RedshiftOdbcDriver-method}
\title{Helper for Connecting to Redshift via ODBC}
\usage{
redshift()

\S4method{dbConnect}{RedshiftOdbcDriver}(
  drv,
  clusterId,
  server,
  database,
  region = NULL,
  driver = NULL,
  uid = NULL,
  pwd = NULL,
  dbUser = uid,
  ...
)
}
\arguments{
\item{drv}{an object that inherits from \link[DBI:DBIDriver-class]{DBI::DBIDriver},
or an existing \link[DBI:DBIConnection-class]{DBI::DBIConnection}
object (in order to clone an existing connection).}

\item{clusterId}{The Redshift cluster identifier. Only one of \code{clusterId} or
the more verbose \code{server} is required.}

\item{server}{The full hostname of the Redshift cluster.}

\item{database}{The name of the Redshift database to connect to.}

\item{region}{The AWS region the Redshift cluster runs in. Ignored when
\code{server} is provided. Defaults to the value of the environment variable
\code{AWS_REGION}, then \code{AWS_REGION}, or uses \code{us-east-1} if both are unset.}

\item{driver}{The name of or path to a Redshift ODBC driver, or \code{NULL} to
locate one automatically.}

\item{uid, pwd}{Disable IAM credentials and manually specify a username and
password for authentication.}

\item{dbUser}{The Redshift database account.}

\item{...}{Further arguments passed on to \code{\link[=dbConnect]{dbConnect()}}.}
}
\value{
An \code{OdbcConnection} object with an active connection to a Redshift
cluster or SQL warehouse.
}
\description{
Connect to Redshift clusters via ODBC.

In particular, the custom \code{dbConnect()} method for Redshift ODBC drivers
automatically determines whether IAM-based credentials are available, much
like other AWS SDKs and tools. This requires the \pkg{paws.common} package.
}
\examples{
\dontrun{
# Connect to Redshift using IAM credentials.
DBI::dbConnect(
  odbc::redshift(),
  clusterId = "my-testing-cluster",
  database = "dev",
  dbUser = "me"
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odbc-package.R
\docType{package}
\name{odbc-package}
\alias{odbc-package}
\title{odbc: Connect to ODBC Compatible Databases (using the DBI Interface)}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A DBI-compatible interface to ODBC databases.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://odbc.r-dbi.org}
  \item \url{https://github.com/r-dbi/odbc}
  \item \url{https://solutions.posit.co/connections/db/}
  \item Report bugs at \url{https://github.com/r-dbi/odbc/issues}
}

}
\author{
\strong{Maintainer}: Hadley Wickham \email{hadley@posit.co}

Authors:
\itemize{
  \item Jim Hester
  \item Oliver Gjoneski
  \item Simon Couch \email{simon.couch@posit.co}
}

Other contributors:
\itemize{
  \item lexicalunit (nanodbc library) [copyright holder]
  \item Google Inc. (cctz library) [copyright holder]
  \item Posit Software, PBC [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbi-result.R
\docType{methods}
\name{OdbcResult}
\alias{OdbcResult}
\alias{OdbcResult-class}
\alias{dbClearResult,OdbcResult-method}
\alias{dbFetch,OdbcResult-method}
\alias{dbHasCompleted,OdbcResult-method}
\alias{dbIsValid,OdbcResult-method}
\alias{dbGetStatement,OdbcResult-method}
\alias{dbColumnInfo,OdbcResult-method}
\alias{dbGetRowCount,OdbcResult-method}
\alias{dbGetRowsAffected,OdbcResult-method}
\alias{dbBind,OdbcResult-method}
\title{Odbc Result Methods}
\usage{
\S4method{dbClearResult}{OdbcResult}(res, ...)

\S4method{dbFetch}{OdbcResult}(res, n = -1, ...)

\S4method{dbHasCompleted}{OdbcResult}(res, ...)

\S4method{dbIsValid}{OdbcResult}(dbObj, ...)

\S4method{dbGetStatement}{OdbcResult}(res, ...)

\S4method{dbColumnInfo}{OdbcResult}(res, ...)

\S4method{dbGetRowCount}{OdbcResult}(res, ...)

\S4method{dbGetRowsAffected}{OdbcResult}(res, ...)

\S4method{dbBind}{OdbcResult}(res, params, ..., batch_rows = getOption("odbc.batch_rows", NA))
}
\arguments{
\item{res}{An object inheriting from \link[DBI:DBIResult-class]{DBI::DBIResult}.}

\item{...}{Other arguments passed on to methods.}

\item{n}{maximum number of records to retrieve per fetch. Use \code{n = -1}
or \code{n = Inf}
to retrieve all pending records.  Some implementations may recognize other
special values.}

\item{dbObj}{An object inheriting from \code{DBIObject}, i.e. \code{DBIDriver},
\code{DBIConnection}, or a \code{DBIResult}.}

\item{params}{For \code{dbBind()}, a list of values, named or unnamed,
or a data frame, with one element/column per query parameter.
For \code{dbBindArrow()}, values as a nanoarrow stream,
with one column per query parameter.}

\item{batch_rows}{The number of rows to retrieve. Defaults to \code{NA}, which
is set dynamically to the minimum of 1024 and the size of the input.
Depending on the database, driver, dataset and free memory, setting this
to a lower value may improve performance.}
}
\description{
Implementations of pure virtual functions defined in the \code{DBI} package
for OdbcResult objects.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/driver-databricks.R
\docType{class}
\name{databricks}
\alias{databricks}
\alias{DatabricksOdbcDriver-class}
\alias{dbConnect,DatabricksOdbcDriver-method}
\title{Helper for Connecting to Databricks via ODBC}
\usage{
databricks()

\S4method{dbConnect}{DatabricksOdbcDriver}(
  drv,
  httpPath,
  workspace = Sys.getenv("DATABRICKS_HOST"),
  useNativeQuery = TRUE,
  driver = NULL,
  HTTPPath,
  uid = NULL,
  pwd = NULL,
  ...
)
}
\arguments{
\item{drv}{an object that inherits from \link[DBI:DBIDriver-class]{DBI::DBIDriver},
or an existing \link[DBI:DBIConnection-class]{DBI::DBIConnection}
object (in order to clone an existing connection).}

\item{httpPath, HTTPPath}{To query a cluster, use the HTTP Path value found
under \verb{Advanced Options > JDBC/ODBC} in the Databricks UI. For SQL
warehouses, this is found under \verb{Connection Details} instead.}

\item{workspace}{The URL of a Databricks workspace, e.g.
\code{"https://example.cloud.databricks.com"}.}

\item{useNativeQuery}{Suppress the driver's conversion from ANSI SQL 92 to
HiveSQL? The default (\code{TRUE}), gives greater performance but means that
paramterised queries (and hence \code{dbWriteTable()}) do not work.}

\item{driver}{The name of the Databricks ODBC driver, or \code{NULL} to use the
default name.}

\item{uid, pwd}{Manually specify a username and password for authentication.
Specifying these options will disable automated credential discovery.}

\item{...}{Further arguments passed on to \code{\link[=dbConnect]{dbConnect()}}.}
}
\value{
An \code{OdbcConnection} object with an active connection to a Databricks
cluster or SQL warehouse.
}
\description{
Connect to Databricks clusters and SQL warehouses via the \href{https://www.databricks.com/spark/odbc-drivers-download}{Databricks ODBC driver}.

In particular, the custom \code{dbConnect()} method for the Databricks ODBC driver
implements a variant of Databricks's \href{https://docs.databricks.com/aws/en/dev-tools/auth/unified-auth}{unified authentication}
model when no \code{uid} or \code{pwd} is supplied, checking for ambient credentials in
the following order:
\itemize{
\item Viewer-based or service principal credentials supplied by Posit Connect
(requires the \pkg{connectcreds} package).
\item Personal access tokens.
\item Workload identity federation.
\item OAuth machine-to-machine credentials.
\item OAuth user-to-machine credentials suplied via Posit Workbench or the
Databricks CLI on desktop.
}

This aims to provide broad compatibility between \pkg{odbc} and the standard
environment variables used by Databricks SDKs in R and other languages.

In addition, on macOS platforms, the \code{dbConnect()} method will check
for irregularities with how the driver is configured,
and attempt to fix in-situ, unless the \code{odbc.no_config_override}
environment variable is set.
}
\examples{
\dontrun{
DBI::dbConnect(
  odbc::databricks(),
  httpPath = "sql/protocolv1/o/4425955464597947/1026-023828-vn51jugj"
)

# Use credentials from the viewer or a service principal (when possible) in
# a Shiny app deployed to Posit Connect.
library(connectcreds)
server <- function(input, output, session) {
  conn <- DBI::dbConnect(
    odbc::databricks(),
    workspace = "https://example.cloud.databricks.com",
    httpPath = "sql/protocolv1/o/4425955464597947/1026-023828-vn51jugj"
  )
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odbc-connection.R
\docType{data}
\name{SUPPORTED_CONNECTION_ATTRIBUTES}
\alias{SUPPORTED_CONNECTION_ATTRIBUTES}
\alias{ConnectionAttributes}
\title{Supported Connection Attributes}
\description{
These (pre) connection attributes are supported and can be passed as
part of the \code{dbConnect} call in the named list \code{attributes} parameter:
}
\details{
\itemize{
\item \code{azure_token}: This should be a string scalar; in particular Azure Active
Directory authentication token.  Only for use with Microsoft SQL Server and
with limited support away from the OEM Microsoft driver.
\item \code{sf_private_key}: This parameter is specific to establishing a connection
to \code{snowflake} and is understood by both OEM, as well as \code{Posit}
pro drivers.  Argument should be a string (scalar); in particular a
PEM-encoded private key.  Note,
if using private key authentication, the \code{authenticator} connection
string attribute must be set to \code{SNOWFLAKE_JWT}.
Using this \emph{connection} attribute is an alternative to using the
\code{PRIV_KEY_FILE} connection string attribute.
\item \code{sf_private_key_password}: If key passed using \code{sf_private_key} is
encrypted, you can use this attribute to communicate the password.
}
}
\examples{
\dontrun{
conn <- dbConnect(
  odbc::odbc(),
  dsn = "my_azure_mssql_db",
  Encrypt = "yes",
  attributes = list("azure_token" = .token)

conn <- dbConnect(
 odbc::odbc(),
 dsn = "snowflake",
 attributes = list("sf_private_key" = paste(readLines("<path-to-private-key-file>"), collapse="\n"),
                   "sf_private_key_password" = "<optional-private-key-encryption-password>"),
 authenticator = "SNOWFLAKE_JWT")
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connection-pane.R
\name{odbcPreviewObject}
\alias{odbcPreviewObject}
\title{Preview the data in an object.}
\usage{
odbcPreviewObject(connection, rowLimit, ...)
}
\arguments{
\item{connection}{A connection object, as returned by \code{dbConnect()}.}

\item{rowLimit}{The maximum number of rows to display.}

\item{...}{Parameters specifying the object.}
}
\value{
A data frame containing the data in the object.
}
\description{
Return the data inside an object as a data frame.
}
\details{
The object to previewed must be specified as one of the arguments
(e.g. \code{table = "employees"}); depending on the driver and underlying
data store, additional specification arguments may be required.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbi-driver.R
\name{odbc}
\alias{odbc}
\alias{dbConnect,OdbcDriver-method}
\alias{dbConnect}
\title{Connect to a database via an ODBC driver}
\usage{
odbc()

\S4method{dbConnect}{OdbcDriver}(
  drv,
  dsn = NULL,
  ...,
  timezone = "UTC",
  timezone_out = "UTC",
  encoding = "",
  name_encoding = "",
  bigint = c("integer64", "integer", "numeric", "character"),
  timeout = 10,
  driver = NULL,
  server = NULL,
  database = NULL,
  uid = NULL,
  pwd = NULL,
  dbms.name = NULL,
  attributes = NULL,
  interruptible = getOption("odbc.interruptible", interactive()),
  .connection_string = NULL
)
}
\arguments{
\item{drv}{An \code{OdbcDriver}, from \code{odbc()}.}

\item{dsn}{The data source name. For currently available options, see the
\code{name} column of \code{\link[=odbcListDataSources]{odbcListDataSources()}} output.}

\item{...}{Additional ODBC keywords. These will be joined with the other
arguments to form the final connection string.

Note that ODBC parameter names are case-insensitive so that (e.g.) \code{DRV}
and \code{drv} are equivalent. Since this is different to R and a possible
source of confusion, odbc will error if you supply multiple arguments that
have the same name when case is ignored.

Any values containing a leading or trailing space, a \code{=}, \verb{;}, \verb{\{},
or \verb{\}} are likely to require quoting. Use \code{\link[=quote_value]{quote_value()}} for a fairly
standard approach or see your driver documentation for specifics.}

\item{timezone}{The server time zone. Useful if the database has an internal
timezone that is \emph{not} 'UTC'. If the database is in your local timezone,
set this argument to \code{\link[=Sys.timezone]{Sys.timezone()}}. See \code{\link[=OlsonNames]{OlsonNames()}} for a complete
list of available time zones on your system.
Note, if the datatype itself carries timezone information, as is the case,
for example, with SQL Server::DATETIMEOFFSET, \code{package:odbc} will make
an effort to respect the timezone declared therein.  In those cases,
this parameter is not used - the timezone that is part of the datatype
takes precedence.}

\item{timezone_out}{The time zone returned to R. If you want to display
datetime values in the local timezone, set to \code{\link[=Sys.timezone]{Sys.timezone()}}.}

\item{encoding}{The text encoding used on the Database. If the database is
not using UTF-8 you will need to set the encoding to get accurate
re-encoding. See \code{\link[=iconvlist]{iconvlist()}} for a complete list of available encodings
on your system. Note strings are always returned \code{UTF-8} encoded.}

\item{name_encoding}{The text encoding for column names used on the
Database.  May be different than the \code{encoding} argument.  Defaults to
empty string which is equivalent to returning the column names without
performing any conversion.}

\item{bigint}{The R type that \code{SQL_BIGINT} types should be mapped to.
Default is \link[bit64:bit64-package]{bit64::integer64}, which allows the full range of 64 bit
integers.}

\item{timeout}{Time in seconds to timeout the connection attempt. Setting a
timeout of \code{Inf} or \code{NA} indicates no timeout. Defaults to 10 seconds.}

\item{driver}{The ODBC driver name or a path to a driver. For currently
available options, see the \code{name} column of \code{\link[=odbcListDrivers]{odbcListDrivers()}} output.}

\item{server}{The server hostname. Some drivers use \code{Servername} as the name
for this argument. Not required when configured for the supplied \code{dsn}.}

\item{database}{The database on the server. Not required when configured for
the supplied \code{dsn}.}

\item{uid}{The user identifier. Some drivers use \code{username} as the name
for this argument. Not required when configured for the supplied \code{dsn}.}

\item{pwd}{The password. Some drivers use \code{password} as the name
for this argument. Not required when configured for the supplied \code{dsn}.}

\item{dbms.name}{The database management system name. This should normally
be queried automatically by the ODBC driver. This name is used as the class
name for the OdbcConnection object returned from \code{\link[=dbConnect]{dbConnect()}}. However, if
the driver does not return a valid value, it can be set manually with this
parameter.}

\item{attributes}{A list of connection attributes that are passed
prior to the connection being established. See \link{ConnectionAttributes}.}

\item{interruptible}{Logical.  If \code{TRUE} calls to \code{SQLExecute} and
\code{SQLExecuteDirect} can be interrupted when the user sends SIGINT ( ctrl-c ).
Otherwise, they block.  Defaults to \code{TRUE} in interactive sessions, and
\code{FALSE} otherwise.  It can be set explicitly either by manipulating this
argument, or by setting the global option \code{odbc.interruptible}.}

\item{.connection_string}{A complete connection string, useful if you are
copy pasting it from another source. If this argument is used, any
additional arguments will be appended to this string.}
}
\description{
The \code{dbConnect()} method documented here is invoked when \code{\link[DBI:dbConnect]{DBI::dbConnect()}}
is called with the first argument \code{odbc()}. Connecting to a database via
an ODBC driver is likely the first step in analyzing data using the odbc
package; for an overview of package concepts, see the \emph{Overview} section
below.
}
\section{Connection strings}{


Internally, \code{dbConnect()} creates a connection string using the supplied
arguments. Connection string keywords are driver-dependent; the arguments
documented here are common, but some drivers may not accept them.

Alternatively to configuring DSNs and driver names with the driver manager,
you can pass a complete connection string directly as the
\code{.connection_string} argument.
\href{https://www.connectionstrings.com}{The Connection Strings Reference} is a
useful resource that has example connection strings for a large variety of
databases.
}

\section{Overview}{


The odbc package is one piece of the R interface to databases with support
for ODBC:

\figure{whole-game.png}

The package supports any \strong{Database Management System (DBMS)} with ODBC
support. Support for a given DBMS is provided by an \strong{ODBC driver}, which
defines how to interact with that DBMS using the standardized syntax of ODBC
and SQL. Drivers can be downloaded from the DBMS vendor or, if you're a Posit
customer, using the \href{https://docs.posit.co/pro-drivers/}{professional drivers}.
To manage information about each driver and the data sources they provide
access to, our computers use a \strong{driver manager}. Windows is bundled with
a driver manager, while MacOS and Linux require installation of one; this
package supports the \href{https://www.unixodbc.org/}{unixODBC} driver manager.

In the \strong{R interface}, the \href{https://dbi.r-dbi.org/}{DBI package} provides a
front-end while odbc implements a back-end to communicate with the driver
manager. The odbc package is built on top of the
\href{https://nanodbc.github.io/nanodbc/}{nanodbc} C++ library.

Interfacing with DBMSs using R and odbc involves three high-level steps:
\enumerate{
\item \emph{Configure drivers and data sources}: the functions \code{\link[=odbcListDrivers]{odbcListDrivers()}},
\code{\link[=odbcListDataSources]{odbcListDataSources()}}, and \code{\link[=odbcListConfig]{odbcListConfig()}} help to interface with the
driver manager.
\item \emph{Connect to a database}: The \code{\link[=dbConnect]{dbConnect()}} function, called with the
first argument odbc(), connects to a database using the specified ODBC
driver to create a connection object.
\item \emph{Interface with connections}: The resulting connection object can be
passed to various functions to retrieve information on database
structure ([DBI::dbListTables()][]), iteratively develop queries
(\code{\link[DBI:dbSendQuery]{DBI::dbSendQuery()}}, \code{\link[DBI:dbColumnInfo]{DBI::dbColumnInfo()}}), and query data objects
(\code{\link[DBI:dbFetch]{DBI::dbFetch()}}).
}
}

\section{Learn more}{


To learn more about databases:
\itemize{
\item \href{https://solutions.posit.co/connections/db/}{"Best Practices in Working with Databases"}
documents how to use the odbc package with various popular databases.
\item \href{https://github.com/mkleehammer/pyodbc/wiki/Drivers-and-Driver-Managers}{The pyodbc "Drivers and Driver Managers" Wiki}
provides further context on drivers and driver managers.
\item \href{https://learn.microsoft.com/en-us/sql/odbc/reference}{Microsoft's "Introduction to ODBC"}
is a thorough resource on the ODBC interface.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connection-pane.R
\name{odbcListObjectTypes}
\alias{odbcListObjectTypes}
\title{Return the object hierarchy supported by a connection.}
\usage{
odbcListObjectTypes(connection)
}
\arguments{
\item{connection}{A connection object, as returned by \code{dbConnect()}.}
}
\value{
The hierarchy of object types supported by the connection.
}
\description{
Lists the object types and metadata known by the connection, and how those
object types relate to each other.
}
\details{
The returned hierarchy takes the form of a nested list, in which each object
type supported by the connection is a named list with the following
attributes:

\describe{
\item{contains}{A list of other object types contained by the object, or
"data" if the object contains data}
\item{icon}{An optional path to an icon representing the type}
}
For instance, a connection in which the top-level object is a schema that
contains tables and views, the function will return a list like the
following:

\preformatted{list(schema = list(contains = list(
                   list(name = "table", contains = "data")
                   list(name = "view", contains = "data"))))

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odbc.R
\name{odbcSetTransactionIsolationLevel}
\alias{odbcSetTransactionIsolationLevel}
\title{Set the Transaction Isolation Level for a Connection}
\usage{
odbcSetTransactionIsolationLevel(conn, levels)
}
\arguments{
\item{conn}{A \link[DBI:DBIConnection-class]{DBI::DBIConnection} object, as returned by
\code{dbConnect()}.}

\item{levels}{One or more of 'read_uncommitted', 'read_committed', 'repeatable_read', 'serializable'.}
}
\description{
Set the Transaction Isolation Level for a Connection
}
\examples{
\dontrun{
# Can use spaces or underscores in between words.
odbcSetTransactionIsolationLevel(con, "read uncommitted")

# Can also use the full constant name.
odbcSetTransactionIsolationLevel(con, "SQL_TXN_READ_UNCOMMITTED")
}
}
\seealso{
\url{https://learn.microsoft.com/en-us/sql/odbc/reference/develop-app/setting-the-transaction-isolation-level}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/driver-snowflake.R
\name{odbcConnectionColumns,Snowflake,character-method}
\alias{odbcConnectionColumns,Snowflake,character-method}
\alias{dbExistsTableForWrite,Snowflake,character-method}
\title{Connecting to Snowflake via ODBC}
\usage{
\S4method{dbExistsTableForWrite}{Snowflake,character}(conn, name, ..., catalog_name = NULL, schema_name = NULL)
}
\arguments{
\item{conn}{A \link[DBI:DBIConnection-class]{DBI::DBIConnection} object, as returned by
\code{dbConnect()}.}

\item{name}{The table name, passed on to \code{\link[DBI:dbQuoteIdentifier]{dbQuoteIdentifier()}}. Options are:
\itemize{
\item a character string with the unquoted DBMS table name,
e.g. \code{"table_name"},
\item a call to \code{\link[DBI:Id]{Id()}} with components to the fully qualified table name,
e.g. \code{Id(schema = "my_schema", table = "table_name")}
\item a call to \code{\link[DBI:SQL]{SQL()}} with the quoted and fully qualified table name
given verbatim, e.g. \code{SQL('"my_schema"."table_name"')}
}}

\item{...}{Other parameters passed on to methods.}

\item{catalog_name, schema_name}{Catalog and schema names.}
}
\description{
\subsection{\code{odbcConnectionColumns()}}{

If the catalog, or the schema arguments are NULL, attempt to infer
by querying for CURRENT_DATABASE() and CURRENT_SCHEMA().  We do this
to aid with performance, as the SQLColumns method is more performant
when restricted to a particular DB/schema.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connection-pane.R
\name{odbcListColumns}
\alias{odbcListColumns}
\title{List columns in an object.}
\usage{
odbcListColumns(connection, ...)
}
\arguments{
\item{connection}{A connection object, as returned by \code{dbConnect()}.}

\item{...}{Parameters specifying the object.}
}
\value{
A data frame with \code{name} and \code{type} columns, listing the
object's fields.
}
\description{
Lists the names and types of each column (field) of a specified object.
}
\details{
The object to inspect must be specified as one of the arguments
(e.g. \code{table = "employees"}); depending on the driver and underlying
data store, additional specification arguments may be required.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/driver-sql-server.R
\docType{class}
\name{Microsoft SQL Server-class}
\alias{Microsoft SQL Server-class}
\alias{dbUnquoteIdentifier,Microsoft SQL Server,SQL-method}
\alias{isTempTable,Microsoft SQL Server,character-method}
\alias{isTempTable,Microsoft SQL Server,SQL-method}
\alias{dbExistsTable,Microsoft SQL Server,character-method}
\alias{dbListTables,Microsoft SQL Server-method}
\alias{dbExistsTable,Microsoft SQL Server,Id-method}
\alias{dbExistsTable,Microsoft SQL Server,SQL-method}
\alias{odbcConnectionSchemas,Microsoft SQL Server-method}
\alias{sqlCreateTable,Microsoft SQL Server-method}
\alias{odbcConnectionColumns,Microsoft SQL Server,character-method}
\alias{odbcConnectionColumns,Microsoft SQL Server,SQL-method}
\title{SQL Server}
\description{
Details of SQL Server methods for odbc and DBI generics.

\subsection{\code{dbUnquoteIdentifier()}}{

\code{conn@quote} returns the quotation mark, but quotation marks and square
brackets can be used interchangeably for delimited identifiers.
(\url{https://learn.microsoft.com/en-us/sql/relational-databases/databases/database-identifiers}).
This function strips the brackets first and then calls the DBI method to
strip the quotation marks.
}

\subsection{\code{isTempTable()}}{

Local temp tables are stored as
\verb{[tempdb].[dbo].[#name]_____[numeric identifier]}, so \code{isTempTable()}
returns \code{TRUE} if \code{catalog_name} is \code{"tempdb"} or \code{"\%"}, or the
name starts with \code{"#"}.
}

\subsection{\code{dbExistsTable()}}{

The default implementation reports temporary tables as non-existent
since they live in a different catalog. This method provides a special
case for temporary tables, as identified by \code{isTempTable()}.
}

\subsection{\code{dbListTables()}}{

The default implementation reports temporary tables as non-existent
when a \code{catalog_name} isn't supplied since they live in a different catalog.
This method provides a special case for temporary tables.
}

\subsection{\code{odbcConnectionSchemas}}{

Method for an internal function. Calls catalog-specific \code{sp_tables} to make
sure we list the schemas in the appropriate database/catalog.
}

\subsection{\code{sqlCreateTable()}}{

Warns if \code{temporary = TRUE} but the \code{name} does not conform to temp table
naming conventions (i.e. it doesn't start with \verb{#}).
}

\subsection{\code{odbcConnectionColumns()}}{

If temp table, query the database for the
actual table name.
}

\subsection{\code{odbcConnectionColumns()}}{

Copied over from odbc-connection to avoid S4 dispatch NOTEs.
}
}
\details{
Note on binding \href{https://learn.microsoft.com/en-us/sql/relational-databases/tables/use-table-valued-parameters-database-engine}{TVPs}:
You can bind \code{data.frame}s to SQL Server TVPs when executing stored procedures with the following caveats:
\itemize{
\item All non-df parameters must be of length 1; and
\item The \code{batch_rows} parameter (to \code{dbBind}, for example) should be set to 1.
}
}
\examples{
\dontrun{
# Bind `data.frame` to a TVP when executing a
# stored procedure that takes three parameters:
# a TVP, an INT and a VARCHAR(max).
res <- dbGetQuery(conn, "{ CALL example_sproc(?, ?, ?) }",
  params = list(df.data, 100, "Lorem ipsum dolor sit amet"))
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/driver-oracle.R
\docType{class}
\name{Oracle-class}
\alias{Oracle-class}
\alias{sqlCreateTable,Oracle-method}
\alias{odbcConnectionTables,Oracle,character-method}
\alias{odbcConnectionColumns,Oracle,character-method}
\title{Oracle}
\usage{
\S4method{sqlCreateTable}{Oracle}(
  con,
  table,
  fields,
  row.names = NA,
  temporary = FALSE,
  ...,
  field.types = NULL
)

\S4method{odbcConnectionTables}{Oracle,character}(
  conn,
  name,
  catalog_name = NULL,
  schema_name = NULL,
  table_type = NULL,
  exact = FALSE
)
}
\description{
Details of Oracle methods for odbc and DBI generics.

\subsection{\code{odbcConnectionTables()}}{

Method for an internal function that otherwise relies on the \code{SQLTables}
ODBC API. While this method is much faster than the OEM implementation, it
does not look through synonyms.
}

\subsection{\code{odbcConnectionColumns()}}{

Query, rather than use \code{SQLColumns} ODBC API, since we bind a \code{BIGINT} to
one of the column results and Oracle's OEM driver can't handle it.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connection-pane.R
\name{odbcPreviewQuery}
\alias{odbcPreviewQuery}
\alias{odbcPreviewQuery.OdbcConnection}
\alias{odbcPreviewQuery.Microsoft SQL Server}
\alias{odbcPreviewQuery.Teradata}
\alias{odbcPreviewQuery.Oracle}
\title{Create a preview query.}
\usage{
odbcPreviewQuery(connection, rowLimit, name)

\method{odbcPreviewQuery}{OdbcConnection}(connection, rowLimit, name)

\method{odbcPreviewQuery}{`Microsoft SQL Server`}(connection, rowLimit, name)

\method{odbcPreviewQuery}{Teradata}(connection, rowLimit, name)

\method{odbcPreviewQuery}{Oracle}(connection, rowLimit, name)
}
\arguments{
\item{connection}{A connection object, as returned by \code{dbConnect()}.}

\item{rowLimit}{The maximum number of rows to display.}

\item{name}{Name of the object to be previewed}
}
\description{
Optimize against the rowLimit argument.  S3 since some
back-ends do not parse the LIMIT syntax.  Internal, not expected that
users would interact with this method.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbi-table.R
\name{DBI-tables}
\alias{DBI-tables}
\alias{dbWriteTable,OdbcConnection,character,data.frame-method}
\alias{dbWriteTable,OdbcConnection,Id,data.frame-method}
\alias{dbWriteTable,OdbcConnection,SQL,data.frame-method}
\alias{dbAppendTable,OdbcConnection-method}
\alias{sqlCreateTable,OdbcConnection-method}
\title{Convenience functions for reading/writing DBMS tables}
\usage{
\S4method{dbWriteTable}{OdbcConnection,character,data.frame}(
  conn,
  name,
  value,
  overwrite = FALSE,
  append = FALSE,
  temporary = FALSE,
  row.names = NULL,
  field.types = NULL,
  batch_rows = getOption("odbc.batch_rows", NA),
  ...
)

\S4method{dbWriteTable}{OdbcConnection,Id,data.frame}(
  conn,
  name,
  value,
  overwrite = FALSE,
  append = FALSE,
  temporary = FALSE,
  row.names = NULL,
  field.types = NULL,
  batch_rows = getOption("odbc.batch_rows", NA),
  ...
)

\S4method{dbWriteTable}{OdbcConnection,SQL,data.frame}(
  conn,
  name,
  value,
  overwrite = FALSE,
  append = FALSE,
  temporary = FALSE,
  row.names = NULL,
  field.types = NULL,
  batch_rows = getOption("odbc.batch_rows", NA),
  ...
)

\S4method{dbAppendTable}{OdbcConnection}(
  conn,
  name,
  value,
  batch_rows = getOption("odbc.batch_rows", NA),
  ...,
  row.names = NULL
)

\S4method{sqlCreateTable}{OdbcConnection}(
  con,
  table,
  fields,
  row.names = NA,
  temporary = FALSE,
  ...,
  field.types = NULL
)
}
\arguments{
\item{conn}{An \link{OdbcConnection} object, produced by \code{\link[DBI:dbConnect]{DBI::dbConnect()}}.}

\item{name}{a character string specifying a table name. Names will be
automatically quoted so you can use any sequence of characters, not
just any valid bare table name.}

\item{value}{A data.frame to write to the database.}

\item{overwrite}{Allow overwriting the destination table. Cannot be
\code{TRUE} if \code{append} is also \code{TRUE}.}

\item{append}{Allow appending to the destination table. Cannot be
\code{TRUE} if \code{overwrite} is also \code{TRUE}.}

\item{temporary}{If \code{TRUE}, will generate a temporary table.}

\item{row.names}{Either \code{TRUE}, \code{FALSE}, \code{NA} or a string.

If \code{TRUE}, always translate row names to a column called "row_names".
If \code{FALSE}, never translate row names. If \code{NA}, translate
rownames only if they're a character vector.

A string is equivalent to \code{TRUE}, but allows you to override the
default name.

For backward compatibility, \code{NULL} is equivalent to \code{FALSE}.}

\item{field.types}{Additional field types used to override derived types.}

\item{batch_rows}{The number of rows to retrieve. Defaults to \code{NA}, which
is set dynamically to the minimum of 1024 and the size of the input.
Depending on the database, driver, dataset and free memory, setting this
to a lower value may improve performance.}

\item{...}{Other arguments used by individual methods.}

\item{con}{A database connection.}

\item{table}{The table name, passed on to \code{\link[DBI:dbQuoteIdentifier]{dbQuoteIdentifier()}}. Options are:
\itemize{
\item a character string with the unquoted DBMS table name,
e.g. \code{"table_name"},
\item a call to \code{\link[DBI:Id]{Id()}} with components to the fully qualified table name,
e.g. \code{Id(schema = "my_schema", table = "table_name")}
\item a call to \code{\link[DBI:SQL]{SQL()}} with the quoted and fully qualified table name
given verbatim, e.g. \code{SQL('"my_schema"."table_name"')}
}}

\item{fields}{Either a character vector or a data frame.

A named character vector: Names are column names, values are types.
Names are escaped with \code{\link[DBI:dbQuoteIdentifier]{dbQuoteIdentifier()}}.
Field types are unescaped.

A data frame: field types are generated using
\code{\link[DBI:dbDataType]{dbDataType()}}.}
}
\description{
Convenience functions for reading/writing DBMS tables
}
\examples{
\dontrun{
library(DBI)
con <- dbConnect(odbc::odbc())
dbListTables(con)
dbWriteTable(con, "mtcars", mtcars, temporary = TRUE)
dbReadTable(con, "mtcars")

dbListTables(con)
dbExistsTable(con, "mtcars")

# A zero row data frame just creates a table definition.
dbWriteTable(con, "mtcars2", mtcars[0, ], temporary = TRUE)
dbReadTable(con, "mtcars2")

dbDisconnect(con)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odbc-connection.R
\name{quote_value}
\alias{quote_value}
\title{Quote special character when connecting}
\usage{
quote_value(x)
}
\arguments{
\item{x}{A string to quote.}
}
\value{
A quoted string, wrapped in \code{I()}.
}
\description{
When connecting to a database using odbc, all the arguments are concatenated
into a single connection string that looks like \verb{name1=value1;name2=value2}.
That means if your value contains \code{=} or \verb{;} then it needs to be quoted.
Other rules mean that you need to quote any values that starts or ends with
white space, or contains \verb{\{} or \verb{\}}.

This function quotes a string in a way that should work for most drivers,
but unfortunately there doesn't seem to be an approach that works everywhere.
If this function doesn't work for you, you'll need to carefully read the
docs for your driver.
}
\examples{
quote_value("abc")
quote_value("ab'c")

# Real usage is more likely to look like:
\dontrun{
library(DBI)

con <- dbConnect(
  odbc::odbc(),
  dsn = "reallycooldatabase"
  password = odbc::quote_value(Sys.getenv("MY_PASSWORD"))
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbi-connection.R
\name{dbListTables,OdbcConnection-method}
\alias{dbListTables,OdbcConnection-method}
\title{List remote tables and fields for an ODBC connection}
\usage{
\S4method{dbListTables}{OdbcConnection}(
  conn,
  catalog_name = NULL,
  schema_name = NULL,
  table_name = NULL,
  table_type = NULL,
  ...
)
}
\arguments{
\item{conn}{A \link[DBI:DBIConnection-class]{DBI::DBIConnection} object, as returned by
\code{dbConnect()}.}

\item{catalog_name, schema_name, table_name}{Catalog, schema, and table names.

By default, \code{catalog_name}, \code{schema_name} and \code{table_name} will
automatically escape underscores to ensure that you match exactly one
table. If you want to search for multiple tables using wild cards, you
will need to use \code{odbcConnectionTables()} directly instead.}

\item{table_type}{The type of the table to return, the default returns
all table types.}

\item{...}{Other parameters passed on to methods.}
}
\value{
A character vector of table or field names respectively.
}
\description{
\code{dbListTables()} provides names of remote tables accessible through this
connection; \code{dbListFields()} provides names of columns within a table.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbi-driver.R
\docType{class}
\name{OdbcDriver}
\alias{OdbcDriver}
\alias{OdbcDriver-class}
\alias{show,OdbcDriver-method}
\alias{dbDataType,OdbcDriver,ANY-method}
\alias{dbDataType,OdbcDriver,list-method}
\alias{dbDataType,OdbcDriver,data.frame-method}
\alias{dbIsValid,OdbcDriver-method}
\alias{dbGetInfo,OdbcDriver-method}
\title{Odbc Driver Methods}
\usage{
\S4method{show}{OdbcDriver}(object)

\S4method{dbDataType}{OdbcDriver,ANY}(dbObj, obj, ...)

\S4method{dbDataType}{OdbcDriver,list}(dbObj, obj, ...)

\S4method{dbDataType}{OdbcDriver,data.frame}(dbObj, obj, ...)

\S4method{dbIsValid}{OdbcDriver}(dbObj, ...)

\S4method{dbGetInfo}{OdbcDriver}(dbObj, ...)
}
\arguments{
\item{object}{Any R object}

\item{dbObj}{A object inheriting from \link[DBI:DBIDriver-class]{DBI::DBIDriver}
or \link[DBI:DBIConnection-class]{DBI::DBIConnection}}

\item{obj}{An R object whose SQL type we want to determine.}

\item{...}{Other arguments passed on to methods.}
}
\description{
Implementations of pure virtual functions defined in the \code{DBI} package
for OdbcDriver objects.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odbc-drivers.R
\name{odbcListDrivers}
\alias{odbcListDrivers}
\title{List Configured ODBC Drivers}
\usage{
odbcListDrivers(
  keep = getOption("odbc.drivers_keep"),
  filter = getOption("odbc.drivers_filter")
)
}
\arguments{
\item{keep, filter}{A character vector of driver names to keep in or remove
from the results, respectively. If \code{NULL}, all driver names will be kept,
or none will be removed, respectively. The \code{odbc.drivers_keep} and
\code{odbc.drivers_filter} options control the argument defaults.

Driver names are first processed with \code{keep}, then \code{filter}. Thus, if a
driver name is in both \code{keep} and \code{filter}, it won't appear in output.}
}
\value{
A data frame with three columns.

\describe{
\item{name}{Name of the driver. The entries in this column can be
passed to the \code{driver} argument of \code{\link[=dbConnect]{dbConnect()}} (as long as the
driver accepts the argument).}
\item{attribute}{Driver attribute name.}
\item{value}{Driver attribute value.}
}

If a driver has multiple attributes, there will be one row per attribute,
each with the same driver \code{name}. If a given driver name does not have any
attributes, the function will return one row with the driver \code{name}, but
the last two columns will be \code{NA}.
}
\description{
Collect information about the configured driver names. A driver must be both
installed and configured with the driver manager to be included in this list.
Configuring a driver name just sets up a lookup table (e.g. in
\code{odbcinst.ini}) to allow users to pass only the driver name to \code{\link[=dbConnect]{dbConnect()}}.

Driver names that are not configured with the driver manager (and thus
do not appear in this function's output) can still be
used in \code{\link[=dbConnect]{dbConnect()}} by providing a path to a driver directly.
}
\section{Configuration}{


This function interfaces with the driver manager to collect information
about the available driver names.

For \strong{MacOS and Linux}, the odbc package supports the unixODBC driver
manager. unixODBC looks to the \code{odbcinst.ini} \emph{configuration file} for
information on driver names. Find the location(s) of your \code{odbcinst.ini}
file(s) with \code{odbcinst -j}.

In this example \code{odbcinst.ini} file:

\if{html}{\out{<div class="sourceCode">}}\preformatted{[MySQL Driver]
Driver=/opt/homebrew/Cellar/mysql/8.2.0_1/lib/libmysqlclient.dylib
}\if{html}{\out{</div>}}

Then the driver name is \verb{MySQL Driver}, which will appear in the \code{name}
column of this function's output. To pass the driver name as the \code{driver}
argument to \code{\link[=dbConnect]{dbConnect()}}, pass it as a string, like \code{"MySQL Driver"}.

\strong{Windows} is \href{https://learn.microsoft.com/en-us/sql/odbc/admin/odbc-data-source-administrator}{bundled}
with an ODBC driver manager.

In this example, function output would include 1 row: the \code{name} column
would read \code{"MySQL Driver"}, \code{attribute} would be \code{"Driver"}, and \code{value}
would give the file path to the driver. Additional key-value pairs
under the driver name would add additional rows with the same \code{name} entry.

When a driver is configured with a driver manager, information on the driver
will be automatically passed on to \code{\link[=dbConnect]{dbConnect()}} when its \code{driver} argument
is set. For an example, see the same section in the \code{\link[=odbcListDataSources]{odbcListDataSources()}}
help-file. Instead of configuring driver information with a driver manager,
it is also possible to provide a path to a driver directly to \code{\link[=dbConnect]{dbConnect()}}.
}

\examples{
\dontshow{if (FALSE) withAutoprint(\{ # examplesIf}
odbcListDrivers()
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=odbcListDataSources]{odbcListDataSources()}}
}
